/*!
 * @file fcp_datamgr.h
 * @brief header file for the external interface for the fcp_datamgr.c
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FCP_DATAMGR_H_
#define FCP_DATAMGR_H_


//--------------------------------------------------------------------
// Include Files
//--------------------------------------------------------------------

// will store the object pointer to the connected camera server in the
// protocol data structure passed as userdata see per_session_data_fcp
#include "fi_cs.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// the FCP Protocol handler has a size allocated for payloads on the WSS this
// is configurable in the protocol definition of the WSS server module (test-server.c)
// currently.  This current value also happens to be the default payload size for a
// wss protocol handler of LWS_MAX_SOCKET_IO_BUF - and it is noted that there are pre
// and post 8-bit bytes that are also used in the buffer
#define FCP_MAX_RX_BUFFER           4096

// the FCP Listener Protocol handler has a size allocated for payloads on the WSS this
// is configurable in the protocol definition of the WSS server module (test-server.c)
// currently.  note that there are pre and post 8-bit bytes that are also used in buffer
#define FCP_LISTENER_MAX_RX_BUFFER  1024

/* --------------------------------------------------------------------------------
 * each fcp_request corresponds to a requested load or store on a fcp register
 * of the Fastec Imaging Camera Server
 * --------------------------------------------------------------------------------*/
#define FCP_REQUEST_STATE_EMPTY     0   // request package is available to use
#define FCP_REQUEST_STATE_READY     1   // request package has a request ready
#define FCP_REQUEST_STATE_SENT      2   // request package has sent a request

/* --------------------------------------------------------------------------------
 * each fcp_response contains the result of a requested load or store on a fcp register
 * of the Fastec Imaging Camera Server
 * --------------------------------------------------------------------------------*/
#define FCP_RESPONSE_STATE_EMPTY     0  // response package is available to use
#define FCP_RESPONSE_STATE_READY     1  // response package has a response ready
#define FCP_RESPONSE_STATE_SENT      2  // response package has sent a response

/* --------------------------------------------------------------------------------
 * each fcp_response contains the result of a requested load or store on a fcp register
 * of the Fastec Imaging Camera Server
 * --------------------------------------------------------------------------------*/
#define FCP_REQUEST_ACTION_NONE     0  // action type is not set
#define FCP_REQUEST_ACTION_LOAD     1  // action type is a fcp load request
#define FCP_REQUEST_ACTION_STORE    2  // action type is a fcp store request
#define FCP_REQUEST_ACTION_NOTIFY   4  // action type is a fcp notify request

//--------------------------------------------------------------------
// Data Structures and Definitions
//--------------------------------------------------------------------

/* --------------------------------------------------------------------------------
 * each fcp_request is capable of containing multiple GIGE fields for loading
 * or storing dynamically, this structure is used to manage the memory associated
 * with each request identifying the Camera Server information necessary for request
 * --------------------------------------------------------------------------------*/
#define FCP_MAX_GIGEREG_LABEL       64  // maximum size for a fcp GIGE register label in string format
#define FCP_MAX_GIGEREG_VALUE       16  // maximum size for a fcp GIGE register value in string format
typedef struct
{
    // the name of the GIGE Register in use
    char cs_reg_name[FCP_MAX_GIGEREG_LABEL];
    // the camera server register value
    uint32_t cs_reg;
    // is this a memory or register request - memory will be an array of 8 bit data and
    // register is simply one 32 bit register if array setting is 1 if not setting is 0
    uint8_t is_register_an_array;
    // if is_register_an_array is 1 and this value is 1 response string is encoded b64 if 0 it is not
    uint8_t encode_response;
    // if is_register_an_array is 0 then the data is stored in cs_reg_data
    uint32_t cs_reg_data;
    // how many bytes exist in the cs_reg_array
    uint32_t cs_reg_array_len;
    // if is_register_an_array is 1 then the data is stored in allocated cs_reg_array
    uint8_t *p_cs_reg_array;
    // if the data in this field record has been set due to a notification change this
    // state must be identified - if set to 1 it means notification change applies if set
    // to 0 it means that notification change does not apply and it was a direct read from CS
    uint8_t is_this_notification;
} fcp_request_field;


/* --------------------------------------------------------------------------------
 * each fcp_request corresponds to a requested load, store, or notification
 * on a fcp register of the Fastec Imaging Camera Server
 * --------------------------------------------------------------------------------*/
typedef struct
{
    // the state of this request, is it empty, ready to be processed, or in process of sending response
    uint16_t state;
    // if this request record is in error this will keep track of what error occurred
    // it will be one of the settings maintained in the FCP_PKGSTATUS_* settings
    uint16_t pkgstatus;
    // if the JSON parser return code is stored here
    int json_return_code;
    // if the Camera Server returned an error the code will be stored here
    uint16_t fi_cs_error;
    // what type of action request is this?  loading, storing, or notifying
    uint8_t action_type;
    // how many fields have been requested
    uint16_t nfields;
    // an allocated collection of ptrs to multiple fcp_request_field data records
    // the value in nfields indicates how many field request records exist initially 0
    fcp_request_field **p_fcp_request_fields;
} fcp_request;

/* --------------------------------------------------------------------------------
 * each session connected to the fcp protocol has a unique set of data that
 * belongs to that session - for the FCP protocol each session is allowed a fixed
 * number of requests for servicing as set in MAX_FCP_PER_SESSION
 * --------------------------------------------------------------------------------*/
#define MAX_FCP_PER_SESSION 4     /* number of requests allowed per session connection*/
typedef struct
{
    fiCameraServerIO* p_fiCS;				// the connected camera server IO object pointer or NULL if not available
    // collects all of the register requests for a fcp connected
    // request session
    fcp_request session_requests[MAX_FCP_PER_SESSION];
} per_session_data_fcp;

/* --------------------------------------------------------------------------------
 * each session connected to the fcp-listener-protocol has a unique set of data that
 * belongs to that session - for the FCP listener protocol each session will only
 * allow for a single listener - listening to specific notifications or all notifications
 * --------------------------------------------------------------------------------*/
#define MAX_FI_PENDING_NOTIFICATIONS 32    /* should never need more than this as writeable happens 2 times a second */

// in the case of an ALL notification state there needs to be a list created that can store
// overlapping asynchronous notifications -- this list is small and will have a maximum size
// because servicing will occur when writeable moments in time are available in the wss
#define FI_ASYNC_NOTIFY_AVAILABLE       0   // async notification record available
#define FI_ASYNC_NOTIFY_CS_RECEIVED     1   // async notification record has received cs info waiting to write
#define FI_ASYNC_NOTIFY_BUILT_FOR_WRITE 2   // async notification record has put info into response build
typedef struct
{
    // what state is this entry in one of the FI_ASYNC_NOTIFY_* states
    uint8_t notify_state;
    // this will contain the notification register that is ready to write or 0 if none
    uint32_t cs_reg;
    // this will contain the data for the notification register that is ready to write
    uint32_t cs_data;
} fcp_listener_notifications;

// THERE IS ONLY ONE OF THESE PER SESSION -- IF THIS NEEDS TO BE CHANGED IN THE FUTURE
// THAN THE fcp_listener_request NEEDS TO BE TURNED INTO AN ARRAY AND MANAGED
typedef struct
{
    fiCameraServerIO* p_fiCS;			//!< The connected camera server IO object pointer or NULL if not available.
    uint8_t listening;                      	//!< Set to 1 if listening 0 of not.
    uint8_t catch_all_notifications;        	//!< Set to 1 if catching all notifications or 0 if specified fields.

    //! If looking for all notifications, this array holds the async register information
    //! when data is received from Camera Server before it can be written to WSS client.
    fcp_listener_notifications all_async_data[MAX_FI_PENDING_NOTIFICATIONS];

    //! Because order should be maintained in the queue store position info in the async data.
    //! Fills queue then wraps around starts at 0.
    uint8_t async_pos;

    //! This will contain initial FCP information for notification processing.
    //! Note in the case of ALL notifications that only one field record exists otherwise one or
    //! more fields can exist in the request record depending on client's JSON request.
    //! What FCP registers are we listening to?
    fcp_request request;

    //! This is the libewebsocket interface record all session data structures maintain.
    //! Identifies the web socket interface.
    struct libwebsocket *wsi;
} per_session_data_fcp_listener;


// ///////////////////////////////////////////////////////////////////
//--------------------------------------------------------------------
// External Interfaces
//--------------------------------------------------------------------
// ///////////////////////////////////////////////////////////////////

//--------------------------------------------------------------------
// External Interfaces for fcp-listener-protocol
//--------------------------------------------------------------------

extern void cs_fcp_reg_notify( void* arg, uint32_t reg, uint32_t val );

//----------------------------------------------------------------------
// Handlers for the management of the fcp-listener-protocol listener
// queue that maintains the list of session data pointers actively listening
//----------------------------------------------------------------------

// call once during program initialization to create empty fcp listener queue
// this is the array of session data pointers to active wss clients that want
// particular cs notifications
extern bool fi_init_fcp_listeners();

// called after when fcp-listener-protocol session has been created and
// the session data pointer needs to be added to the listener queue
extern int fcp_add_listener(per_session_data_fcp_listener *pss);

// called when the fcp-listener-protocol session is being destroyed
// and the session data pointer needs to be removed from the listener queue
// and any allocated memory released
extern bool fcp_listener_cleanup(per_session_data_fcp_listener *pss);

// called when a new request is made to an existing fcp-listener-protocol
// session and the protocol needs to determine if the session data pointer
// exists in the listener queue already
extern int is_pss_in_fcp_listening_queue(per_session_data_fcp_listener *pss);

//----------------------------------------------------------------------
// Handlers for the management of the fcp-listening-protocol session
// data structure itself
//----------------------------------------------------------------------

// called when fcp-listener-protocol session has been established and
// it is also called by the fcp_reset_listener function to reset for
// a new requests - it's common initialization logic
extern int fcp_init_listener(per_session_data_fcp_listener *pss);

// called by the fcp-listener-protocol when a request for notifications
// is received by the wss client - it will re-initialize the session data
// for the new request
extern int fcp_reset_listener(per_session_data_fcp_listener *pss);

// responsible for building up the ALL response buffer for notification requests
// that are looking to monitor ALL notifcation entries.
extern int
fcp_bld_all_notifications(per_session_data_fcp_listener *pss, uint8_t *p_fcp_buffer, uint16_t fcp_max_bufsize );

extern void fcp_listeners_lock();

extern void fcp_listeners_unlock();

//--------------------------------------------------------------------
// External Interfaces for fcp-protocol
//--------------------------------------------------------------------

// entry point to initialize fcp protocol session data
extern bool init_fcp_session(per_session_data_fcp *pss);

// called to pull out a pointer to the available request record to use for
// the fcp protocol session
extern fcp_request* fcp_get_empty_request(per_session_data_fcp *pss);

// entry point to reset all session requests and free up any allocated memory
// this is called from the websocket server in response to the fcp protocol
// reason = LWS_CALLBACK_PROTOCOL_DESTROY
extern bool reset_all_session_requests( per_session_data_fcp *pss );

// return the next available request in the queue if one exists or null if not
extern fcp_request* fcp_get_next_request(per_session_data_fcp *pss);

// return the number of pending requests in this session queue
extern int fcp_get_nbr_session_requests(per_session_data_fcp *pss);

// post request processing
extern void fcp_request_sent(per_session_data_fcp *pss, fcp_request* p_request);

//--------------------------------------------------------------------
// General Interfaces Used By All FCP Protocol Handlers
//--------------------------------------------------------------------

// entry point to parse the request data received by the fcp protocol handler
extern bool fcp_parse_request(fcp_request *p_request, void* p_json_pkg, bool is_fcp );

// called when a JSON response needs to be created and returned to the caller
extern int fcp_handle_response(fcp_request *p_request, uint8_t *p_fcp_buffer, uint16_t fcp_max_bufsize );

#endif	/* !FCP_DATAMGR_H_ */
