/*
 * fi_buffer.h
 *
 *  Created on: Jan 25, 2016
 *      Author: eric
 */

#ifndef FI_BUFFER_H_
#define FI_BUFFER_H_

#include <stdint.h>
#include <libwebsockets.h>

#define CEILING(x,y) ( (y) * (((x) + (y) - 1) / (y)))

/**
 * @brief Buffers for Websocket streaming/download protocols
 *
 * This Buffer class, together with the associated BufferQueue class, serves as
 * the data path between the Fastec LSSM steaming interface and WssProtocol
 * implementations that support streaming image data through Websockets
 * connections.
 *
 * At one end of the data path, the LSSM interface is the Producer of image
 * data. At the other end of the path the Websockets server is the consumer.
 *
 * <H3>Alignment and Padding</H3>
 * Buffers are allocated to maintain 4K page alignment on the Buffer's base
 * address. The size of the data area available at the base address is given
 * by Buffer::BUFF_DATA_SIZE.
 *
 * Buffer's also implement <I>prefix padding</I> below
 * the base address. The space available for prefix padding is given by
 * Buffer::BUFF_PREFIX_SIZE. The prefix area is sized to accommodate a 2048-byte
 * (assumed max) image file header. The prefix area also transparently
 * accommodates additional space for the liobwesockets LWS_SEND_BUFFER_PRE_PADDING
 * area.
 *
 * Finally, Buffers include space for post-padding as required by libwebsockets.
 * Note however that libwebsockets no longer actually requires post-padding. The
 * libwebsockets LWS_SEND_BUFFER_POST_PADDING macro evaluates to zero.
 * Buffer::POST_BUFF_SIZE is therefore also zero.
 *
 * <H3>Fastec WebSocket Streaming Frame Header</H3>
 * The Fastec WebSocket streaming protocols place a Frame Header in front of
 * the binary data that presents each image frame. The Frame Header gives the
 * total number of bytes in the frame data (including an image header, when
 * an image header is present) and all the pixel data. The Frame header can
 * also carry a frame identifier (frame number) for streaming protocoles that
 * include such a notion. The Buffer class include accessor functions for these
 * parameters. They are not part of the Buffer's data area.
 */
class Buffer
{
public:

    static const int BUFF_DATA_SIZE    = (64 * 1024); //!< Space available for data
    static const int BUFF_PREFIX_SIZE  = 2048;        //!< Space available for image header

    Buffer();
    virtual ~Buffer();

    /** @name Getter methods: */
    ///@{
    char* getDataPtr(size_t offset);           //!< @brief Get base address of Buffer data
    uint32_t getFlags();                       //!< @brief Get protocol-specific flag word
    size_t getSize();                          //!< @brief Get current size of Buffer data
    uint32_t get_frame_data_size( void );      //!< @brief Get Fastec WebSocket stream header frame data size
    int32_t get_frame_id( void );              //!< @brief Get Fastec WebSocket stream header frame ID
    ///@}

    /** @name Setter methods (generally used by Producer): */
    ///@{
    void setSize(size_t s);                    //!< @brief Set current size of Buffer data
    void prefix_reserve( size_t size );        //!< @brief Set space reserved for file header prefix
    void setFlags(uint32_t f);                 //!< @brief Set protocol-specific flag word
    void set_frame_data_size( uint32_t fds );  //!< @brief Set Fastec WebSocket stream header frame data size
    void set_frame_id( int32_t n );            //!< @brief Set Fastec WebSocket stream header frame ID
    ///@}

    /** @name BufferQueue methods to contorl ownership: */
    ///@{
    void setProducer();                        //!< @brief Declare this Buffer owned by Producer
    Buffer* isProducer();                      //!< @brief Return true iff this Buffer is owned by Producer
    void setConsumer();                        //!< @brief Declare this Buffer owned by Consumer
    Buffer* isConsumer();                      //!< @brief Return true iff this Buffer is owned by Consumer
    ///@}

private:
    static const int BUFF_ALIGN        = 4096;        //!< (Assumed) system memory page size
    static const int PRE_BUFF_SIZE     = CEILING(LWS_SEND_BUFFER_PRE_PADDING + BUFF_PREFIX_SIZE, BUFF_ALIGN);
    static const int POST_BUFF_SIZE    = CEILING(LWS_SEND_BUFFER_POST_PADDING, BUFF_ALIGN);

    /** @brief Size, in bytes, requested from malloc() */
    static const int BUFF_SIZE         = PRE_BUFF_SIZE + BUFF_DATA_SIZE + POST_BUFF_SIZE;

    /**
     * @brief Buffer owners
     *
     * Each Buffer is owned by the Producer (places data into a Buffer)
     * or the Consumer (removes data from the Buffer), but never both
     * at the same time.
     *
     */
    typedef enum
    {
        PRODUCER,
        CONSUMER
    } OwnerType;

    char* data_malloc;          //!< Buffer base address, as returned from malloc()
    char* data;                 //!< Buffer base address, as set by Producer
    size_t size;                //!< Data size (bytes) as set by Producer
    OwnerType owner;            //!< Current owner of this Buffer
    uint32_t flags;             //!< Protocol-defined flags
    uint32_t frame_data_size;   //!< Fastec WebSocket streaming header data size for this frame
    int32_t id;                 //!< Fastec WebSocket streaming header ID
};

#endif /* FI_BUFFER_H_ */
