/*
 * fi_buffer_q.h
 *
 *  Created on: Jan 25, 2016
 *      Author: eric
 */

#ifndef FI_BUFFER_Q_H_
#define FI_BUFFER_Q_H_

#include <stdint.h>
#include <pthread.h>

#include "fastec/lssm.h"
#include "fi_buffer.h"

class StreamingSession;

/**
 * @brief BufferQueue: Interface between LSSM thread and the main WSS thread
 *
 * Instances of this class provide the interface between the LSSM (camera's
 * Local Streaming Shared Memory interface) thread and the main WSS thread
 * (where the main libwebsockets loop and Fastec websockets protocols execute).
 * There is one BufferQueue instance for each Fastec WebSocket streaming
 * protocol supported by the Fastec WebSocket Server.
 *
 * The BufferQueue class actually manages two queues:
 *     - A queue of Buffers, which itself may be considered two queues.
 *       There is a set of empty Buffers that belong to the Producer, and
 *       another set of full buffers that belong to the Consumer(s). As
 *       the Producer fills buffers with image data, it adds them to the
 *       Consumers' input queue. As the Consumer StreamingSessions finish
 *       transmitting image data, they add empty Buffers to the Producer's
 *       input queue.
 *     - A queue of Streaming Sessions waiting to transmit image data from
 *       the Consumer Buffer currently at the head of the Consumer Buffer
 *       queue. When a new Consumer Buffer arrives at the head of the
 *       Consumer queue, all the currently active StreamingSesisons are
 *       placed on the session Tx queue. As each session finished
 *       transmitting, it is removed from the session Tx queue. When the
 *       Sesison Tx queue becomes empty, the now-empty Consumer Buffer is
 *       handed back to the Producer, and a new Consumer Buffer can arrive
 *       at the head of the Consumer queue.
 *
 *The diagram below illustrates the Scope of BufferQueue's responsibilities:
 *
 * \image html BufferQueue_overview.svg
 *
 * \note The diagram suggests that many Buffer instances may circulate between
 * Producer and consumer, and that many StreamingSessions may be active. In
 * practice (so far) BufferQueues are configured with only two Buffer instances
 * each. The number of active session will vary among the Fastec streaming
 * protocols, but generally a camera will be able to effectively support only
 * two or three streaming sessions.
 */
class BufferQueue
{

public:
    /** @brief The number of Buffers that circulate between Producer and Consumer */
    static const int NUM_BUFFS = 2;

    BufferQueue();
    virtual ~BufferQueue();

    /** @name Consumer methods: */
    //! @{
    /**
     * @brief Transfer ownership of the current Consumer Buffer to the Producer
     *
     * This executes in the Consumer thread, and will indicate that the Consumer
     * Buffer currently at the head of the Consumer queue is now empty, and should
     * be given back to the Producer. If the Producer queue was previously empty,
     * this method will also signal the Producer thread that an empty Buffer is
     * now available on its Buffer queue.
     */
    virtual void finished();

    /**
     * @brief Move all Consumer Buffers immediately to the Producer
     *
     * This executes in the Consumer thread. Any and all currently queued
     * Consumer Buffers will be immediately handed back to the Producer by
     * placing them on the Producer queue. If the Producer queue was previously
     * empty, this method will also signal the Producer thread that an empty
     * Buffer is now available on its Buffer queue.
     */
    virtual void flush();

    /**
     * @brief Place a StreamingSession onto the sesison Tx queue
     *
     * This executes in the Consumer thread. It will place a StreamingSession
     * onto the session transmit (Tx) queue, allowing it to transmit data
     * from the Buffer at the head of the Consumer queue. This method will
     * also arrange for the StreamingSesison's write() callback to execute
     * at a future time that its WebSocket is available for writing.
     *
     * @param s Points to the StreamingSession that gets added to the Tx queue
     */
    void fill(StreamingSession* s);

    /**
     * @brief Test head of the Consumer queue
     *
     * This returns true if and only if there is a Consumer-owned buffer at
     * the head of the Consumer queue. A return value false implies that the
     * Consumer queue is empty.
     *
     * @return true iff there is a Consumer-owned buffer at the head of
     *         the Consumer queue
     */
    bool consumerOwnsCurrent();
    //! @}

    //! @name Fault flag accessors
    //! @{

    //! @brief Set fault indicator for this BufferQueue
    void set_fault( bool f );

    //! @brief Get fault indicator for this BufferQeueu
    bool get_fault( void );
    //! @}

    /** @name Consumer methods: */
    ///@{
    uint32_t getCurrentConsumerFlags();
    bool txQueueIsEmpty();
    size_t getSize();
    char* consumerGetDataPtr(size_t offset);
    size_t consumer_get_size( void );
    size_t consumer_get_size( size_t sf );
    int32_t consumer_get_frame_id( void );
    uint32_t consumer_get_frame_data_size( void );
    int consumer_index( void );
    void removeFromTxQueue(StreamingSession* s);
    bool isOnTxQueue(StreamingSession* s);
    bool get_oldest(void);
    void set_oldest(bool mode);
    ///@}

    /** @name Producer methods: */
    ///@{
    bool producer_owns_current();
    void producer_wait( void );
    void producer_set_flags(uint32_t f);
    void producer_set_size(size_t s);
    void producer_set_frame_id( int32_t n );
    void producer_set_frame_data_size(uint32_t fds);
    void producer_set_signal( void (*func)( void* ), void* arg );
    char* producer_get_data_ptr(size_t offset);
    void producer_reserve_prefix( size_t size );
    int producer_finished();
    int producer_index( void );
    ///@}

    /** @name Other methods */
    ///@{
    int get_producer_buffer_count( void );    /**< @brief Return number of Buffers currently owned by Producer */
    ///@}

    /** @name Transfer Parameters, managed by the Producer Process (LSSM thread)
     *
     *  The transfer parameters define a buffer in System Memory that contains
     *  pixel data ready to transfer into one or more Buffers.  These parameters
     *  also track progress of the transfer, as sections of the source image are
     *  transferred into successive Buffers. It is normally the case that the
     *  source image is far larger than the data size in a single Buffer, so multiple
     *  transfer operations are required to process one image frame.
     */
    ///@{
    size_t image_size;
    size_t bytes_to_read_from_im;
    size_t bytes_to_read_from_header;
    uint32_t frame_num;
    uint64_t xfer_base;             	//!< Base address of the frame in unmapped system memory.
    size_t xfer_offset;			//!< Byte offset into the frame.
    int buf_id;                        	//!< LSSM Buffer ID, where applicable.
    bool fault;                        	//!< LSSM has declared a fault for this protocol.
    ///@}

    volatile bool buffer_queue_trace;

private:
    //! @brief Serialize access to the Buffer queue state
    pthread_mutex_t lock;

    //! @name Producer/Consumer queues
    //!
    //! All the Buffer instances reside in an array of Buffers, treated as
    //! a circular list. The list incorporates two logical queues: one queue
    //! of Producer-owned buffers and one queue of Consumer-owned buffers.
    //! Either queue may be empty, but at all times all the Buffers are on one
    //! or the other of the queues.
    //!
    //! Buffers are moved from one queue to the other by manipulating their
    //! Buffer::owner field, and the producerOwns, consumerCurrent, and
    //! producerCurrent variables declared below.
    //!
    static const int PRODUCER          = 0;
    static const int CONSUMER          = 1;

    pthread_cond_t producer_check_ready;

    //! @brief All the Buffer instances for this BufferQueue
    Buffer b[NUM_BUFFS];

    //! @brief Number of Buffers on the Producer queue
    //!
    //! Implies that the number of Buffers on the Consumer queue is
    //! always NUM_BUFFS - producerOwns.
    //!
    int producerOwns;

    //! @brief Head of the Consumer queue
    //!
    //! Only valid if the Consumer queue is not empty. When the Consumer
    //! queue is not empty, b[consumerCurrent] will be owned by the consumer,
    //! and producerOwns will be less than NUM_BUFS.
    //!
    int consumerCurrent;

    //! @brief Head of the Producer queue
    //!
    //! Only valid if the Producer queue is not empty. When the Producer
    //! queue is not empty, b[producerCurrent] will be owned by the producer,
    //! and producerOwns will be greater than 0.
    //!
    int producerCurrent;

    size_t size;
    bool get_oldest_flag;

    //! @name Consumer queue
    //!
    //! This is the queue of streaming data consumers. Consumers (streaming
    //! protocol sessions) enter the queue when a new Buffer arrives at the
    //! head of the consumer Buffer queue. They leave the consumer queue after
    //! transmitting all the data in the Buffer, or if their !WebSocket
    //! connection closes (perhaps due to a network error, or because the
    //! client has initiated a disconnect).
    //!
    //! The queue itself is formed as a linked list of StreamingSession
    //! instances.
    //! @{

    //! @brief Head of consumer queue
    StreamingSession* txQueueHead;

    //! @brief Tail of consumer queue
    StreamingSession* txQueueTail;
    //! @}

    void (*producer_signal)( void* arg );
    void* producer_signal_arg;
};

#endif /* FI_BUFFER_Q_H_ */
