/*!
 * @file fi_cs.h
 * @brief header file for the external interface for the fi_cs.cpp
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_CSIO_H
#define FI_CSIO_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <stdint.h>
#include <string>

#include "fcp.h"

using namespace std;

//--------------------------------------------------------------------
// Forward Class Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Constant Declarations or Enumerations Used in class
//--------------------------------------------------------------------

//! States for the Camera Server IO object.
enum FI_CSIO_STATES
{
    // Always first in the list
    FI_CSIO_UNINITIALIZED = 0,		//!< Empty uninitialized state.
    FI_CSIO_CONNECTED,			//!< The cs is initialized and ready.
    FI_CSIO_INIT_FAILED,		//!< The cs initialziation failed.
    FI_CSIO_MISSING_PORT,		//!< There was no port specified for CS use.
    FI_CSIO_MISSING_IP,			//!< There was no IP specified for CS use.
    FI_CSIO_STATE_COUNT			//!< The number of states.
};

//--------------------------------------------------------------------
// Class: fiCameraServerIO
//--------------------------------------------------------------------

class fiCameraServerIO
{

public:
    // Camera Server IO constructor
    // creates a Fastec Imaging Camera Server I/O object
    explicit fiCameraServerIO();

    // Camera Server IO destructor
    // cleans up and destructs a web socket server object
    ~fiCameraServerIO();

    // Is the Camera Server Connected?
    // common call to determine if the camera server is connected and
    // therefore ready to use.
    bool isCSConnected();


    // Initializes Camera Server Connection
    // responsible for initializing and setting up camera server connection
    FI_CSIO_STATES initCSConnection( const char *in_p_cam_ip, int in_cam_port );

    // Camera Address accessor method
    // returns the ip address of the camera in array pointer form */
    const char *    getCameraIPStr( void )
    {
        return cam_ip_arg.c_str();
    }

    // Camera Address accessor method
    // returns the ip address of the camera in array pointer form */
    uint32_t getCameraIP( void )
    {
        return cam_ip;
    }

    // Camera Port accessor method
    // returns the 32 bit camera port to use for camera server interaction
    unsigned short  getCameraPort( void )
    {
        return cam_port;
    }

    // Camera Server connection is closed and object reflects
    // this state
    void quitCSConnection( void );

    // Camera Server FCP Read Register Request
    // returns 1 if successful or 0 if failed
    // passes through call to cs_get_reg_l
    int readFCPReg( uint32_t reg, uint32_t *val );

    // Camera Server FCP Read Register Request
    // return 1 on success, 0 on failure. On failure, refer to the contents of istat.
    // passes through call to cs_set_reg_l
    int writeFCPReg( uint32_t reg, uint32_t val, int *istat );

    // Camera Server FCP Read Array Data from FCP Register Request
    // returns number of bytes read or < 0 if error
    // passes through to call cs_read_array
    int readFCPArray( int array_id, void *data_ptr, uint16_t data_len );

    // Camera Server FCP Read Array Data from FCP Register Request
    // returns number of bytes read or < 0 if error
    // passes through to call cs_write_array
    int writeFCPArray( int array_id, const void *data_ptr, uint16_t data_len );

    // The Camera Server IO object interfaces with fcp.c and in that module
    // there is only support for one registered notification handler.  In order
    // to differentiate between FCP register notifications that are specific
    // to FCP data versus Live View Image ready notifications, the ability to
    // register different callback handlers was added to fcp.c - these next
    // methods will take care of registration for each type of notification
    // callback handler.


    // the FCP Register change notification handler for wss is registered
    // thru this interface - note still uses CS_REG_CALLBACK function signature
    int regFCPListener( CS_REG_CALLBACK fcp_listener );

private:
    std::string     cam_ip_arg;             //!< Camera IP argument in string form
    uint32_t        cam_ip;                 //!< Camera we're interested in.
    unsigned short  cam_port;               //!< Network port we're supposed to connect to.

    FI_CSIO_STATES	cs_state;				//!< what state is camera server interface
};

#endif // FI_CSIO_H
