/*!
 * @file fi_jsonpkg.h
 * @brief header file for the external interface for the fi_jsonpkg.c
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdbool.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/types.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// the core of a JSON package is the key or data value and this structure represents
// the data maintained for these value pairs
typedef struct
{
    uint16_t len;
    uint8_t *p_buf;
} json_buf, *json_buf_ptr;

// each JSON element can be comprised of a key and value pair.  sometimes the key is
// not present as in the case of arrays with one primary key and many values contained
// within [] brackets and separated by commas.  This data structure maintains a key
// value pair within a json record
typedef struct
{
    json_buf key;
    json_buf value;
} json_rec, *json_rec_ptr;


// if a JSON package contains an array it will have a primary key and a series of
// key value pairs where the key may or may not be present
#define MAX_JSON_ARRAY_SIZE 128	// can only request 128 array items at a time
typedef struct
{
    json_buf array_key;			// primary key for the array
    uint8_t array_items_inuse;	// how many array items in use
    uint8_t array_items_size;	// max size of array items
    json_rec array_items[MAX_JSON_ARRAY_SIZE];
} json_array_rec, *json_array_rec_ptr;

//--------------------------------------------------------------------
// External Interfaces and data
//--------------------------------------------------------------------

// the JSON parsing package is the structure provided to this module's external interface
// allowing for consumers to query and interpret the data parsed out of a WSS JSON request
//
// At this time the structure contains all of the informaiton parsed from
// the input buffer but perhaps it may make more sense to use it more as an
// interruptable parse where each call can read a key value token out of the
// input stream -- this structure could be used to store internal state to restart
// the internal JSON parser of this module.  Maybe this is a more efficient use
// of memory?  TODO - decide which is better.  Typically the JSON packages will
// be lite - small in size so for now the implementation is to pull it all into
// this package structure which is initialized, populated, and freed after use

#define MAX_FI_JSON_KEYS	100	// can only request 100 key-value records at a time
#define MAX_FI_JSON_ARRAYS	1	// can only request 1 array per FI Request

typedef struct
{
    // list of normal key/value pairs
    uint8_t key_values_inuse;	// how many key value pairs in use
    uint8_t key_values_size;		// max size for key value pairs
    json_rec key_values[MAX_FI_JSON_KEYS];

    // list of array data - there may or may not be an embedded array
    uint8_t json_arrays_inuse;		// how many array records in use in the list
    uint8_t json_arrays_size;		// max size available for arrays
    json_array_rec json_arrays[MAX_FI_JSON_ARRAYS];
} fi_json_parse_pkg, *fi_json_parse_pkg_ptr;


/* initialize output parsing data structure for persistent result*/
//int init_json_parse(json_parsed *json_parsed_ptr);
extern uint8_t init_fi_json_parse_pkg( fi_json_parse_pkg_ptr parsed_pkg_ptr );

/*
 * Find and free() allocated memory attached to caller's fi_json_pkg
 * (but not the fi_json_pkg itself)
 */
//int free_json_parse(json_parsed *json_parsed_ptr);
extern void free_fi_json_parse_pkg( fi_json_parse_pkg_ptr pkg );

/* entry point where a json object is passed in for intrepretation */
/* the output is stored in the json_parsed structure in key/value pairs */
extern int fi_json_parse(const char *json_input,  fi_json_parse_pkg_ptr parsed_pkg_ptr );

// find a value associated with a key in a parsed JSON package
extern const uint8_t * find_value_for_key( fi_json_parse_pkg_ptr parsed_pkg_ptr, uint8_t* key_ptr);

// return a pointer to the array primay key value for an array index returns null if no array
// exists otherwise it will return a pointer to the key value in the list which also may be null
extern const uint8_t * get_array_key( fi_json_parse_pkg_ptr parsed_pkg_ptr, uint16_t array_key_index);



/* translate an error code to a readable error string */
const /*@observer@*/char *json_error_string(int);

void json_enable_debug(int, FILE *);
#ifdef __cplusplus
}
#endif

#define JSON_ERR_OBSTART	1	/* non-WS when expecting object start */
#define JSON_ERR_ATTRSTART	2	/* non-WS when expecting attrib start */
#define JSON_ERR_BADATTR	3	/* unknown attribute name */
#define JSON_ERR_ATTRLEN	4	/* attribute name too long */
#define JSON_ERR_NOARRAY	5	/* saw [ when not expecting array */
#define JSON_ERR_NOBRAK 	6	/* array element specified, but no [ */
#define JSON_ERR_STRLONG	7	/* string value too long */
#define JSON_ERR_TOKLONG	8	/* token value too long */
#define JSON_ERR_BADTRAIL	9	/* garbage while expecting comma or } or ] */
#define JSON_ERR_ARRAYSTART	10	/* didn't find expected array start */
#define JSON_ERR_OBJARR 	11	/* error while parsing object array */
#define JSON_ERR_SUBTOOLONG	12	/* too many array elements */
#define JSON_ERR_BADSUBTRAIL	13	/* garbage while expecting array comma */
#define JSON_ERR_SUBTYPE	14	/* unsupported array element type */
#define JSON_ERR_BADSTRING	15	/* error while string parsing */
#define JSON_ERR_CHECKFAIL	16	/* check attribute not matched */
#define JSON_ERR_NOPARSTR	17	/* can't support strings in parallel arrays */
#define JSON_ERR_BADENUM	18	/* invalid enumerated value */
#define JSON_ERR_QNONSTRING	19	/* saw quoted value when expecting nonstring */
#define JSON_ERR_NONQSTRING	19	/* didn't see quoted value when expecting string */
#define JSON_ERR_MISC		20	/* other data conversion error */
#define JSON_ERR_BADNUM		21	/* error while parsing a numerical argument */
#define JSON_ERR_NULLPTR	22	/* unexpected null value or attribute pointer */

/* fi_jsonpkg.h ends here */
