/*
 * fi_playback_protocol.h
 *
 *  Created on: Jan 25, 2016
 *      Author: eric
 */

#ifndef FI_PLAYBACK_PROTOCOL_H_
#define FI_PLAYBACK_PROTOCOL_H_

#include "fi_streaming_session.h"
#include "fi_streaming_protocol.h"
#include <iterator>
#include <pthread.h>

class BufferQueue;

/**
 * @brief StreamingProtocol tailored for camera's playback stream
 *
 * Clients can use this WebSocket protocol to receive streams of image
 * data from the camera, when the camera is reviewing captured images.
 * Image data is always compressed JPEG.
 */
class fiWSSPlaybackProtocol: public StreamingProtocol
{
public:
    fiWSSPlaybackProtocol(struct lws_protocols* p);
    virtual ~fiWSSPlaybackProtocol();

    void process_consumer(); // Override

    virtual void release_image_memory_buffer( void* shmem );

    void notify( uint32_t reg, uint32_t val );

    virtual int established(struct lws *wsi, WssSession** s);
    virtual int closed(struct lws *wsi, WssSession** user);
    virtual int destroy(struct lws *wsi, WssSession* user);
    virtual int writable(struct lws *wsi, WssSession* s);
    virtual int receive(struct lws *wsi, WssSession* s, uint8_t* in, size_t len);

private:
    static const char* PROTOCOL_NAME;
    static const char* JSON_TAG;

    //! @name Track camera states related to playback operation
    //! @{

    //! @brief Control initialization of the playback state
    typedef enum
    {
        NEED_INIT,
        INIT_DONE,
        CHANGE_PENDING,
        OK
    } playback_state_status_t;

    //! @brief Control initialization and update of the playback state
    playback_state_status_t playback_state_status;

    //! @brief Serialize access to these state variables
    pthread_mutex_t playback_state_mutex;

    //! @brief Camera is in a "playable" state
    //!
    //! True when both:
    //!    GIGE_REG_CAMERA_STATE == GIGE_CAMERA_STATE_REVIEW
    //!    GIGE_REG_PART_STATE == GIGE_PART_STATE_COMPLETE
    //!
    bool playable;

    //! @brief Camera hardware is playing recorded video
    //!
    //! True when:
    //!     GIGE_PBC_ENABLE is set in GIGE_REG_PB_CONTROL
    //!
    bool playing;

    //! @brief GIGE_REG_CAMERA_STATE
    uint32_t gige_reg_camera_state;

    //! @brief GIGE_REG_PB_CONTROL
    uint32_t gige_reg_pb_control;

    //! @brief GIGE_REG_PART_STATE
    uint32_t gige_reg_part_state;
    //! @}
};

#endif /* FI_PLAYBACK_PROTOCOL_H_ */
