/*
 * fi_wss_streamer.h
 *
 *  Created on: Jan 25, 2016
 *      Author: eric
 */

#ifndef FI_STREAMING_PROTOCOL_H_
#define FI_STREAMING_PROTOCOL_H_

#include <vector>

#include "fi_wss_session.h"
#include "fi_wss_protocol.h"
#include "fi_buffer_q.h"

class StreamingSession;

/**
 * @brief Base class for streaming Websockets protocols
 *
 * This base class abstracts most of the interaction with Buffers and the
 * BufferQueue required to move image data through Websockets sessions.
 * Specific protocols must sub-class this to implement specific protocol-
 * dependent functionality. (For example, adding headers to image frames.)
 *
 */
class StreamingProtocol: public WssProtocol
{
public:
    static const uint32_t MAGIC_ID = 0x46417374;
    
    typedef enum
    {
        START_OF_FRAME = 0x00000001,
        END_OF_FRAME = 0x00000002
    } StreamFlags;

    /**
     * @brief two-word (8-byte) header placed at the beginning of each Frame
     *
     * Always transmitted in network byte order.
     */
#pragma pack(1)
    typedef struct
    {
	uint32_t magic;	/**< Magic ID, aka "FAst". */
        uint32_t size;  /**< Size, in bytes, of the Image Frame data */
        int32_t id;     /**< Frame number, relative to Trigger (when applicable) */
    } FRAME_HEADER;
#pragma pack(0)

    /**
     * @brief Structure containing a FRAME_HEADER, along with pre- and post- padding for lilbwebsockets
     */
    typedef struct
    {
        uint8_t prepad[ LWS_SEND_BUFFER_PRE_PADDING ];    /**< Pre-buffer padding for libwebsockets */
        FRAME_HEADER h;                                   /**< Streaming protocol frame header */
        uint8_t postpad [ LWS_SEND_BUFFER_POST_PADDING ]; /**< Post-buffer padding for libwebsockets */
    } FRAME_HEADER_LWS;

    StreamingProtocol(struct lws_protocols* p);
    virtual ~StreamingProtocol();

    BufferQueue& buffer_q( void );

    void set_fault( void );

    virtual bool consumer_queue_has_data( void ); /**< True iff the Consumer Queue for this protocol has data */
    virtual bool producer_queue_has_buffer( void ); /**< True iff the Producer Queue for this protocol has an empty buffer */
    virtual bool tx_queue_is_empty( void ); /**< True iff the protocols's Tx queue is empty */
    bool get_bq_fault( void ); /**< Test fault status on protocol's BufferQueue */
    void flush_consumer_queue( void );

    void producer_wait( void );
    void producer_set_flags(uint32_t f);
    void producer_set_size(size_t s);
    void producer_set_frame_id( int32_t n );
    void producer_set_frame_data_size(uint32_t fds);
    void producer_set_signal( void (*func)( void* ), void* arg );
    char* producer_get_data_ptr(size_t offset);
    void producer_reserve_prefix( size_t size );
    int producer_index( void );

    virtual void process_consumer( void );

    virtual void release_image_memory_buffer( void* shmem ) = 0;

    virtual int established(struct lws *wsi, WssSession** s) = 0; /**< New session established */
    virtual int destroy(struct lws *wsi, WssSession* s) = 0;
    virtual int writable(struct lws *wsi, WssSession* s) = 0;
    virtual int receive(struct lws *wsi, WssSession* s, uint8_t* in, size_t len) = 0;

    bool streaming_protocol_trace; //!< @brief True to enable tracing

protected:
    BufferQueue bq;
    int listeners;  //*<  Number of active (ON_REQ, ON_FRAME, ON_IDLE, OFF_REQ) listeners */
};

#endif /* FI_STREAMING_PROTOCOL_H_ */
