/*
 * fi_wss_stream.h
 *
 *  Created on: Jan 25, 2016
 *      Author: eric
 */

#ifndef FI_WSS_PROTOCOL_H_
#define FI_WSS_PROTOCOL_H_
#include <vector>

#include "libwebsockets.h"
extern void lws_signal();

#include "fi_wss_session.h"

/**
 * @brief Base class for Fastec libwebsockets protocol modules
 *
 * This base class captures the essential interface between WSS protocol implementations
 * and the libwebsockets server framework. Specifically, this class implements the
 * callback linkage that the server uses to invoke protocol-specific processing as it
 * monitors activity on its network interfaces. The base class implements the callback
 * function, and in turn makes calls to individual (virtual) protocol methods based on
 * the callback reason. Protocols that sub-class WssProtocol must provide implementations
 * for the four pure virtual methods defined here:
 *
 *    - established (LWS_CALLBACK_ESTABLISHED)
 *    - destroy (LWS_CALLBACK_PROTOCOL_DESTROY)
 *    - writable (LWS_CALLBACK_SERVER_WRITEABLE)
 *    - receive (LWS_CALLBACK_RECEIVE)
 *
 * For each WssProtocol sub-class that implements a Websockets protocol, there should be
 * a corresponding sub-class of WssProtocolSession that represents individual sessions
 * (client connections) that use the protocol.
 *
 * In all cases, the libwebsockets per-session user data will point to an instance of
 * WssProtocolSession for with the protocol.
 */
class WssProtocol
{
public:
    /**
     * @brief WssProtocol base class constructor
     *
     * The constructor takes a pointer to a libwebsockets struct lws_protocols.
     * The constructor will fill in these fields:
     *
     * 	.name                    Default string "unknown." Derived classes should
     * 	                         over-write this with their own names.
     * 	.rx_buffer_size          Set to 0; Derived classes should set this based
     * 	                         on their own receive data buffering requirements.
     *
     * 	Derived classes should not over-write any of these fields:
     *
     *  .callback                Pointer to the static callback handler.
     *  .per_session_data_size   Set to sizeof(WssSession*).
     *  .user                    Set to NULL; not used by WssProtocol.
     *  .id                      Set to 0; not used by WssProtocol.
     *
     * @param p  (Output) pointer to struct lws_protocols
     */
    WssProtocol(struct lws_protocols* p);

    virtual ~WssProtocol();

    void set_lws_context( struct lws_context* c );

    /**
     * @name Callback functions
     *
     * These all execute in the context of the single WSS Thread. The
     * base class' static callback handler parses the libwebsockets
     * "callback reason," and calls the appropriate derived class
     * method, along with a pointer to the derived class' session.
     */
    /**@{*/
    virtual int accept( struct lws *wsi, WssSession** s );
    virtual int established(struct lws *wsi, WssSession** s) = 0;   /**< @brief New session established */
    virtual int closed(struct lws *wsi, WssSession** s) = 0;		/**< @brief Session closed */
    virtual int destroy(struct lws *wsi, WssSession* s) = 0;		/**< @brief Session terminated */
    virtual int writable(struct lws *wsi, WssSession* s) = 0;		/**< @brief Session is writable */
    virtual int receive(struct lws *wsi, WssSession* s, uint8_t* in, size_t len) = 0;		/**< @brief Data has arrived for a sesison */
    /**@}*/

    /**
     * @name Iterator support for protocols
     *
     * Since protocols may handle multiple sessions, the iterator here
     * supports walking through all the currently connected sessions.
     */
    /**@{*/
    typedef std::vector<WssSession*>::iterator iterator;
    typedef std::vector<WssSession*>::const_iterator const_iterator;
    iterator begin();
    iterator end();
    /**@}*/

protected:
    std::vector<WssSession*> sessions;	/**< Currently active sessions for this protocol */
    struct lws_context* ctx;
};

#endif /* FI_WSS_PROTOCOL_H_ */
