/*!
 * @file fi_wss_protocols.h
 * @brief header file for fi_wss_protocols
 *
 * central point for handling all supported libwebsocket protocol
 * handlers available in the fastec imaging web socket server
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_WSS_PROTOCOLS_H
#define FI_WSS_PROTOCOLS_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "fi_buffer_q.h"

//--------------------------------------------------------------------
// Forward Class Declarations
//--------------------------------------------------------------------
class fiHTTPProtocol;
class fiWSSFCPProtocol;
class fiWSSFCPListenProtocol;
class fiWSSLiveViewProtocol;
class fiWSSPlaybackProtocol;
class fiWSSDownloadProtocol;

//--------------------------------------------------------------------
// Class: fiWSSProtocols
//--------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Constant, Macro and Type Definitions
//-----------------------------------------------------------------------------

// The PROTOCOL_TYPES define the order of the protocols in the protocol
// static array managed by this object
enum PROTOCOL_TYPES
{
    // always first in the list
    PROTOCOL_HTTP = 0,

    // first customized protocol handler after the http
    PROTOCOL_FCP,				//!< The Fastec Imaging control register protocol.
    PROTOCOL_FCP_LISTENER,			//!< The Fastec Imaging asynchronous notification protocol.
    PROTOCOL_FI_LIVEVIEWER,			//!< The Fastec Imaging asynchronous live image view protocol.
    PROTOCOL_FI_PLAYBACK,
    PROTOCOL_FI_DOWNLOAD,

    // last entry just dummy placeholder for count
    PROTOCOL_COUNT
};

// The FI_WSSPROTO_STATES define the state of the protocols used for wss
// as a collective entity - individual protocol state is defined in the
// protocol object itself
enum FI_WSSPROTOCOLS_STATE
{
    // always first in the list
    FI_WSSPROTOCOLS_UNINITIALIZED = 0,	// empty uninitialized state
    FI_WSSPROTOCOLS_CREATED,			// the wss prototypes are created and ready
    FI_WSSPROTOCOLS_CREATION_FAILED,	// the cs initialziation failed
    FI_WSSPROTO_STATE_COUNT
};

class fiWSSProtocols
{
public:
    // Web Socket Server Protocol constructor
    fiWSSProtocols();

    // Web Socket Server Protocol destructor
    ~fiWSSProtocols();

    // Web Socket Server Protocols Creation
    // returns a 0 if fails or a 1 if all are protocols are ready
    FI_WSSPROTOCOLS_STATE createProtocols(fiCameraServerIO* pfiCS);

    // Get Web Socket Server Protocols List
    void setListOfProtocols(struct lws_context_creation_info *pfiWSSCreateInfo);

    // Time Slice Is Available to Write to any connected wss
    // Will walk through protocols giving them a slice of time
    void canWriteIfDataReady(struct lws_context *ctx);

    fiWSSLiveViewProtocol* getLiveViewProto();
    fiWSSPlaybackProtocol* getPlaybackProto();
    fiWSSDownloadProtocol* getDownloadProto();

private:

    //holds the state of readiness for the protocols collectively
    FI_WSSPROTOCOLS_STATE ProtocolsAreReady;

    // List of the libwebsocket protocol records supported in this
    // protocols object - this structure contains all of the callback
    // information and important buffersizes that define protocol
    // communication package sizes and is necessary to make calls into
    // the libwebsocket interface
    static struct lws_protocols protocols[PROTOCOL_COUNT + 1];

    fiHTTPProtocol *pHTTPProto;

    // the FCP protocol object
    fiWSSFCPProtocol *pFCPProto;

    // the FCP listener protocol object
    fiWSSFCPListenProtocol* pFCPListenerProto;

    // the LiveView listener protocol object
    fiWSSLiveViewProtocol* pLVListenerProto;

    // Playback protocol
    fiWSSPlaybackProtocol* pPlaybackProto;

    // Download protocol
    fiWSSDownloadProtocol* pDownloadProto;
};

#endif // FI_WSS_PROTOCOLS_H
