/*
 * lssm_thread.h
 *
 *  Created on: Feb 2, 2016
 *      Author: eric
 */

#ifndef LSSM_THREAD_H_
#define LSSM_THREAD_H_

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>

#include "fi_liveview_protocol.h"
#include "fi_playback_protocol.h"
#include "fi_download_protocol.h"

#include "fastec/lssm.h"

#include "fi_buffer_q.h"

/**
 * @brief Handle interactions with the Camera Server's local stream
 *
 * This class is a client for the Camera Server's local stream interface.
 * Note that in the camera's WebSocket Server, this class plays the client
 * role on the local stream interface, and the "producer" role within the
 * WebSocket Server. The Camera Server implements the local stream server.
 * The WebSocket Server's main thread executes the "consumer" role within
 * the WebSocket Server. The Buffer and BufferQeue class implement the
 * data and control paths between the "producer" and "consumer" threads:
 *
 * {{{
 * Camera Server
 * -------------
 *              Local Stream
 *                  LSSM library (server)
 *                      ^^
 *                      || <-- shared memory interface
 *                      ||
 *                      ||
 * WebSocket Server     ||
 * ----------------     vv
 *                  LSSM library (client)
 *     LSSM_Thread ("Producer") <----> BufferQueue/Buffers <--> Main thread ("Consumer"):
 *                                                              libwebsockets
 *                                                              and protocols
 * }}}
 *
 * The local stream server supplies a stream of processed image frames
 * in the camera's image memory, and signals their availability across
 * the LSSM shared memory interface. The WebSocket Server catches these
 * signals, copies image data into its own system memory, and pushes
 * them into WebSocket connections (through the WebSocket Server BufferQueue/
 * Buffer system).
 *
 * The Camera Server can supply images in a number of formats:
 *     - BMP (Microsoft Windows Bitmap)
 *     - DNG (Digital Negative)
 *     - JPEG (Joint Photographic Experts Group compressed)
 *     - TIF (Tagged Image File format)
 *
 * The details of the image format in a stream are largely immaterial to
 * LSSM_Thread.
 * The Camera Server supplies the necessary parameters, along with information about each
 * image's availabliilty in image memory, on the LSSM shared-memory interface.
 *
 * Signaling and Synchronization
 * -----------------------------
 * The LSSM library makes a synchronization/signaling API available to
 * the LSSM server and client. The LSSM_Thread class makes use of these
 * functions to synchronize itself with events from the Camera Server's
 * Local Stream. LSSM_Thread also leverages the LSSM library functions
 * to receive asynchronous events from the WebSocket Server "consumer"
 * thread (Main thread).
 *
 * The Camera Server can signal these events through the LSSM:
 *     - Image is available in image memory (hands ownership of
 *       an image memory buffer to the LSSM Client (in this case,
 *       the WebSocket Server LSSM_Thread).
 *     - Local Stream has terminated stream processing. This is an
 *       indication that the client (LSSM_Thread) must release
 *       image memory buffers that it is currently holding back to
 *       the LSSM server.
 *
 * The WebSocket Server "consumer" thread also uses the LSSM to signal
 * events to the LSSM_Thread. Specifically:
 *     - The Consumer has finished processing data in a Buffer (on the
 *       Buffer Queue), so an empty buffer is now available for the
 *       "Producer."
 *
 * The LSSM_Thread uses a "custom signaling" hook available in the
 * BufferQueue class to leverage the LSSM synchronization/signaling
 * mechanisms for this purpose.
 */
class LSSM_Thread
{
public:
    static void* lssm_entry(void* self)
    {
        static_cast<LSSM_Thread*>(self)->run((void*)NULL);
        return NULL;
    }

    LSSM_Thread(fiWSSLiveViewProtocol* lv,
                fiWSSPlaybackProtocol* pb,
                fiWSSDownloadProtocol* dl,
                struct lws_context* ctx);

    virtual ~LSSM_Thread();

    void consumer_signal();

protected:
    virtual void run(void* t);

    struct lws_context* lws_ctx;
    pthread_t thread;
    void* shmem;
    int mmap_fd;
    int current_mode;		

    fiWSSLiveViewProtocol* liveview_proto;
    fiWSSPlaybackProtocol* playback_proto;
    fiWSSDownloadProtocol* download_proto;

private:
    void process_xfer( StreamingProtocol* prot );
    void check_lssm( void );
    bool lssm_idle( void );

    size_t compute_xfer_size(size_t space_available, size_t data_available);
    void deal_with_xfer(BufferQueue& bq, size_t nbytes, uint8_t *dst);

    /**
     * @brief Propagate an LSSM fault indication to one of our streaming protocols
     *
     * The LSSM has indicated a fault condition. It should be associated with one of
     * our streaming protocols, and the LSSM should have specified what is was doing
     * (playback or download) at the time of the fault. Propagate the fault to the
     * appropriate protocol.
     *
     * TODO: This needs a parameter to identify the protocol (playback/download)
     */
    void producer_set_fault( StreamingProtocol* prot );

    void fault_recovery( BufferQueue& bq );
};

#endif /* LSSM_THREAD_H_ */
