/*!
 * @file fcp.cxx
 * @brief implements methods for communicating with the camera server.
 *
 * Copyright 2011,2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <pthread.h>

#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <signal.h>
#include <sys/signalfd.h>

#include <sys/ioctl.h>

#include <fastec/fas_log.h>
#include <fastec/gige_msg.h>
#include "fcp.h"

//--------------------------------------------------------------------
// Constants, Defines, Typedefs and Macros
//--------------------------------------------------------------------
static uint8_t cs_verbose = 1;      // control when log messages are written
static bool local_server = false;   // must be set in cs_init
static uint32_t dest_ip = 0;
static unsigned short dest_port = 0;

#define CS_KEY_PATH             "/var/run/CameraServer.socket"
#define CS_QUEUE_FILE           "/tmp/wss-fcp.queue"

// Timeout period for our communications thread.
#define CS_RCV_TIMEOUT_SEC      1
#define CS_RCV_TIMEOUT_USEC     0

// Define the number of times we'll retry when waiting for a message.
#define CS_MAX_RETRIES          5

// If we are just getting too too many errors, we will want to give up and stop logging them
// so we don't fill up the logs.
#define MAX_COMM_FAIL_ERRORS        50

// Number of nanoseconds per second.
#define NANOSEC_PER_SEC         1000000000

// Define the timeout period for our select call.
#define CS_PSELECT_TIMEOUT_SEC      5
#define CS_PSELECT_TIMEOUT_NSEC     000000000

// Define our communications thread states.
#define CS_THREAD_IDLE              0
#define CS_THREAD_SEND_PENDING      1
#define CS_THREAD_WRITE_FD_PENDING  2
#define CS_THREAD_REPLY_PENDING     3
#define CS_THREAD_MSG_FAIL          4
#define CS_THREAD_REPLY_OK          5
#define CS_THREAD_SEND_RETRY        6

typedef struct
{
    CS_REG_CALLBACK func;
    void *arg;
} CB_ENTRY;

static CB_ENTRY callbacks[8] = { {0, NULL}, };

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

static int  cs_fd = -1;         //!< Local copy of unix domain socket.
static char cs_sock_path[ PATH_MAX ];   //!< Null terminated path of our socket.
static uint16_t cs_last_msg_id = 0;     //!< Last message ID sent to the Camera Server.
static uint32_t cs_last_reg_id = 0;

static volatile int cs_exit_requested = 0;  //!< Flag to indicate when we'd like the thread to exit.
static volatile uint32_t cs_thread_state = CS_THREAD_IDLE;//!< Used to manage the state of the communications thread.
static pthread_mutex_t cs_com_mutex;        //!< Used to serialize communication with camera server.
static int regrw_fail = 0;                  //!< Used to reduce error spewage.

static char temp_buffer[ FCP_MSG_DATA_SIZE ];   //!< Temp buffer for use below.

static pthread_cond_t cs_write_ready;           //!< Used to control access to the communications thread.
static pthread_cond_t cs_reply_ready;           //!< Used to wait for a response from the communications thread.
static fcp_msg  cs_msg_out;             //!< Structure used to send a GiGE message.
static fcp_msg  cs_msg_in;              //!< Structure used to receive a GiGE message.
static pthread_t cs_thread;             //!< Our communications thread.

//--------------------------------------------------------------------
// Static Function Definitions
//--------------------------------------------------------------------

/*!
 * @brief Register for a callback for register change notifications.
 *
 * Note that this method only supports a single callback.  If the application
 * wishes to have multiple callbacks for some reason, a method must be designed
 * to save the old callback function and have the new callback function call
 * the old.
 *
 * DO NOT make calls back into the cs_client module from within this callback.
 * A deadlock in such circumstances is possible if not likely.
 *
 * @param new_callback: pointer to the callback function.
 *
 * @return Old calback function or NULL if there was no function registered.
 */
void cs_register_callback(CS_REG_CALLBACK new_callback, void *arg)
{
    // IMPORTANT NOTE:
    // local_server should be set to true when the Camera Server is running
    // on the same process space as the caller of this function.  If the Camera Server
    // is remote (that is what happens process is running on an Ubuntu system
    // that is not hosting a Camera Server) - and in that case the Notification
    // handling will NOT work.  So the best course of action is simply to return
    // early in this function.  In production the process calling will be local
    // to the same processor running the Camera Server.
    if (!local_server)
        return; // circumvent for testing in debugger on Ubuntu

    // otherwise this is the code that sets up the notification thread
    int status;

    CB_ENTRY *cbe = callbacks;
    while (cbe != &callbacks[sizeof(callbacks) / sizeof(callbacks[0])])
    {
        if (cbe->func == NULL)
        {
            cbe->arg = arg;
            cbe->func = new_callback;
            break;
        }
        cbe++;
    }
    if (cbe == &callbacks[sizeof(callbacks) / sizeof(callbacks[0])])
        fi_log(LOG_ERR, "(%s) Too many callback registrations", __FUNCTION__);

    // Register with the camera server to get asynchronous notifications. Do this
    // at the end, because once registered we have to assume we could receive a
    // notification at any time.
    //
    // Note that cs_set_reg_l() proxies the call to the worker thread, so by
    // definition, the message is not sent until the worker thread is ready to listen.
    while (!cs_set_reg_l(GIGE_REG_EVENT_NOTIFY, 1, &status))
    {
        fi_log(LOG_WARNING, "(%s) unable to set GIGE_REG_EVENT_NOTIFY.  Retrying...", __FUNCTION__);

        // Sleep for 1 second.
        sleep(1);
    }

    return;
}

/*!
 * @brief Perfroms result = x - y, and returns -1 if x < y, 0 if x = y, and 1 if x > y.
 */
int timespec_subtract(struct timespec *result, struct timespec *x, struct timespec *y)
{
    uint64_t    op1, op2, ans;
    int     rc;

    // Sanity check.
    if ((x == NULL) || (y == NULL))
    {
        fi_log(LOG_ERR, "(%s) Invalid operands", __FUNCTION__);
        return -1;
    }

    // Convert to nanosecond values.
    op1 = (x->tv_sec * NANOSEC_PER_SEC) + x->tv_nsec;
    op2 = (y->tv_sec * NANOSEC_PER_SEC) + y->tv_nsec;

    // Find our answer.
    ans = op1 - op2;

    // Convert back to our structure.
    if (result)
    {
        result->tv_sec = ans / NANOSEC_PER_SEC;
        result->tv_nsec = (ans - (result->tv_sec * NANOSEC_PER_SEC));
    }

    // Return 1 if result is negative.
    if (op1 < op2)
        rc = -1;
    else if (op1 == op2)
        rc = 0;
    else
        rc = 1;

    // Return our comparison value.
    return rc;
}

/*!
 * @brief Wait for and read a message from the camera server
 *
 * Wait for at most CS_RCV_TIMEOUT_USEC for a message from the camera server.
 *
 *
 * @param p_msg: pointer to buffer in which to put the message
 * @param num_bytes: number of bytes in the buffer
 *
 * @return 1 on success, 0 on error.
 */
static uint32_t cs_ipc_get_msg(unsigned char *p_msg, int num_bytes)
{
    struct sockaddr from;       // socket address structure
    int     fromlen = sizeof(from); // Number of bytes in our structure.
    ssize_t msg_size = 0;   // How big was that message?
    int retries = CS_MAX_RETRIES;   // Number of times we'll try to receive a message.

    // Make sure we have a valid descriptor.
    if (cs_fd < 0)
    {
        fi_log(LOG_ERR, "(%s) invalid socket fd", __FUNCTION__);
        return 0;
    }

    // Initialize our receive structure.
    memset(&from, 0, sizeof(from));
//    from.sun_family = AF_LOCAL;

    // Try to get our message from the CS.
    do
    {
        msg_size = recvfrom(cs_fd, (char *)p_msg, num_bytes, 0, &from, (socklen_t *)&fromlen);
        retries--;
    } while ((msg_size <= 0) && (errno == EAGAIN) && (retries > 0));

    // Make sure we got a valid message.
    if ((msg_size == -1) || (msg_size == 0))
    {
        fi_log(LOG_ERR, "(%s) error receiving after %d retries, errno:%d(%s)",
               __FUNCTION__, CS_MAX_RETRIES - retries, errno, strerror(errno));
        return 0;
    }

    // All done.
    return 1;
}

/*!
 * @brief Send a message
 *
 * Sends a message message to the camera server and waits for then validates
 * the id of the reply.  It is the reponsibility of the caller to validate
 * the proper command in the response.
 *
 * Note: this method also tracks and fills in the request ID in a thread safe way.
 *
 * @param p_msg: pointer to the message to send
 * @param num_bytes: number of bytes in the message
 *
 * @return bytes sent on success, 0 on error.
 */
static uint32_t cs_ipc_send_msg(unsigned char *p_msg, int num_bytes)
{
    ssize_t msg_size = 0;           // Number of bytes actually sent.

    // Make sure we have a valid descriptor.
    if (cs_fd < 0)
    {
        fi_log(LOG_ERR, "(%s) invalid socket fd", __FUNCTION__);
        return 0;
    }

    // NOTE:
    // if connecting to the Camera Server Locally will use UNIX domain
    // sockets - if not will use INET sockets for communication

    if (local_server)
    {
        struct sockaddr_un to;  // UNIX domain socket address structure.
        int tolen = sizeof(to); // Number of bytes in our structure.

        // Populate the socket.
        memset(&to, 0, sizeof(to));
        to.sun_family = AF_LOCAL;
        strncpy(to.sun_path, CS_KEY_PATH, sizeof(to.sun_path));

        // Send the message.
        msg_size = sendto(cs_fd, (char *)p_msg, num_bytes, 0,
                          (struct sockaddr *)&to, tolen);
    }
    else
    {
        struct sockaddr_in to;  // INET socket address structure.
        int tolen = sizeof(to); // Number of bytes in our structure.

        memset(&to, 0, sizeof(to));
        to.sin_family = AF_INET;
        to.sin_addr.s_addr = htonl(dest_ip);
        to.sin_port = htons(dest_port);

        // Send the message.
        msg_size = sendto(cs_fd, (char *)p_msg, num_bytes, 0,
                          (struct sockaddr *)&to, tolen);
    }

    if (msg_size != num_bytes)
    {
        fi_log(LOG_ERR, "(%s) error sending, errno:%d(%s)",
               __FUNCTION__, errno, strerror(errno));
        return 0;
    }

    return (msg_size == num_bytes ? 1 : 0);
}

/*!
 * @brief Event thread
 *
 * This is an event thread that handles 4 events:
 *    A signal indicating that there is a pending CS send request.
 *    A signal indicating that the thread should exit.
 *    The IPC has read data.
 *    The IPC can accept write data (only done if there is a pending CS write reqest).
 *
 * The signals are handled synchronously through signalfd() so the select() call will
 * wake up. Once it wakes, it uses static variables to decide which situation applies.
 *
 * It is important to note that while this thread access variables that would
 * typically be guarded by the cs_com_mutex, at the time it is using/modifying
 * these variables, all other threads would be blocked waiting for the
 * request to complete or this thread to go idle.
 *
 * @param arg: pointer to a thread argument.  Currently unused
 *
 * @return NULL
 */
static void *event_thread(void * /*arg*/)
{
    int     comm_fail = 0;
    int     result;
    uint16_t    rcv_msg_id, rcv_msg_cmd, rcv_event_reg;
    uint32_t    retries = 0;
    uint32_t    rcv_event_val;
    int     sig_fd;
    int     nfds;
    sigset_t    usr1_set;
    fd_set  read_fds;
    fd_set  write_fds;
    fcp_msg tmp_msg_in;
    struct timespec     to;
    struct timespec sleep_req;

    // Create a signal file descriptor that we can "select()" on.
    sigemptyset(&usr1_set);
    sigaddset(&usr1_set, SIGUSR1);
    sig_fd = signalfd(-1, &usr1_set, 0);
    if (sig_fd == -1)
    {
        fi_log(LOG_ERR, "(%s) signalfd failed %d(%s)", __FUNCTION__,
               errno, strerror(errno));
        goto init_failure;
    }
    nfds = (sig_fd > cs_fd) ? (sig_fd + 1) : (cs_fd + 1);

    while (!cs_exit_requested)
    {
        // Initialize the read descriptor set - descriptors we are
        // interested about.
        FD_ZERO(&read_fds);
        FD_SET(cs_fd, &read_fds);
        FD_SET(sig_fd, &read_fds);

        // Initialize the write descriptor set.
        FD_ZERO(&write_fds);
        if ((cs_thread_state == CS_THREAD_SEND_PENDING) ||
            (cs_thread_state == CS_THREAD_SEND_RETRY))
        {
            // When we have a send request waiting, before writing, we must wait
            // for the fd to be writable.
            FD_SET(cs_fd, &write_fds);

            // Only reset the retries on the first send, not any retries
            if (cs_thread_state == CS_THREAD_SEND_PENDING)
                retries = CS_MAX_RETRIES;

            // Update our state.
            cs_thread_state = CS_THREAD_WRITE_FD_PENDING;
        }

        // Initialize our timeout period.
        to.tv_sec = CS_PSELECT_TIMEOUT_SEC;
        to.tv_nsec = CS_PSELECT_TIMEOUT_NSEC;
        result = pselect(nfds, &read_fds, &write_fds, NULL, &to, NULL);

        // Are we being asked to bail?
        if (cs_exit_requested)
        {
            fi_log(LOG_DEBUG, "%s: cs_exit_requested, so bailing...", __PRETTY_FUNCTION__);
            break;
        }

        // Did we get an interrupt/signal we didn't expect?
        if ((result < 0) && (errno == EINTR))
        {
            if (++comm_fail < MAX_COMM_FAIL_ERRORS)
                fi_log(LOG_ERR, "(%s) Unexpected EINTR in select", __FUNCTION__);
            continue;   // Don't want to process fd sets on signal
        }

        // Was there a problem?
        if (result < 0)
        {
            if (++comm_fail < MAX_COMM_FAIL_ERRORS)
                fi_log(LOG_ERR, "(%s) Error waiting with pselect() %d:%s",
                       __FUNCTION__, errno, strerror(errno));

            // Sleep for 500 milliseconds.
            sleep_req.tv_sec = 0;
            sleep_req.tv_nsec = 500000000;
            nanosleep(&sleep_req, NULL);
            continue;   // Don't want to process fd sets on error
        }

        // Did we timeout?
        if (result == 0)
        {
            // A timeout occurred. Check to see if we had anyone
            // waiting (PENDING) on us.
            if (cs_thread_state == CS_THREAD_WRITE_FD_PENDING)
            {
                if (++comm_fail < MAX_COMM_FAIL_ERRORS)
                    fi_log(LOG_ERR, "(%s) Timeout waiting for write ready on IPC socket", __FUNCTION__);

                // Must hold mutex to signal condition
                pthread_mutex_lock(&cs_com_mutex);
                {
                    cs_thread_state = CS_THREAD_MSG_FAIL;
                    pthread_cond_signal(&cs_reply_ready);
                }
                pthread_mutex_unlock(&cs_com_mutex);
            }
            else if (cs_thread_state == CS_THREAD_REPLY_PENDING)
            {
                if (++comm_fail < MAX_COMM_FAIL_ERRORS)
                    fi_log(LOG_WARNING, "(%s) Timeout waiting for Camera Server reply reg: 0x%x id:%d, try #%d", __FUNCTION__, cs_last_reg_id, cs_last_msg_id, retries);
                retries--;
                if (retries != 0)
                {
                    // Decrement retries and re-transmit same request.
                    cs_thread_state = CS_THREAD_SEND_RETRY;
                }
                else
                {
                    if (++comm_fail < MAX_COMM_FAIL_ERRORS)
                        fi_log(LOG_ERR, "(%s) retries exhausted", __FUNCTION__);

                    // Must hold mutex to signal condition
                    pthread_mutex_lock(&cs_com_mutex);
                    {
                        cs_thread_state = CS_THREAD_MSG_FAIL;
                        pthread_cond_signal(&cs_reply_ready);
                    }
                    pthread_mutex_unlock(&cs_com_mutex);
                }
            }

            continue;   // Don't want to process fd sets on timeout either
        }

        // At this point, we are fairly certain that we actually had activity
        // on one of the descriptors we were interested in.

        // Check to see if one of the read descriptors had activity.
        if (FD_ISSET(cs_fd, &read_fds))
        {
            // Try to read the message from the CS.
            memset(&tmp_msg_in, 0, sizeof(fcp_msg));
            result = cs_ipc_get_msg((unsigned char *)&tmp_msg_in, sizeof(fcp_msg));
            if (result == 0)
            {
                // In this case, the fd is supposed to be read-ready, but the read
                // failed.  Abort and flag an error
                if (++comm_fail < MAX_COMM_FAIL_ERRORS)
                    fi_log(LOG_ERR, "(%s) failure receving message", __FUNCTION__);
                if (cs_thread_state == CS_THREAD_REPLY_PENDING)
                {
                    // Must hold mutex to signal condition
                    pthread_mutex_lock(&cs_com_mutex);
                    {
                        cs_thread_state = CS_THREAD_MSG_FAIL;
                        pthread_cond_signal(&cs_reply_ready);
                    }
                    pthread_mutex_unlock(&cs_com_mutex);
                }

                // Sleep for 500 milliseconds.
                sleep_req.tv_sec = 0;
                sleep_req.tv_nsec = 500000000;
                nanosleep(&sleep_req, NULL);

                continue;   // Don't want to process fd sets on timeout either
            }

            // We had a successful communication. All is forgiven.
            comm_fail = 0;

            // Get the information byte swapped so we can process it.
            rcv_msg_id = ntohs(tmp_msg_in.header.resp.ack_id);
            rcv_msg_cmd = ntohs(tmp_msg_in.header.resp.command);

            // Was the message an event?
            if (rcv_msg_cmd == GIGE_MSG_EVENTDATA_CMD)
            {
                int n_found = 0;

                // Get the data into a format we can use.
                rcv_event_reg = ntohs(tmp_msg_in.fed.event_id);
                rcv_event_val = ntohl(tmp_msg_in.fed.reg_val_l);

                // pass the notification through to interested parties.
                CB_ENTRY *cbe = callbacks;
                while (cbe != &callbacks[sizeof(callbacks) / sizeof(callbacks[0])])
                {
                    if (cbe->func != NULL)
                    {
                        (*(cbe->func))(cbe->arg, rcv_event_reg, rcv_event_val);
                        ++n_found;
                    }
                    cbe++;
                }

                // If we didn't have any callbacks register, make sure we disable
                // the notification event.
                if (n_found == 0)
                {
                    int status;

                    if (!cs_set_reg_l(GIGE_REG_EVENT_NOTIFY, 0, &status))
                    {
                        fi_log(LOG_ERR, "(%s) unable to set GIGE_REG_EVENT_NOTIFY", __FUNCTION__);
                    }
                }
            }

            // Were we waiting for a reply?
            else if (cs_thread_state == CS_THREAD_REPLY_PENDING)
            {
                // Sanity check the reply - was this the one we were waiting for?
                if (cs_last_msg_id != rcv_msg_id)
                {
                    if (++comm_fail < MAX_COMM_FAIL_ERRORS)
                        // Stay in reply pending state and wait for next message, or a timeout.
                        fi_log(LOG_ERR, "(%s) Ignoring received msg: msg_id(%d) does not match send msg_id(%d)",
                               __FUNCTION__, rcv_msg_id, cs_last_msg_id);
                }
                else
                {
                    // Looks like we got our message, copy it over so we can
                    // pass it on to the waiting thread.
                    memcpy(&cs_msg_in, &tmp_msg_in, sizeof(fcp_msg));

                    // Must hold mutex to signal condition
                    pthread_mutex_lock(&cs_com_mutex);
                    {
                        cs_thread_state = CS_THREAD_REPLY_OK;
                        pthread_cond_signal(&cs_reply_ready);
                    }
                    pthread_mutex_unlock(&cs_com_mutex);
                }
            }
            else
            {
                if (++comm_fail < MAX_COMM_FAIL_ERRORS)
                    fi_log(LOG_WARNING, "(%s) unknown async message id:%d cmd:%x received", __FUNCTION__, rcv_msg_id, rcv_msg_cmd);
            }
        }

        // Check to see if it was one of the write descriptors that woke us up.
        if ((cs_thread_state == CS_THREAD_WRITE_FD_PENDING) && FD_ISSET(cs_fd, &write_fds))
        {
            // Try to send our message.
            result = cs_ipc_send_msg((unsigned char *)&cs_msg_out, sizeof(fcp_msg));
            if (result == 0)
            {
                // Back off the last message ID, since it didn't work.
                cs_last_msg_id--;
                fi_log(LOG_ERR, "(%s) error sending message: %d(%s)", __FUNCTION__, errno, strerror(errno));

                // Indicate that we're a failure and notify any pending threads.
                if (++comm_fail < MAX_COMM_FAIL_ERRORS)
                    fi_log(LOG_ERR, "(%s) error sending, errno:%d(%s)",
                           __FUNCTION__, errno, strerror(errno));

                // Must hold mutex to signal condition
                pthread_mutex_lock(&cs_com_mutex);
                {
                    cs_thread_state = CS_THREAD_MSG_FAIL;
                    pthread_cond_signal(&cs_reply_ready);
                }
                pthread_mutex_unlock(&cs_com_mutex);

                // Sleep for 500 milliseconds.
                sleep_req.tv_sec = 0;
                sleep_req.tv_nsec = 500000000;
                nanosleep(&sleep_req, NULL);
                continue;
            }

            // All is well - we should be getting a reply anytime now.
            cs_thread_state = CS_THREAD_REPLY_PENDING;
        }

        // If it was our signal fd, read it to flush it. We just
        // needed it to wake us up but we don't care about the info.
        if (FD_ISSET(sig_fd, &read_fds))
        {
            struct signalfd_siginfo buf;
            ssize_t unused __attribute__((unused));

            // Compiler trivia: In C we can avoid the warning about ignoring the return value
            // with a (void) cast but in C++ that doesn't do the job so we need this construction.
            unused = read(sig_fd, &buf, sizeof(buf));
        }
    }
    close(sig_fd);

init_failure:
    // Exit this thread.
    pthread_exit(NULL);
}

/*!
 * @brief Wait for the cs event_thread to be idle (write ready)
 *
 * The caller should hold the cs_com_mutex.
 *
 * This method was added to provide a specific function to time for profiling.
 *
 */
__attribute__((noinline))
static void cs_wait_thread_idle(void)
{
    while ((cs_thread_state != CS_THREAD_IDLE) && !cs_exit_requested)
    {
        // It is not idle, so we'll wait for it here. When this returns
        // we will still own the mutex.
        pthread_cond_wait(&cs_write_ready, &cs_com_mutex);
    }
}


/*!
 * @brief Wait for the cs event_thread to get a reply
 *
 * The caller should hold the cs_com_mutex.
 *
 * This method was added to provide a specific function to time for profiling.
 *
 */
__attribute__((noinline))
static void cs_wait_thread_reply(void)
{
    while (!cs_exit_requested && (cs_thread_state != CS_THREAD_REPLY_OK) &&
           (cs_thread_state != CS_THREAD_MSG_FAIL))
    {
        pthread_cond_wait(&cs_reply_ready, &cs_com_mutex);
    }
}


/*!
 * @brief Send a message and wait for ack.
 *
 * Sends a message to the camera server and waits for then validates the id of
 * the reply.  It is the reponsibility of the caller to validate the expected
 * command in the response.
 *
 * Note: this method also tracks and fills in the request ID in a thread safe way.
 *
 * @param out: pointer to the message to send
 * @param in: pointer to the structure in which to store the reply.
 *
 * @return 1 on success, 0 on error.
 */
static uint32_t cs_send_msg_get_ack(fcp_msg *out, fcp_msg *in)
{
    int istat;          // Our return status.

    // Sanity check our parameters.
    if ((out == NULL) || (in == NULL))
    {
        fi_log(LOG_ERR, "(%s) NULL pointer in parameters", __FUNCTION__);
        return 0;
    }

    // Lock access.
    pthread_mutex_lock(&cs_com_mutex);

    // Make sure the thread is idle.
    cs_wait_thread_idle();

    if (cs_exit_requested)
    {
        fi_log(LOG_ERR, "(%s) CS work thread exiting", __FUNCTION__);
        pthread_mutex_unlock(&cs_com_mutex);
        return 0;
    }

    // Increment our message ID, and set it in our message structure.
    cs_last_msg_id++;
    out->header.req.req_id = htons(cs_last_msg_id);

    // Copy the new message over.
    memcpy(&cs_msg_out, out, sizeof(fcp_msg));

    // Change the state
    cs_thread_state = CS_THREAD_SEND_PENDING;

    // Does not really kill the thread, just sends the signal to the thread.
    pthread_kill(cs_thread, SIGUSR1);

    // Wait for our reply.
    cs_wait_thread_reply();

    if (cs_thread_state != CS_THREAD_REPLY_OK)
    {
        // Details on the error should have been logged already
        istat = 0;
    }
    else
    {
        memcpy(in, &cs_msg_in, sizeof(fcp_msg));
        istat = 1;
    }

    // We are done - move the thread to the idle state and let
    // anyone pending know that the thread is available now.
    cs_thread_state = CS_THREAD_IDLE;
    pthread_cond_broadcast(&cs_write_ready);

    // Release our mutex and return our result.
    pthread_mutex_unlock(&cs_com_mutex);

    return istat;
}


/*!
 * @brief Get a register value from CameraServer
 *
 * If a value is in the cache and has not timed out, it is stored in the pointer
 *
 * @param reg: the register to lookup in the cache
 * @param val: pointer to a location to store the value
 *
 * @return 1 on success, 0 on error.
 */
static uint32_t cs_get_reg_from_cs(uint32_t reg, uint32_t *val)
{
    fcp_msg  msg_out;  // structure used to send a gige message
    fcp_msg  msg_in;   // structure used to receive a gige message
    int istat = 0;      // status variable
    unsigned char rcv_cmd;

    memset(&msg_out, 0, sizeof(msg_out));
    msg_out.header.req.code = GIGE_REQ_CODE;
    msg_out.header.req.flag = GIGE_FLAG_MUST_ACK_BE;
    msg_out.header.req.command = htons(GIGE_MSG_READREG_CMD);
    msg_out.header.req.length = htons(sizeof(uint32_t));  // 1 registers
    msg_out.data_l[0] = htonl(reg);

    istat = cs_send_msg_get_ack(&msg_out, &msg_in);
    if (!istat)
    {
        if (++regrw_fail < MAX_COMM_FAIL_ERRORS)
            fi_log(LOG_ERR, "(%s) error reading register %04x", __FUNCTION__, reg);
        return 0;
    }

    // Make sure we got back our ACK.
    rcv_cmd = ntohs(msg_in.header.resp.command);
    if (rcv_cmd != GIGE_MSG_READREG_ACK)
    {
        fi_log(LOG_ERR, "(%s) received invalid command type:%d expected GIGE_MSG_READREG_ACK:%d",
               __FUNCTION__, rcv_cmd, GIGE_MSG_READREG_ACK);
        return 0;

    }

    istat = ntohs(msg_in.header.resp.status);
    if (istat != GEV_STATUS_SUCCESS)
    {
        fi_log(LOG_INFO, "(%s) received error 0x%04x in response to GIGE_MSG_READREG_CMD", __FUNCTION__, istat);
        fi_log(LOG_ERR, "(%s) error reading register %04x", __FUNCTION__, reg);
        return 0;
    }

    // Get the value into a useable form.
    *val = ntohl(msg_in.data_l[0]);

    regrw_fail = 0;
    return 1;
}

//--------------------------------------------------------------------
// Global Function Definitions
//--------------------------------------------------------------------

/*!
 * @brief initialize the camera server client interface
 *
 * Initializes the ipc socket and other necessary resources dedicated to communication
 * with the camera server.  If the camera ip is set to zero then the connection is
 * assumed to be local otherwise the ip represents the remote ip for connected camera
 *
 * @param camera_ip:            The IP address of the camara or 0 if local server
 * @param camera_port:          The PORT argument or NULL if local server
 *
 * @return 1 on success, 0 on error.
 */
int cs_init(uint32_t camera_ip, unsigned short camera_port)
{
    int size;                       // Stores the length of the unix domain socket addr
    int err;                        // Used to test return values from pthread initialization calls
    struct timeval  time_out;       // Timeout period for receive.
    sigset_t    usr1_set;
    //int status;

    // NOTE on local_server static
    // if Camera Server is a local process or if it is a remote process this
    // module needs to know.  If remote process that means the calling process
    // is not able to receive notification messages on either liveview or about
    // FCP registers.  The only time a remote process is typically used is when
    // debugging on an Ubuntu system remotely to a Camera through IP connection

    // check the input arguments and setup the camera connection data based on them
    if (camera_ip != 0)
    {
        local_server = false;   // assume a remotely connected camera server
        dest_ip = camera_ip;
        dest_port = camera_port;
    }
    else
    {
        local_server = true;    // assume a locally connected camera server (same os)
        dest_ip = 0;
        dest_port = 0;
    }

    // Make sure we get rid of any old queues.
    if (local_server)
        unlink(CS_QUEUE_FILE);

    // Create a signal set containing only SIGUSR1 and block it for the process.
    sigemptyset(&usr1_set);
    sigaddset(&usr1_set, SIGUSR1);
    if (sigprocmask(SIG_BLOCK, &usr1_set, NULL) < 0)
    {
        fi_log(LOG_WARNING, "(%s) unable to block SIGUSR1 signal %d(%s)", __FUNCTION__,
               errno, strerror(errno));
    }

    // Initialize our thread conditional variables. These are used to
    // coordinate actions to the communications thread.
    err = pthread_cond_init(&cs_write_ready, NULL);
    if (err)
    {
        fi_log(LOG_CRIT, "(%s) pthread_cond_init failed %d(%s)", __FUNCTION__, err, strerror(err));
        return 0;
    }
    err = pthread_cond_init(&cs_reply_ready, NULL);
    if (err)
    {
        fi_log(LOG_CRIT, "(%s) pthread_cond_init failed %d(%s)", __FUNCTION__, err, strerror(err));
        return 0;
    }

    // Initialize our mutex - used to give us exclusive access to shared
    // variables, structures, etc.
    err = pthread_mutex_init(&cs_com_mutex, NULL);
    if (err)
    {
        fi_log(LOG_CRIT, "(%s) pthread_mutex_init failed %d", __FUNCTION__, err);
        return 0;
    }

    // NOTE:
    // if connecting to the Camera Server Locally will use UNIX domain
    // sockets - if not will use INET sockets for communication
    if (local_server)
    {
        struct sockaddr_un un;          // UNIX domain socket address structure

        // Create a socket to communicate to the camera server with.
        if ((cs_fd = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0)
        {
            fi_log(LOG_CRIT, "(%s) error creating UNIX domain socket errno:%d(%s)",
                   __FUNCTION__, errno, strerror(errno));
            return 0;
        }

        // Initialize our socket address structures.
        memset(&un, 0, sizeof(un));
        un.sun_family = AF_UNIX;
        memset(cs_sock_path, 0, sizeof(cs_sock_path));

        // Copy the file name for our receive queue. The camera server will use this
        // socket to send us information.
        strncpy(un.sun_path, CS_QUEUE_FILE, sizeof(un.sun_path));
        size = SUN_LEN(&un);

        // Bind the socket to a unique filename
        if (bind(cs_fd, (struct sockaddr *)&un, size) < 0)
        {
            fi_log(LOG_CRIT, "(%s) error binding to UNIX domain socket key:%s errno:%d(%s)",
                   __FUNCTION__, un.sun_path, errno, strerror(errno));
            return 0;
        }
        else
        {
            if (cs_verbose >= 3)
                fi_log(LOG_INFO, "(%s) UNIX domain socket bound to key %s",
                       __FUNCTION__, un.sun_path);
            strncpy(cs_sock_path, un.sun_path, sizeof(cs_sock_path));
        }
    }
    else
    {
        // Create a socket to communicate to the camera server with.
        if ((cs_fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
        {
            fi_log(LOG_CRIT, "(%s) error creating INET socket errno:%d(%s)",
                   __FUNCTION__, errno, strerror(errno));
            return 0;
        }
    }

    // Set the timeout.
    time_out.tv_sec = CS_RCV_TIMEOUT_SEC;
    time_out.tv_usec = CS_RCV_TIMEOUT_USEC;
    setsockopt(cs_fd, SOL_SOCKET, SO_RCVTIMEO, &time_out, sizeof(struct timeval));

    // Start thread
    if (pthread_create(&cs_thread, NULL, event_thread, NULL) != 0)
    {
        fi_log(LOG_CRIT, "(%s) unable to start event thread", __FUNCTION__);
        return 0;
    }


#ifdef COMMENT_OUT
    this isn't working right now.'
    // Register with the camera server to get asynchronous notifications. Do this
    // at the end, because once registered we have to assume we could receive a
    // notification at any time.
    //
    // Note that cs_set_reg_l() proxies the call to the worker thread, so by
    // definition, the message is not sent until the worker thread is ready to listen.
    while (!cs_set_reg_l(GIGE_REG_EVENT_NOTIFY, 1, &status))
    {
        fi_log(LOG_WARNING, "(%s) unable to set GIGE_REG_EVENT_NOTIFY.  Retrying...",
               __FUNCTION__);

        // Sleep for 1 second.
        sleep(1);
    }
#endif

    return 1;
}

/*!
 * @brief Close down the IPC channels to make sure everything is cleaned up
 *        before shutting down.
 *
 * @return 0 for success
 */
void cs_destroy()
{
    void *rval;
    int status;

    // Turn off asynchronous notifications from the camera server, but
    // only if we registered a callback because that's when we turned on notifications.
    CB_ENTRY *cbe = callbacks;
    int n_callbacks = 0;
    while (cbe != &callbacks[sizeof(callbacks) / sizeof(callbacks[0])])
    {
        if (cbe->func != NULL)
        {
            cbe->arg = NULL;
            cbe->func = NULL;
            n_callbacks++;
        }
        cbe++;
    }
    if (n_callbacks != 0)
    {
        if (!cs_set_reg_l(GIGE_REG_EVENT_NOTIFY, 0, &status))
        {
            fi_log(LOG_ERR, "(%s) unable to set GIGE_REG_EVENT_NOTIFY", __FUNCTION__);
        }
    }

    // Ask thread to quit
    cs_exit_requested = 1;

    // Does not really kill the thread, just sends the signal to the thread
    // to wake it up, so that it'll notice we've asked it to stop.
    pthread_kill(cs_thread, SIGUSR1);

    // Wait for thread to exit.
    if (pthread_join(cs_thread, &rval) != 0)
    {
        fi_log(LOG_WARNING, "(%s) unable to pthread_join", __FUNCTION__);
    }

    // If we have a valid descriptor, we'll want to close it.
    if (cs_fd >= 0)
    {
        close(cs_fd);
    }

    // Remove our queue.
    unlink(cs_sock_path);

    // Make sure any blocked threads are released.
    pthread_cond_broadcast(&cs_write_ready);
    pthread_cond_broadcast(&cs_reply_ready);
    pthread_mutex_destroy(&cs_com_mutex);
}


/*!
 * @brief Read a 32-bit register from the camera server (or cache).
 * @note This will update the cache if we do a read from the camera server.
 *
 * @param reg   Register ID we are interested in.
 * @param val   Pointer for where we want the data to go.
 *
 * @return 1 on success, 0 on failure.
 */
uint32_t cs_get_reg_l(uint32_t reg, uint32_t *val)
{
    // Sanity check our parameters.
    if (val == NULL)
    {
        fi_log(LOG_ERR, "(%s) NULL pointer in parameters", __FUNCTION__);
        return 0;
    }

    // Remember the last request.
    cs_last_reg_id = reg;

    // Go to the camera server and get the value.
    return cs_get_reg_from_cs(reg, val);

}

/*!
 * @brief Write a 32-bit register to the camera server.
 * @note This will update the cache.
 *
 * @param reg   Register ID we are interested in.
 * @param val   Value we want to write to the register.
 * @param istat Pointer for returning the status information.
 *
 * @return 1 on success, 0 on failure. On failure, refer to the
 *     contents of istat.
 */
uint32_t cs_set_reg_l(uint32_t reg, uint32_t val, int *istat)
{
    fcp_msg  msg_out;  // structure used to send a gige message
    fcp_msg  msg_in;   // structure used to receive a gige message
    unsigned char rcv_cmd;

    *istat = 0;      // status variable

    // Remember the last request.
    cs_last_reg_id = reg;

    if (cs_verbose >= 3)
        fi_log(LOG_INFO, "(%s) Entry (0x%x) (0x%x) istat(%x)", __FUNCTION__, reg, val, *istat);

    memset(&msg_out, 0, sizeof(msg_out));
    msg_out.header.req.code = GIGE_REQ_CODE;
    msg_out.header.req.flag = GIGE_FLAG_MUST_ACK_BE;
    msg_out.header.req.command = htons(GIGE_MSG_WRITEREG_CMD);
    msg_out.header.req.length = htons(2 * sizeof(uint32_t));

    msg_out.data_l[0] = htonl(reg);
    msg_out.data_l[1] = htonl(val);

    *istat = cs_send_msg_get_ack(&msg_out, &msg_in);

    if (istat == 0)
    {
        if (++regrw_fail < MAX_COMM_FAIL_ERRORS)
            fi_log(LOG_ERR, "(%s) error writing %04x register", __FUNCTION__, reg);
        return 0;
    }

    rcv_cmd = ntohs(msg_in.header.resp.command);
    if (rcv_cmd != GIGE_MSG_WRITEREG_ACK)
    {
        fi_log(LOG_ERR, "(%s) received invalid command type:%d expected GIGE_MSG_WRITEREG_ACK:%d",
               __FUNCTION__, rcv_cmd, GIGE_MSG_WRITEREG_ACK);

        return 0;
    }

    *istat = ntohs(msg_in.header.resp.status);
    if (*istat != GEV_STATUS_SUCCESS)
    {
        fi_log(LOG_INFO, "(%s) received error 0x%04x in response to GIGE_MSG_WRITEREG_CMD",
               __FUNCTION__, *istat);
        fi_log(LOG_ERR, "(%s) error writing %04x register", __FUNCTION__, reg);
        return 0;
    }

    if (cs_verbose >= 3)
        fi_log(LOG_INFO, "(%s) Exit (0x%x) (0x%x) istat(%x)", __FUNCTION__, reg, val, *istat);

    regrw_fail = 0;
    return 1;
}

/*!
 * @brief Write a 32-bit register to the camera server, with a
 *    lock and unlock around it.
 *
 * @param reg   Register ID we are interested in.
 * @param val   Value we want to write to the register.
 * @param istat Pointer for returning the status information.
 *
 * @return 1 on success, 0 on failure. On failure, refer to the
 *     contents of istat.
 */
uint32_t cs_set_reg_lock(uint32_t reg, uint32_t val, int *istat)
{
    fcp_msg  msg_out;  // structure used to send a gige message
    fcp_msg  msg_in;   // structure used to receive a gige message
    unsigned char rcv_cmd;

    *istat = 0;      // status variable

    memset(&msg_out, 0, sizeof(msg_out));
    msg_out.header.req.code = GIGE_REQ_CODE;
    msg_out.header.req.flag = GIGE_FLAG_MUST_ACK_BE;
    msg_out.header.req.command = htons(GIGE_MSG_WRITEREG_CMD);
    msg_out.header.req.length = htons(6 * sizeof(uint32_t));

    msg_out.data_l[0] = htonl(GIGE_REG_FASTEC_CCP);
    msg_out.data_l[1] = htonl(GIGE_FASTEC_CCP_CA_BE);
    msg_out.data_l[2] = htonl(reg);
    msg_out.data_l[3] = htonl(val);
    msg_out.data_l[4] = htonl(GIGE_REG_FASTEC_CCP);
    msg_out.data_l[5] = 0;

    *istat = cs_send_msg_get_ack(&msg_out, &msg_in);
    if (!istat)
    {
        if (++regrw_fail < MAX_COMM_FAIL_ERRORS)
            fi_log(LOG_ERR, "(%s) error writing %04x register", __FUNCTION__, reg);
        return 0;
    }

    rcv_cmd = ntohs(msg_in.header.resp.command);
    if (rcv_cmd != GIGE_MSG_WRITEREG_ACK)
    {
        fi_log(LOG_ERR, "(%s) received invalid command type:%d expected GIGE_MSG_WRITEREG_ACK:%d",
               __FUNCTION__, rcv_cmd, GIGE_MSG_WRITEREG_ACK);

        return 0;
    }

    *istat = ntohs(msg_in.header.resp.status);
    if (*istat != 0)
    {
        fi_log(LOG_INFO, "(%s) received error 0x%04x in response to GIGE_MSG_WRITEREG_CMD", __FUNCTION__, *istat);
        fi_log(LOG_ERR, "(%s) error writing %04x register", __FUNCTION__, reg);
        return 0;
    }

    return 1;
}

/*!
 * @brief Get a 16 bit register value
 *
 * @param reg: the register value to read
 * @param val: a pointer to the uint16_t location to store the value
 *
 * @return 1 on success, 0 on error.
 */
uint32_t cs_get_reg_w(uint32_t reg, uint16_t *val)
{
    fcp_msg  msg_out;  // structure used to send a gige message
    fcp_msg  msg_in;   // structure used to receive a gige message
    int istat = 0;      // status variable
    unsigned char rcv_cmd;

    if (val == NULL)
    {
        fi_log(LOG_ERR, "(%s) NULL pointer in parameters", __FUNCTION__);
        return 0;
    }

    memset(&msg_out, 0, sizeof(msg_out));
    msg_out.header.req.code = GIGE_REQ_CODE;
    msg_out.header.req.flag = GIGE_FLAG_MUST_ACK_BE;
    msg_out.header.req.command = htons(GIGE_MSG_READREG_CMD);
    msg_out.header.req.length = htons(sizeof(uint32_t));  // 1 registers
    msg_out.data_l[0] = htonl(reg);

    istat = cs_send_msg_get_ack(&msg_out, &msg_in);
    if (!istat)
    {
        if (++regrw_fail < MAX_COMM_FAIL_ERRORS)
            fi_log(LOG_ERR, "(%s) error reading 0x%04x", __FUNCTION__, reg);
        return 0;
    }

    rcv_cmd = ntohs(msg_in.header.resp.command);
    if (rcv_cmd != GIGE_MSG_READREG_ACK)
    {
        fi_log(LOG_ERR, "(%s) received invalid command type:%d expected GIGE_MSG_READREG_ACK:%d",
               __FUNCTION__, rcv_cmd, GIGE_MSG_READREG_ACK);
        return 0;

    }

    *val = ntohs(msg_in.data_w[0]);

    regrw_fail = 0;
    return 1;
}


/*!
 * @brief Get a 8 bit register value
 *
 * @param reg: the register value to read
 * @param val: a pointer to the char location to store the value
 *
 * @return 1 on success, 0 on error.
 */
uint32_t cs_get_reg_b(uint32_t reg, char *val)
{
    fcp_msg  msg_out;  // structure used to send a gige message
    fcp_msg  msg_in;   // structure used to receive a gige message
    int istat = 0;      // status variable
    unsigned char rcv_cmd;

    if (val == NULL)
    {
        fi_log(LOG_ERR, "(%s) NULL pointer in parameters", __FUNCTION__);
        return 0;
    }

    memset(&msg_out, 0, sizeof(msg_out));
    msg_out.header.req.code = GIGE_REQ_CODE;
    msg_out.header.req.flag = GIGE_FLAG_MUST_ACK_BE;
    msg_out.header.req.command = htons(GIGE_MSG_READREG_CMD);
    msg_out.header.req.length = htons(sizeof(uint32_t));  // 1 registers
    msg_out.data_l[0] = htonl(reg);

    istat = cs_send_msg_get_ack(&msg_out, &msg_in);
    if (!istat)
    {
        if (++regrw_fail < MAX_COMM_FAIL_ERRORS)
            fi_log(LOG_ERR, "(%s) error reading GIGE_REG_SENSOR_TYPE", __FUNCTION__);
        return 0;
    }

    rcv_cmd = ntohs(msg_in.header.resp.command);
    if (rcv_cmd != GIGE_MSG_READREG_ACK)
    {
        fi_log(LOG_ERR, "(%s) received invalid command type:%d expected GIGE_MSG_READREG_ACK:%d",
               __FUNCTION__, rcv_cmd, GIGE_MSG_READREG_ACK);
        return 0;

    }

    *val = msg_in.data_b[0];
    regrw_fail = 0;
    return 1;

}

uint32_t cs_send_msg(uint16_t command, uint16_t resp_cmd,
                     void *send_data_ptr, uint16_t send_data_len,
                     void *recv_data_ptr, uint16_t recv_buf_len)
{
    fcp_msg  msg_out;  // structure used to send a gige message
    fcp_msg  msg_in;   // structure used to receive a gige message
    int istat = 0;      // status variable
    uint16_t rcv_cmd;
    uint16_t recv_len;

    memset(&msg_out, 0, sizeof(msg_out));
    msg_out.header.req.code = GIGE_REQ_CODE;
    msg_out.header.req.flag = GIGE_FLAG_ALLOW_BROADCAST_BE | GIGE_FLAG_MUST_ACK_BE;
    msg_out.header.req.command = htons(command);
    msg_out.header.req.length = htons(send_data_len);

    // Is there any data?
    if (send_data_len > 0)
        (void)memcpy(msg_out.data_b, send_data_ptr, send_data_len);

    istat = cs_send_msg_get_ack(&msg_out, &msg_in);
    if (!istat)
    {
        if (++regrw_fail < MAX_COMM_FAIL_ERRORS)
            fi_log(LOG_ERR, "(%s) error reading GIGE_REG_SENSOR_TYPE", __FUNCTION__);
        return 0;
    }
    regrw_fail = 0;

    rcv_cmd = ntohs(msg_in.header.resp.command);
    if (rcv_cmd != resp_cmd)
    {
        fi_log(LOG_ERR, "(%s) received invalid command type:%d expected GIGE_MSG_READREG_ACK:%d",
               __FUNCTION__, rcv_cmd, resp_cmd);
        return 0;

    }

    // Are we supposed to copy back any data?
    recv_len = 0;
    if ((recv_data_ptr != NULL) && (recv_buf_len > 0))
    {
        // How many bytes are in the receive buffer?
        recv_len = ntohs(msg_in.header.resp.length);
        if (recv_buf_len < recv_len)
        {
            // Truncate the buffer.
            fi_log(LOG_WARNING, "%s: Response length (%d) > recv_buf_len (%d)",
                   __FUNCTION__, recv_len, recv_buf_len);
            recv_len = recv_buf_len;
        }

        // Copy the received data over.
        (void)memcpy(recv_data_ptr, msg_in.data_b, recv_len);
    }

    // Return the number of bytes copied to the receive buffer.
    return (uint32_t)recv_len;
}

/* added this from the camera_server.c file of the ts3 for microwindows on camera gui */

static void preprocess_string_fetch(int reg, int reg_data_len, uint16_t **word_ptr,  uint32_t **msg_ptr, uint16_t *write_len, uint16_t *read_len)
{
    // Build up the message.
    // The data portion of the message has a 32 bit address,
    // 16 bit reserved space, and a 16 bit length value.
    *msg_ptr = ((uint32_t *)temp_buffer);
    (*msg_ptr)[ GIGE_READMEM_REQ_ADDR_L ] = htonl(((uint32_t)reg));

    // Put the length of the array after the address(32 bits entered above)
    // and the 16 bits of GigE Vision reserved space.
    *word_ptr = ((uint16_t *)&temp_buffer);
    (*word_ptr)[ GIGE_READMEM_REQ_COUNT_W ] = htons(((uint16_t)reg_data_len));

    // Setup the write and read lengths.
    *write_len = sizeof(uint32_t) + (2 * sizeof(uint16_t));
    *read_len = sizeof(uint32_t) + reg_data_len;
}

/*!
 * @brief This writes camera server array structures, such as strings.
 *
 * @param array_id The ID of the array we want to write.
 * @param data_ptr A pointer to a buffer with data to write to the server.
 * @param data_len The number of bytes which we are to write to the server.
 *
 * @return The number of bytes written, or < 0 on error.
 */
int cs_write_array(int array_id, const void *data_ptr, uint16_t data_len)
{
    int     rc;
    uint16_t    write_len;
    uint32_t    *msg_ptr;
    uint16_t    reg_data_len = 0;

    // Sanity Check.
    // If the node or the list are non-existent then the node is not visible.
    if (data_ptr == NULL)
    {
        fi_log(LOG_ERR, "%s: NULL pointer passed to function.", __FUNCTION__);
        return -1;
    }
    if ((data_len == 0) || (data_len > GIGE_WRITEMEM_B_LEN))
    {
        fi_log(LOG_ERR, "%s: Invalid data len (0x%x).", __FUNCTION__, data_len);
        return -1;
    }

    // set the known lengths for the array item being stored
    //TODO - revisit this!
    switch (array_id)
    {
    //generic registers
    case GIGE_REG_USER_DEF_NAME:
    case GIGE_REG_IMAGE_TAG:
    case GIGE_REG_COMMENT:
    case GIGE_REG_WIFI_SSID:
    case GIGE_REG_WIFI_PASSPHRASE:
    case GIGE_REG_LOAD_FILE_NAME:
    case GIGE_REG_FILEPATH:
    case GIGE_REG_NTP_SERVER:
        switch (array_id)
        {
        // figure out the max length
        case GIGE_REG_USER_DEF_NAME:
            reg_data_len = GIGE_REG_USER_DEF_NAME_B_LEN;
            break;
        case GIGE_REG_IMAGE_TAG:
            reg_data_len = GIGE_IMAGE_TAG_MAX_LEN;
            break;
        case GIGE_REG_COMMENT:
            reg_data_len = GIGE_COMMENT_MAX_LEN;
            break;
        case GIGE_REG_WIFI_SSID:
            reg_data_len = GIGE_WIFI_SSID_MAX;
            break;
        case GIGE_REG_WIFI_PASSPHRASE:
            reg_data_len = GIGE_WIFI_PASSPHRASE_MAX;
            break;
        case GIGE_REG_LOAD_FILE_NAME:
            reg_data_len = GIGE_LOAD_FILE_NAME_MAX;
            break;
        case GIGE_REG_FILEPATH:
            reg_data_len = GIGE_FILEPATH_LEN;
            break;
        case GIGE_REG_NTP_SERVER:
            reg_data_len = GIGE_NTP_SERVER_LEN;
            break;
        default:
            break;
        }

        // Make sure the length makes sense.
        if (data_len > reg_data_len)
        {
            fi_log(LOG_WARNING, "%s: Register 0x%04x data_len (%u) is wrong.", __FUNCTION__, array_id, data_len);
            rc = -1;
            write_len = -1;
        }
        else
        {
            // Build up the message.
            // The message without headers has the address
            // and the data to write to it.
            msg_ptr = ((uint32_t *)temp_buffer);
            msg_ptr[ GIGE_WRITEMEM_REQ_ADDR_L ] = htonl(((uint32_t)array_id));
            memcpy(&temp_buffer[ GIGE_WRITEMEM_REQ_DATA_B ], data_ptr, data_len);

            // Initialize the number of bytes we'll be writing.
            write_len = sizeof(uint32_t) + data_len;
        }
        break;

    // other non-generic cases go here
    //case GIGE....
    //    break;

    default:
        fi_log(LOG_ERR, "%s: Invalid update (%d)", __FUNCTION__, array_id);
        rc = -1;
        write_len = -1;
        break;
    }

    // Do we have data to write?
    if ((write_len > 0) && (write_len != ((uint16_t) -1)))
    {
        // Send the message to write the register.
        rc = cs_send_msg(GIGE_MSG_WRITEMEM_CMD, GIGE_MSG_WRITEMEM_ACK,
                         temp_buffer, write_len, NULL, 0);
    }

    // Return the bytes written on success.
    if (rc == 0)
        rc = data_len;

    return rc;
}

/*!
 * @brief This reads camera server array structures, such as strings.
 *
 * @param array_id The ID of the array we want to read.
 * @param data_ptr A pointer to a buffer we are going to fill with data.
 * @param data_len The maximum number of bytes which can go into that buffer.
 *
 * @return The number of bytes read, or < 0 on error.
 */
int cs_read_array(int array_id, void *data_ptr, uint16_t data_len)
{
    int     rc;
    char    *byte_ptr;
    uint16_t    write_len;
    uint16_t    read_len;
    uint16_t    *word_ptr;
    uint32_t    *msg_ptr;
    uint16_t     msg_len = 0;

    // Sanity check. Make sure there is a data_ptr and room for the data.
    if ((data_ptr == NULL) || (data_len == 0) || (data_len > GIGE_READMEM_B_LEN))
    {
        fi_log(LOG_ERR, "%s: sanity check failed. data_len = %u", __FUNCTION__, data_len);
        return -1;
    }

    // Perform any array specific pre-processing which might be required.
    switch (array_id)
    {
    case GIGE_REG_DEVICE_VER:
        msg_len = GIGE_REG_DEVICE_VER_B_LEN;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_MODEL_NAME data_len (%u) is wrong.", __FUNCTION__, data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_BOOTLOADER_VER:
        msg_len = 32;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_BOOTLOADER_VER data_len (%u) is wrong.", __FUNCTION__, data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_SENSOR_VER:
        msg_len = GIGE_SENSOR_VER_LEN;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_SENSOR_VER data_len (%u) is wrong.", __FUNCTION__, data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_SERIAL_NUM:
        msg_len = GIGE_REG_SERIAL_NUM_B_LEN;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_SERIAL_NUM data_len (%u) is wrong.", __FUNCTION__, data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_MODEL_NAME:
        msg_len = GIGE_REG_MODEL_NAME_B_LEN;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_MODEL_NAME data_len (%u) is wrong.", __FUNCTION__, data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_WIFI_SSID:
        msg_len = GIGE_WIFI_SSID_MAX;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_WIFI_SSID data_len (%u) is wrong.", __FUNCTION__, data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_WIFI_PASSPHRASE:
        msg_len = GIGE_WIFI_PASSPHRASE_MAX;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_WIFI_PASSPHRASE data_len (%u) is wrong.", __FUNCTION__, data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_USER_DEF_NAME:
        // Sanity check.
        msg_len = GIGE_REG_USER_DEF_NAME_B_LEN;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_USER_DEF_NAME data_len (%u) is wrong.", __FUNCTION__, data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_FRAME_METADATA:
        // Sanity check.
        msg_len = GIGE_FRAME_METADATA_LEN;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_FRAME_METADATA data_len (%u) is wrong.", __FUNCTION__, data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_NTP_SERVER:
        msg_len = GIGE_NTP_SERVER_LEN;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_NTP_SERVER data_len (%u) is wrong.", __FUNCTION__, data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_IMAGE_TAG:
        msg_len = GIGE_IMAGE_TAG_MAX_LEN;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_IMAGE_TAG data_len (%u) is wrong.", __FUNCTION__, data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_COMMENT:
        msg_len = GIGE_COMMENT_MAX_LEN;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_COMMENT data_len (%u) is wrong.", __FUNCTION__,
                   data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_MAN_NAME:
        msg_len = GIGE_REG_MAN_NAME_B_LEN;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_MAN_NAME_B_LEN data_len (%u) is wrong.", __FUNCTION__,
                   data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;

    case GIGE_REG_MAN_INFO:
        msg_len = GIGE_REG_MAN_INFO_B_LEN;
        if (data_len != msg_len)
            fi_log(LOG_WARNING, "%s: GIGE_REG_MAN_INFO_B_LEN data_len (%u) is wrong.", __FUNCTION__,
                   data_len);
        preprocess_string_fetch(array_id, msg_len, &word_ptr, &msg_ptr, &write_len, &read_len);
        break;


    case GIGE_REG_MARKER_LIST:
        // Build up the message.
        msg_ptr = ((uint32_t *)temp_buffer);
        msg_ptr[ GIGE_READMEM_REQ_ADDR_L ] = htonl(((uint32_t)array_id));
        word_ptr = ((uint16_t *)temp_buffer);
        word_ptr[GIGE_READMEM_REQ_COUNT_W] = htons((uint16_t)data_len);

        // Set up the write and read lengths.
        msg_len = data_len;
        write_len = GIGE_READMEM_REQ_B_LEN;
        read_len = sizeof(uint32_t) + data_len;
        break;

    case GIGE_REG_HISTOGRAM_RAW:
    case GIGE_REG_HISTO_RGB:
    case GIGE_REG_WIFI_SCAN_RESULTS:
        // Build up the message.
        // The data portion of the message has a 32 bit address,
        msg_ptr = ((uint32_t *)temp_buffer);
        msg_ptr[ GIGE_READMEM_REQ_ADDR_L ] = htonl(((uint32_t)array_id));
        word_ptr = ((uint16_t *)temp_buffer);
        word_ptr[GIGE_READMEM_REQ_COUNT_W] = htons((uint16_t)512);

        // Setup the write and read lengths.
        write_len = GIGE_READMEM_REQ_B_LEN;
        read_len = sizeof(uint32_t) + 512;
        break;

    default:
        fi_log(LOG_ERR, "%s: Invalid array (%d)", __FUNCTION__, array_id);
        return -1;
    }

    // Send the request for the information.
    // If there are bytes to return, re-use the temp_buffer for the reply.
    rc = cs_send_msg(GIGE_MSG_READMEM_CMD, GIGE_MSG_READMEM_ACK,
                     temp_buffer, write_len,
                     ((read_len > 0) ? temp_buffer : NULL), read_len);

    // Just leave if we got an error or there is nothing left to process.
    if (rc <= 0)
        return rc;

    // Perform any post processing required.
    switch (array_id)
    {
    case GIGE_REG_IMAGE_TAG:
    case GIGE_REG_WIFI_SSID:
    case GIGE_REG_WIFI_PASSPHRASE:
    case GIGE_REG_NTP_SERVER:
        // Set our return count appropriately.
        rc = msg_len;

        // Copy the data over to our return pointer. Make sure we skip
        // over the ID word.
        memcpy(data_ptr, &temp_buffer[4], rc);

        // Make sure we are null terminated. The 'data_len' argument is
        // the maximum number of bytes which can be held by the buffer
        // pointed to by 'data_ptr'. By setting the last byte to a 0, we
        // are effectively null terminating the string.
        byte_ptr = (char *)data_ptr;
        byte_ptr[ data_len - 1 ] = 0;
        break;

    // For these, the return length may be smaller than what was requested.
    case GIGE_REG_DEVICE_VER:
    case GIGE_REG_MODEL_NAME:
    case GIGE_REG_COMMENT:
    case GIGE_REG_MARKER_LIST:
    case GIGE_REG_SERIAL_NUM:
    case GIGE_REG_SENSOR_VER:
    case GIGE_REG_BOOTLOADER_VER:
    case GIGE_REG_MAN_NAME:
    case GIGE_REG_MAN_INFO:
    case GIGE_REG_USER_DEF_NAME:
        // Set our return count appropriately.
        if (rc > 0)
            rc -= sizeof(uint32_t);

        // Copy the data over to our return pointer. Make sure we skip
        // over the ID word.
        if (rc > 0)
            memcpy(data_ptr, &temp_buffer[4], rc);
        break;

    case GIGE_REG_FRAME_METADATA:
        // Set our return count appropriately.
        rc = msg_len;

        // Copy the data over to our return pointer. Make sure we skip
        // over the ID word.
        memcpy(data_ptr, &temp_buffer[4], rc);
        break;

    case GIGE_REG_HISTOGRAM_RAW:
    case GIGE_REG_HISTO_RGB:
    case GIGE_REG_WIFI_SCAN_RESULTS:
        // Set our return count appropriately.
        rc = 512;

        // Copy the data over to our return pointer. Make sure we skip
        // over the ID word.
        memcpy(data_ptr, &temp_buffer[4], rc);
        break;

    // Don't know why this function requires explicitly listing every register it supports.
    default:
        fi_log(LOG_DEBUG, "%s: Not copying data for 0x%04x rc=%d", __FUNCTION__, array_id, rc);
        break;
    }

    return rc;
}
