/*!
 * @file fcp_datamgr.cxx
 * @brief all support for managing request/response data for the FCP protocol
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>

#include <libwebsockets.h>

#include <fastec/gige_msg.h>
#include <fastec/fas_log.h>

#include "fcp.h"

#include "fi_wss_main.h"
#include "fi_cs.h"
#include "fi_jsonpkg.h"
#include "fcp_datamgr.h"

// TODO turn this into a C++ object - remember that the callback function
// is a static member of the protocol object and will not have access to
// member variables that are not static.

// create a define to represent the maximum size allowed for a FCP memory
// buffer - even though the message size is FCP_MSG_DATA_SIZE this is the
// size we will restrict the payload to...
#define MAX_FCP_MEMORY_BUFFER      512

/* --------------------------------------------------------------------------------
 * these defines hold the different conditions that can be recorded when servicing
 * a request session for the fcp protocol
 * --------------------------------------------------------------------------------*/
#define FCP_PKGSTATUS_NOT_STARTED               0   // the requesting or responding has not started
#define FCP_PKGSTATUS_OK                        1   // the processing of the request was successful
#define FCP_PKGSTATUS_PARSE_BAD_INIT            2   // unable to initialize JSON Parsing data
#define FCP_PKGSTATUS_PARSE_FAILED              3   // the JSON parser failed to parse the input
#define FCP_PKGSTATUS_PARSE_ACTION_FAILED       4   // there is no action code in the request
#define FCP_PKGSTATUS_PARSE_FIELDS_FAILED       5   // the number of fields value is missing or bad
#define FCP_PKGSTATUS_PARSE_BAD_FCP_REG        6   // the FCP register requested is bad
#define FCP_PKGSTATUS_PARSE_BAD_FCP_DATA       7   // the FCP data for a store is bad (either missing or invalid)
#define FCP_PKGSTATUS_BUILD_FAILED              8   // the request or response record was unable to build
#define FCP_PKGSTATUS_FIELDSIZE_EXCEEDED        9   // the fieldsize exceeds package payload size
#define FCP_PKGSTATUS_B64_ENCODE_FAILED         10  // base 64 encoding failure
#define FCP_PKGSTATUS_CS_ERROR                  11  // the CS was not able to service the request
#define FCP_PKGSTATUS_NO_CS_AVAIL               12  // the Camera Server Connection was not found
#define FCP_PKGSTATUS_SKIPPED                   13  // Skipped a load/store field due to an earlier error
#define FCP_PKGSTATUS_CS_LOCK_FAILED            14  // could not get Fastec CCP lock
#define FCP_PKGSTATUS_CS_DENIED_ACCESS          15  // Camera Server denied access (no FASTEC_CCP)


/* --------------------------------------------------------------------------------
 ***********************************************************************************
 * Data Management for:   fcp-listener protocol listeners
 ***********************************************************************************
 * --------------------------------------------------------------------------------*/
#define MAX_FCP_LISTENERS 16    // right now this is set to 16 but what should it be set to? TODO

/* --------------------------------------------------------------------------------
 * the fcp-listener-protocol will allow up to a maximum of MAX_FCP_LISTENERS
 * and if the listener is active it will be
 * --------------------------------------------------------------------------------*/
static per_session_data_fcp_listener *all_fcp_listeners[MAX_FCP_LISTENERS];
static int fcp_listeners_initialized = 0;  // make sure we know if the caller initialized the listeners
static pthread_mutex_t fcp_listeners_mutex; //!< Serialize access to all_fcp_listeners[]

/* --------------------------------------------------------------------------------
 ***********************************************************************************
 * Data Management for:   fcp protocol RESPONSE Header Strings
 ***********************************************************************************
 * --------------------------------------------------------------------------------*/
static char fcp_response_action_load[]      = "\"action\":\"load\"";            // action field for load
static char fcp_response_action_store[]     = "\"action\":\"store\"";           // action field for store
static char fcp_response_action_notify[]    = "\"action\":\"notify\"";          // action field for store
static char fcp_response_action_bad[]       = "\"action\":\"unknown\"";         // action field for unknown
static char fcp_response_pkgstatus[]        = "\"pkginfo\"";                    // package information
static char fcp_response_pkg_field_list[]   = "\"fields\"";                     // response package fields in sub-package

static char fcp_response_field_delimeter[]  = ":";                           // what goes between key and data
static char fcp_response_field_sep[]        = ",";                           // what goes between fields if multiple
static char fcp_response_field_list_start[] = "[";                          // what starts json response field list
static char fcp_response_field_list_end[]   = "]";                           // what ends json response field list
static char fcp_response_field_keyquote[]   = "\"";                           // how are key strings in a field item quoted

// JSON package and field item delimeters
static char fcp_response_pkg_start[]       = "{";                            // what starts json response package
static char fcp_response_pkg_end[]         = "}";                            // what ends json response package

static char fcp_all_notifications[]         = "ALL";

// package status - status strings - the string data returned to caller in addition to the codes
// if there is an error the format of the string returned for pkg_status is in Error.XX.Description.YY
// where the error string is noted followed by a dot followed by the fi fcp protocol error code followed by
// a dot then the description followed by a dot if additional sub-error code YY is also found if one was
// received from the JSON parser or if one was received from the Camera Server
static char fcp_pkgstatus_complete[]             = "Complete";                               //FCP_PKGSTATUS_OK
static char fcp_pkgstatus_cs_error[]             = "CS_Error";                               //FCP_PKGSTATUS_CS_ERROR

static char fcp_pkgstatus_parse_not_started[]    = "Waiting_for_JSON Request";               //FCP_PKGSTATUS_NOT_STARTED
static char fcp_pkgstatus_parse_bad_init[]       = "Unable_to_Initialize_JSON_Parser";       //FCP_PKGSTATUS_PARSE_BAD_INIT
static char fcp_pkgstatus_parse_failed[]         = "Invalid_JSON_Format";                    //FCP_PKGSTATUS_PARSE_FAILED
static char fcp_pkgstatus_bad_action[]           = "Invalid_Request_Action";                 //FCP_PKGSTATUS_PARSE_ACTION_FAILED
static char fcp_pkgstatus_bad_fields[]           = "Invalid_Fields";                         //FCP_PKGSTATUS_PARSE_FIELDS_FAILED
static char fcp_pkgstatus_bad_fcp[]             = "Unknown_FCP_Register";                  //FCP_PKGSTATUS_PARSE_BAD_FCP_REG
static char fcp_pkgstatus_bad_fcp_data[]        = "FCP_Data_For_Store_Missing_or_Invalid"; //FCP_PKGSTATUS_PARSE_BAD_FCP_DATA
static char fcp_pkgstatus_build_failed[]         = "Package_Build_Failure";                  //FCP_PKGSTATUS_BUILD_FAILED
static char fcp_pkgstatus_no_cs_avail[]          = "No Camera Server Connection Available";  //FCP_PKGSTATUS_NO_CS_AVAIL
static char fcp_pkgstatus_exceeds_size[]         = "Exceeds_Max_Field_Size";                 //FCP_PKGSTATUS_EXCEED_SIZE
static char fcp_pkgstatus_b64_failed[]           = "Base64_Encoder_Failure";                 //FCP_PKGSTATUS_B64_ENCODE_FAILED
static char fcp_pkgstatus_skipped[]              = "Skipped";                                //FCP_PKGSTATUS_SKIPPED
static char fcp_pkgstatus_cs_lock_failed[]       = "Could not lock";             //FCP_PKGSTATUS_CS_LOCK_FAILED
static char fcp_pkgstatus_unknown[]              = "Unknown_Error";                          // catch-all for unknown package status

// Serial number for FCP and FCP Notify packets
static uint8_t fcp_sn = 0;

/* --------------------------------------------------------------------------------
 ***********************************************************************************
 * Support for Package Status Errors
 ***********************************************************************************
 * --------------------------------------------------------------------------------*/

/* --------------------------------------------------------------------------------
 * @brief Translate Package Status Code into User Readable String
 *
 * translates the package status integer value into a character readable message
 *
 * @param pkg_status:                   the package status setting for
 * @param p_output:                     the pointer to the output buffer
 *
 * @return pointer to the user readable string that maps to the package status value
 * --------------------------------------------------------------------------------*/
static char *xlate_pkgstatus_to_str(int pkg_status)
{
    char *p_pkg_info;

    switch (pkg_status)
    {
    case FCP_PKGSTATUS_NOT_STARTED:
        p_pkg_info = fcp_pkgstatus_parse_not_started;
        break;
    case FCP_PKGSTATUS_OK:
        p_pkg_info = fcp_pkgstatus_complete;
        break;
    case FCP_PKGSTATUS_PARSE_BAD_INIT:
        p_pkg_info = fcp_pkgstatus_parse_bad_init;
        break;
    case FCP_PKGSTATUS_PARSE_FAILED:
        p_pkg_info = fcp_pkgstatus_parse_failed;
        break;
    case FCP_PKGSTATUS_PARSE_ACTION_FAILED:
        p_pkg_info = fcp_pkgstatus_bad_action;
        break;
    case FCP_PKGSTATUS_PARSE_FIELDS_FAILED:
        p_pkg_info = fcp_pkgstatus_bad_fields;
        break;
    case FCP_PKGSTATUS_PARSE_BAD_FCP_REG:
        p_pkg_info = fcp_pkgstatus_bad_fcp;
        break;
    case FCP_PKGSTATUS_PARSE_BAD_FCP_DATA:
        p_pkg_info = fcp_pkgstatus_bad_fcp_data;
        break;
    case FCP_PKGSTATUS_BUILD_FAILED:
        p_pkg_info = fcp_pkgstatus_build_failed;
        break;
    case FCP_PKGSTATUS_CS_ERROR:
        p_pkg_info = fcp_pkgstatus_cs_error;
        break;
    case FCP_PKGSTATUS_FIELDSIZE_EXCEEDED:
        p_pkg_info = fcp_pkgstatus_exceeds_size;
        break;
    case FCP_PKGSTATUS_B64_ENCODE_FAILED:
        p_pkg_info = fcp_pkgstatus_b64_failed;
        break;
    case FCP_PKGSTATUS_NO_CS_AVAIL:
        p_pkg_info = fcp_pkgstatus_no_cs_avail;
        break;
    case FCP_PKGSTATUS_SKIPPED:
        p_pkg_info = fcp_pkgstatus_skipped;
        break;
    case FCP_PKGSTATUS_CS_LOCK_FAILED:
        p_pkg_info = fcp_pkgstatus_cs_lock_failed;
        break;

    default:
        p_pkg_info = fcp_pkgstatus_unknown;
        break;
    }

    return p_pkg_info;
}

/* --------------------------------------------------------------------------------
 * @brief Build user message for the package status
 *
 * build the package status user message
 *
 * @param pkg_status:                   the package status setting for
 * @param json_code:                    the result of the json parse
 * @param p_output:                     the pointer to the output buffer
 *
 * @return void
 * --------------------------------------------------------------------------------*/
static void get_pkg_status_msg(int pkg_status, int json_code, char *p_output)
{
    if (p_output)
    {
        char *p_pkg_info = xlate_pkgstatus_to_str(pkg_status);

        // pkgstatus will be a sub-object comprised of keyfields to denote all information from
        // the request
        sprintf((char *)p_output, "{\"status_code\":\"%d\",\"status_info\":\"%s\",\"json_parser_code\":\"%d\"}",
                pkg_status,
                p_pkg_info,
                json_code);
    }
}


/* --------------------------------------------------------------------------------
 ***********************************************************************************
 * Utility Support for FCP Request Registers
 ***********************************************************************************
 * --------------------------------------------------------------------------------*/

/* --------------------------------------------------------------------------------
 * @brief Determines if the fcp register is a memory register
 *
 * call this to determine if the fcp register is a memory register
 *
 * @param in_cs_reg:                    the cs actual register value
 * @param is_fcp_mem_req_ptr:          pointer to boolean to return whether or not this is mem array
 * @param nbytes_in_fcp_mem_ptr:       pointer to the number of bytes that should exist for this memory
 * @param p_encode_this:                pointer to boolean to indicate whether or not cs reg needs response encoding
  *
 * @return 1 if the register was processed or 0 if the pointer(s) were invalid
 * --------------------------------------------------------------------------------*/
static int is_cs_reg_mem(uint32_t cs_reg, bool *p_is_fcp_mem_req, uint16_t *p_nbytes_in_fcp_mem, bool *p_encode_this)
{
    int result = 0;

    // make sure the output pointers are valid before continuing
    if ((p_is_fcp_mem_req != (bool *) NULL) && (p_nbytes_in_fcp_mem != (uint16_t *) NULL) && (p_encode_this != (bool *)NULL))
    {
        result = 1;
        *p_encode_this = false; // most are not encoded on output
        switch (cs_reg)
        {
        case GIGE_REG_USER_DEF_NAME:
            *p_is_fcp_mem_req = true;
            *p_encode_this = true;  // for now as a test let's encode the camera name
            *p_nbytes_in_fcp_mem = GIGE_REG_USER_DEF_NAME_B_LEN;
            break;
        case GIGE_REG_MODEL_NAME:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_REG_MODEL_NAME_B_LEN;
            break;
        case GIGE_REG_SERIAL_NUM:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_REG_SERIAL_NUM_B_LEN;
            break;
        case GIGE_REG_SENSOR_VER:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_SENSOR_VER_LEN;
            break;
        case GIGE_REG_IMAGE_TAG:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_IMAGE_TAG_MAX_LEN;
            break;
        case GIGE_REG_BOOTLOADER_VER:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = 32;
            break;
        case GIGE_REG_FILEPATH:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_FILEPATH_LEN;
            break;
        case GIGE_REG_LOAD_FILE_NAME:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_LOAD_FILE_NAME_MAX;
            break;
        case GIGE_REG_WIFI_SSID:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_WIFI_SSID_MAX;
            break;
        case GIGE_REG_WIFI_PASSPHRASE:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_WIFI_PASSPHRASE_MAX;
            break;
        case GIGE_REG_WIFI_SCAN_RESULTS:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_WIFI_SCAN_RESULTS_READ_SIZE;
            break;
        case GIGE_REG_COMMENT:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_COMMENT_MAX_LEN;
            break;

        // pulled out of gigemsg.h that had a definite length associated with it indicating string
        case GIGE_REG_MAN_NAME:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_REG_MODEL_NAME_B_LEN;
            break;
        case GIGE_REG_DEVICE_VER:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_REG_DEVICE_VER_B_LEN;
            break;
        case GIGE_REG_MAN_INFO:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_REG_MAN_INFO_B_LEN;
            break;

        // pulled out of camera_server.c that wasn't in the INGE above and was being pre or post processed
        case GIGE_REG_FRAME_METADATA:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_FRAME_METADATA_LEN;
            break;
        case GIGE_REG_NTP_SERVER:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = GIGE_NTP_SERVER_LEN;
            break;
        case GIGE_REG_MARKER_LIST:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = MAX_FCP_MEMORY_BUFFER;
            break;
        case GIGE_REG_HISTOGRAM_RAW:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = MAX_FCP_MEMORY_BUFFER;
            break;
        case GIGE_REG_HISTO_RGB:
            *p_is_fcp_mem_req = true;
            *p_nbytes_in_fcp_mem = MAX_FCP_MEMORY_BUFFER;
            break;
        default:
            *p_is_fcp_mem_req = false;
            *p_nbytes_in_fcp_mem = 0;
            break;
        }
    }

    return result;
    \
}

/* --------------------------------------------------------------------------------
 ***********************************************************************************
 * Data Management for:   fcp_request_field
 ***********************************************************************************
 * --------------------------------------------------------------------------------*/

/* --------------------------------------------------------------------------------
 * @brief Creates a fcp_request_field record
 *
 * This is the mechanism to create a fcp_request record.  Memory is allocated that
 * must be freed after use by calling free_fcp_request_field this is called outside
 * of the normal field generation use in the parsing logic of FCP request
 *
 * @param p_cs_reg_name:                read only pointer to the cs register name
 * @param in_cs_reg:                    the cs actual register value
 * @param in_is_register_an_array:      is the cs register a memory array
 * @param in_cs_memsize:                if cs register is memory array how big does it need to be
 * @param encode_response:              if using_memory this indicates whether or not response data needs to be encoded
 *
 * @return allocated pointer to the fcp_request_field that was created or NULL if failed
 * --------------------------------------------------------------------------------*/
static fcp_request_field *
create_fcp_request_field(const char *p_cs_reg_name,              // the camera server register name
                         uint32_t in_cs_reg,             // the camera server register value
                         bool in_is_register_an_array,   // is the data a memory array or 32 bit register
                         uint32_t in_cs_memsize,         // how many bytes for data if memory array
                         bool in_encode_response)        // if the response for this field needs to be encoded or not
{
    // creation status is set false until all is proven to be valid
    bool created_ok = false;

    // create a null pointer to return if an error ccurs
    fcp_request_field *p_result = (fcp_request_field *)NULL;
    uint8_t *p_memory_data = (uint8_t *)NULL;

    // if need to allocate pointer to hold a memory request do so -
    if (in_is_register_an_array)
    {
        // allocate the memory data but make sure it is within the range of fcp memory message size
        if ((in_cs_memsize > 0) && (in_cs_memsize <= MAX_FCP_MEMORY_BUFFER))
        {
            size_t nbytes = in_cs_memsize + 1;
            p_memory_data = (uint8_t *) malloc(nbytes);
            // make the memory buffer marked with zeros
            if (p_memory_data)
            {
                memset((void *) p_memory_data, 0, nbytes);
            }
        }
    }
    // allocate the request field record
    p_result = (fcp_request_field *) malloc((size_t)sizeof(fcp_request_field));

    // if have a valid field pointer populate with the correct information
    if (p_result != (fcp_request_field *) NULL)
    {
        // copy over the string that names the GIGE register restrict to max size just in case
        strncpy((char *)p_result->cs_reg_name, (char *)p_cs_reg_name, (size_t)FCP_MAX_GIGEREG_LABEL);

        // now copy over the camera server register in integer form
        p_result->cs_reg = in_cs_reg;

        // if the request is a memory (array) request then set up the data properly
        if (in_is_register_an_array)
        {
            if (p_memory_data != (uint8_t *) NULL)
            {
                p_result->p_cs_reg_array = p_memory_data;
                p_result->cs_reg_array_len = in_cs_memsize;
                p_result->is_register_an_array = 1; // this is a memory field
                if (in_encode_response)
                    p_result->encode_response = 1;
                else
                    p_result->encode_response = 0;
                created_ok = true;
            }
        }
        else
        {
            p_result->p_cs_reg_array = (uint8_t *)NULL;
            p_result->cs_reg_array_len = 0;
            p_result->is_register_an_array = 0; // this is not a memory field
            p_result->encode_response = 0;  // can't encode a non-array response
            created_ok = true;
        }

        // this field may contain data that is set from notification but initially
        // it must be set to zero
        p_result->is_this_notification = 0;
    }

    // if memory was not created properly free any allocations that were made
    if (!created_ok)
    {
        if (p_result != (fcp_request_field *)NULL)
        {
            free(p_result);
            p_result = (fcp_request_field *) NULL;  // reset empty
        }
        if (p_memory_data != (uint8_t *) NULL)
            free(p_memory_data);
    }

    // return the pointer to the created request field
    return p_result;
}


/* --------------------------------------------------------------------------------
 * @brief Frees memory allocated in a fcp_request_field record
 *
 * This is the mechanism to free allocated memory in a fcp_request record.  This is
 * normally called by the supporting functions that manage the request list.
 *
 * @param p_field_to_free:  pointer to the fcp_request_field to free
 *
 * @return void
 * --------------------------------------------------------------------------------*/
static void
free_fcp_request_field(fcp_request_field *p_field_to_free)
{
    // if the pointer is not null need to verify if it has any embedded allocations
    if (p_field_to_free != (fcp_request_field *) NULL)
    {
        if (p_field_to_free->p_cs_reg_array != (uint8_t *) NULL)
        {
            free(p_field_to_free->p_cs_reg_array);
            p_field_to_free->p_cs_reg_array = (uint8_t *) NULL;
            p_field_to_free->cs_reg_array_len = 0;
        }

        free(p_field_to_free);
    }
    return;
}

/* --------------------------------------------------------------------------------
 ***********************************************************************************
 * Data Management for:   fcp_request
 ***********************************************************************************
 * --------------------------------------------------------------------------------*/

/* --------------------------------------------------------------------------------
 * @brief Common logic used to initialize a request record to an empty state
 *
 * @param request_ptr:  pointer to the request record to initialize
 *
 * @return true if successful or false if input was invalid
 * --------------------------------------------------------------------------------*/
static bool
init_request_rec(fcp_request *p_request)
{
    if (p_request != (fcp_request *) NULL)
    {
        p_request->state = FCP_REQUEST_STATE_EMPTY;
        p_request->pkgstatus = FCP_PKGSTATUS_NOT_STARTED;
        p_request->json_return_code = 0;
        p_request->fi_cs_error = 0;
        p_request->action_type = FCP_REQUEST_ACTION_NONE;
        p_request->nfields = 0;
        p_request->p_fcp_request_fields = NULL;
        return true;
    }
    return false;
}

/* --------------------------------------------------------------------------------
 * @brief Installs an allocated field request to a fcp request record
 *
 * This is the mechanism to create and add an allocated fcp field record to the
 * specified request record - this allocated data is freed when the request record
 * is reset to empty or deleted upon session closure
 *
 * @param request_ptr:  pointer to the request record to use
 * @param cs_reg_name:  character string array identifying gige register
 * @param cs_reg:       the gige register value
 * @param using_memory: if true the field is a memory array field
 * @param memsize:      if using memory this is how big the memory array is
 * @param encode_response:  if using_memory this indicates whether or not response data needs to be encoded
 *
 * @return true if successfully added false if error
 * --------------------------------------------------------------------------------*/
static bool
add_field_to_request(fcp_request *request_ptr, char *cs_reg_name, uint32_t cs_reg,
                     bool using_memory, uint32_t memsize, bool encode_response)
{
    bool added = false;
    fcp_request_field *p_field = (fcp_request_field *) NULL;

    // make sure the request record is valid
    if (request_ptr != (fcp_request *) NULL)
    {
        // allocate the field data record
        p_field = create_fcp_request_field(cs_reg_name, cs_reg, using_memory, memsize, encode_response);

        // if pointer is valid add in the field request
        if (p_field)
        {
            // make sure the request pointer is not null
            size_t new_nfields = request_ptr->nfields + 1;

            // if no fields are allocated yet and this is the first one use malloc else use realloc
            // and this is building up an array of pointers to store the fcp_request * pointers
            if (request_ptr->nfields == 0)
            {
                request_ptr->p_fcp_request_fields = (fcp_request_field **) malloc(sizeof(fcp_request_field *));
            }
            else
            {
                request_ptr->p_fcp_request_fields =
                    (fcp_request_field **)realloc(request_ptr->p_fcp_request_fields, sizeof(fcp_request_field *) * new_nfields);
            }

            if (request_ptr->p_fcp_request_fields != NULL)
            {
                request_ptr->p_fcp_request_fields[request_ptr->nfields] = p_field;  // install the allocated pointer
                request_ptr->nfields++;
                added = true;
            }

            // if the allocated p_field was not able to be installed in array of fcp_request_fields free the memory
            if (!added)
            {
                free_fcp_request_field(p_field);
            }
        }
    }
    return added;
}


/* --------------------------------------------------------------------------------
 * @brief Frees and cleans up allocated memory used for a FCP request record
 *
 * Always call this method to free up any memory allocations for FCP request
 *
 * @param index:  pointer to a fcp request record to clean up
 *
 * @return void
 * --------------------------------------------------------------------------------*/
static void
free_field_data_for_request(fcp_request *p_fcp_request)
{
    if (p_fcp_request != (fcp_request *) NULL)
    {
        // determine how many fields were allocated
        int request_nfields = p_fcp_request->nfields;
        int field_index;
        // loop through and free the allocated data for the field
        for (field_index = 0; field_index < request_nfields; field_index++)
        {
            // this will pull the fcp_request_field pointer for the field_index and free all associated memory
            free_fcp_request_field(p_fcp_request->p_fcp_request_fields[field_index]);
        }
        free(p_fcp_request->p_fcp_request_fields);
        p_fcp_request->nfields = 0;
        p_fcp_request->p_fcp_request_fields = NULL;
    }
}

/* --------------------------------------------------------------------------------
 ***********************************************************************************
 * Data Management for:   per_session_data_fcp
 ***********************************************************************************
 * --------------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------------
* @brief Initialize a session data record
*
* call this from the fcp protocol when the callback handler receives a
* LWS_CALLBACK_ESTABLISHED message from the web socket server
*
* @param pss:  The user pointer per_session_data_fcp holding the session data
*
* @return true if successful or false if error
* --------------------------------------------------------------------------------*/
bool
init_fcp_session(per_session_data_fcp *pss)
{
    // this is supposed to be first opportunity to initialize the state and content of
    // the session data structure if the pointer is NULL that's an error otherwise assume
    // the structure is valid
    if (pss == (per_session_data_fcp *)NULL)
        return false;
    // capture the camara server pointer and have it available for
    // the callback routine handlers.
    pss->p_fiCS = fiGetCS();

    // can't continue if we don't have a connection to the camera server
    if (pss->p_fiCS == (fiCameraServerIO *) NULL)
        return false;

    // otherwise initialize the session data
    int x;
    for (x = 0; x < MAX_FCP_PER_SESSION; x++)
    {
        init_request_rec(&pss->session_requests[x]);
    }
    return true;
}

/* --------------------------------------------------------------------------------
* @brief Is there room in this session for another request?
*
* Called from the fcp protocol callback handler when a receive request package
* is about to be parsed - if there are not session request records free there are no
* more connections available that is what this function verifies
*
* @param pss:   The user pointer per_session_data_fcp holding the session data pointer
*
* @return pointer to the request record if one is available or null pointer if not
* --------------------------------------------------------------------------------*/
fcp_request *
fcp_get_empty_request(per_session_data_fcp *pss)
{
    int index_available = -1;   // set to none available

    if (pss != (per_session_data_fcp *) NULL)
    {
        // are there free slots available
        int x;
        for (x = 0; x < MAX_FCP_PER_SESSION; x++)
        {
            if (pss->session_requests[x].state == FCP_REQUEST_STATE_EMPTY)
                index_available = x;
        }
    }

    if (index_available != -1)
    {
        return &pss->session_requests[index_available];
    }
    return (fcp_request *)NULL;
}

/* --------------------------------------------------------------------------------
* @brief How many requests are pending
*
* walks the session data to see if any requests are pending
*
* @param pss:   The user pointer per_session_data_fcp holding the session data pointer
*
* @return number of request that are pending
* --------------------------------------------------------------------------------*/
int
fcp_get_nbr_session_requests(per_session_data_fcp *pss)
{
    int num_requests = 0;   // set to none available

    if (pss != (per_session_data_fcp *) NULL)
    {
        // are there free slots available
        int x;
        for (x = 0; x < MAX_FCP_PER_SESSION; x++)
        {
            if (pss->session_requests[x].state == FCP_REQUEST_STATE_EMPTY)
                num_requests++;
        }
    }
    return num_requests;
}


/* --------------------------------------------------------------------------------
* @brief Return the next available request record in session queue
*
* Called from the fcp protocol callback handler when it receives time for a
* LWS_CALLBACK_SERVER_WRITEABLE.
*
* @param pss:   The user pointer per_session_data_fcp holding the session data pointer
*
* @return pointer to the request record or null if no requests are pending
* --------------------------------------------------------------------------------*/
fcp_request *fcp_get_next_request(per_session_data_fcp *pss)
{
    fcp_request *p_result = (fcp_request *)NULL;

    if (pss != (per_session_data_fcp *) NULL)
    {
        // starting at 0 loop through to find the pending request
        int x;
        for (x = 0; x < MAX_FCP_PER_SESSION; x++)
        {
            // the next one to return should be ready to process
            if (pss->session_requests[x].state == FCP_REQUEST_STATE_READY)
            {
                p_result = &pss->session_requests[x];
                return p_result;    // got one!
            }
        }
    }
    return p_result;
}


/* --------------------------------------------------------------------------------
* @brief The request record has been sent to the caller this does housekeeping
*
* Called from the fcp protocol callback handler when it receives time for a
* LWS_CALLBACK_SERVER_WRITEABLE.
*
* @param pss:   The user pointer per_session_data_fcp holding the session data pointer
*
* @return pointer to the request record or null if no requests are pending
* --------------------------------------------------------------------------------*/
void fcp_request_sent(per_session_data_fcp *pss, fcp_request *p_result)
{
    // locate the request pointer that was processed
    if (pss != (per_session_data_fcp *) NULL)
    {
        // make sure the pointer matches one of this session's
        int x;
        for (x = 0; x < MAX_FCP_PER_SESSION; x++)
        {
            if (p_result == &pss->session_requests[x])
            {
                // free up any allocated resources for this request
                free_field_data_for_request(&pss->session_requests[x]);
                init_request_rec(&pss->session_requests[x]);
            }
        }
    }
}

/* --------------------------------------------------------------------------------
* @brief Resets all Session FCP request record pointers to empty state
*
* Use this to reset and free up allocated memory associated with a session request list
* called in response to a fcp protocol reason LWS_CALLBACK_PROTOCOL_DESTROY
*
* @param pss:   The user pointer per_session_data_fcp holding the session data
*
* @return true if successful or false if invalid pointer or failure to free allocated memory
* --------------------------------------------------------------------------------*/
bool
reset_all_session_requests(per_session_data_fcp *pss)
{
    // is the session pointer valid
    if (pss == (per_session_data_fcp *) NULL)
        return false;

    // walk the list for any request that is not empty
    int index;

    /* walk the request queue for this session and free anything that is pending */
    for (index = 0; index < MAX_FCP_PER_SESSION; index++)
    {
        // reset the session record to empty state which frees any allocated memory
        // if it is not in an empty state
        if (pss->session_requests[index].state != FCP_REQUEST_STATE_EMPTY)
            free_field_data_for_request(&pss->session_requests[index]);

        // and re-init all request records TODO This may be unnecessary...
        init_request_rec(&pss->session_requests[index])        ;
    }
    return true;
}

/* --------------------------------------------------------------------------------
* @brief Processes a JSON request package and translates into fcp request data
*
* Called from the fcp protocol callback handler when a receive request is made from
* a session client.  The JSON package is expected in a proper format for fcp requests
*
* @param pss:          The user pointer per_session_data_fcp holding the session data pointer
* @param p_json_pkg:   The JSON package received from fcp protocol to parse
* @param is_fcp       Indicate whether the calling protocol is FCP (true) or Notify (false)
*
* @return true if successful parse or false if error occurred
 * --------------------------------------------------------------------------------*/
bool
fcp_parse_request(fcp_request *p_request, void *p_json_pkg, bool is_fcp)
{
    int nfields = 0;            // how many fields to parse?
    int field_index = 0;

    // arrays to parse out a json value string for a field
    char gige_reg_label[FCP_MAX_GIGEREG_LABEL];
    char gige_reg_value[FCP_MAX_GIGEREG_VALUE];

    // only continue if input pointers are not null
    if ((p_request == (fcp_request *)NULL) || (p_json_pkg == (void *) NULL))
    {
	lwsl_err("%s: Invalid parameters!\n", __FUNCTION__);
        return false;  // invalid pointers no need to continue
    }

    // make sure the request is initialized properly
    init_request_rec(p_request);

    // the pkgstatus will keep track of the parsing progress and will record
    // any errors that occur while the JSON is parsed and set up for processing
    p_request->pkgstatus = FCP_PKGSTATUS_NOT_STARTED;

    /* adding new parsing logic to only pull out the key and data pair */
    /* initialize the json parser */
    fi_json_parse_pkg my_json_data;
    if (init_fi_json_parse_pkg(&my_json_data) == 0)
    {
	lwsl_debug("%s: init_fi_json_parse_pkg() failed.\n", __FUNCTION__);
        p_request->pkgstatus = FCP_PKGSTATUS_PARSE_BAD_INIT;
    }
    else
    {
        p_request->json_return_code = 0;    // zero means successful

        // parse the JSON data
        p_request->json_return_code = fi_json_parse((const char *)p_json_pkg, &my_json_data);
        if (p_request->json_return_code != 0)
        {
            p_request->pkgstatus = FCP_PKGSTATUS_PARSE_FAILED;
            lwsl_debug("%s: JSON Read Result: %s\n", __FUNCTION__, json_error_string(p_request->json_return_code));
        }
        else
        {
            p_request->pkgstatus = FCP_PKGSTATUS_OK;
        }
    }

    // if all is ok so far keep interpreting the request data
    if (p_request->pkgstatus == FCP_PKGSTATUS_OK)
    {
        // first determine the action code -
        const char *action_value_ptr = (const char *)find_value_for_key(&my_json_data, (uint8_t *) "action");

        // determine if action type of the request
        if (action_value_ptr != (char *)NULL)
        {
            if (is_fcp && (strcmp((const char *)action_value_ptr, (const char *)"load") == 0))
            {
                p_request->action_type = FCP_REQUEST_ACTION_LOAD;
            }
            else if (is_fcp && (strcmp((const char *)action_value_ptr, (const char *)"store") == 0))
            {
                p_request->action_type = FCP_REQUEST_ACTION_STORE;
            }
            else if ((! is_fcp) && (strcmp((const char *)action_value_ptr, (const char *)"notify")) == 0)
            {
                p_request->action_type = FCP_REQUEST_ACTION_NOTIFY;
            }
            else
            {
		lwsl_debug("%s: Bad action!\n", __FUNCTION__);
                p_request->pkgstatus = FCP_PKGSTATUS_PARSE_ACTION_FAILED;
            }
        }
        else
        {
	    lwsl_debug("%s: Parsing action failed!\n", __FUNCTION__);
            p_request->pkgstatus = FCP_PKGSTATUS_PARSE_ACTION_FAILED;
        }
    }

    // if action code was valid can not pull out the field information for the action requested
    if (p_request->pkgstatus == FCP_PKGSTATUS_OK)
    {
        // the data for the load or store should be built up in a field array - if loading there will
        // be no key data associated with each array entry only the value data which contains the
        // register that requested for loading.  If storing then both the key and the data exist where the
        // key represents the register and the value represents the new value to store

        // set the status to fail and reset to ok if all is ok
        p_request->pkgstatus = FCP_PKGSTATUS_PARSE_FIELDS_FAILED;

        // look for the fields key in the first array of the json parsed package and
        // set the number of fields in the array to process
        if (my_json_data.json_arrays_inuse > 0)
        {
            // the array primary key for the fields indicator should be in index 0
            uint16_t array_key_index = 0;
            const char *p_array_key = (const char *) get_array_key(&my_json_data, array_key_index);
            if (p_array_key != (char *)NULL)
            {
                int cmp_result = strncmp((const char *)"fields", (const char *) p_array_key, 6);
                if (cmp_result == 0)
                {
                    // how many fields were found?
                    nfields = my_json_data.json_arrays[0].array_items_inuse;
                    // only at this point when it is determined that fields were requested is
                    // the package status reset to parsed ok
                    if (nfields > 0)
                        p_request->pkgstatus = FCP_PKGSTATUS_OK;
                }
            }
        }

	if (p_request->pkgstatus != FCP_PKGSTATUS_OK)
	{
	    lwsl_debug("%s: Error parsing fields!\n", __FUNCTION__);
	}
    }

    // if made it here should be able to process the number of fields in this request
    while ((p_request->pkgstatus == FCP_PKGSTATUS_OK) && (field_index < nfields))
    {
        // action types for loading and storing and notify requests will
        // package the JSON data in the same way - the only difference is that
        // when storing the value to store is included as a value in the field
        // data so in that case both a key and data pair should exist

        // pull out the request field data from the json parsed package
        char *field_ptr = (char *) NULL;
        char *data_ptr = (char *) NULL;

        // the size of the gige memory or register data if storing
        uint32_t gige_bufsize = 0;  // each loop through reset this to zero

        // depending on action type for the data the parsed JSON is intrepreted differently
        if (p_request->action_type == FCP_REQUEST_ACTION_LOAD)
            field_ptr = (char *)my_json_data.json_arrays[0].array_items[field_index].value.p_buf;
        else if (p_request->action_type == FCP_REQUEST_ACTION_STORE)
        {
            field_ptr = (char *)my_json_data.json_arrays[0].array_items[field_index].key.p_buf;
            data_ptr = (char *)my_json_data.json_arrays[0].array_items[field_index].value.p_buf;
        }
        else if (p_request->action_type == FCP_REQUEST_ACTION_NOTIFY)
        {
            // as in the load there is no "data" associated with a key so the actual
            // notification field is stored in the data element
            field_ptr = (char *)my_json_data.json_arrays[0].array_items[field_index].value.p_buf;
        }


        // process the field request
        if (field_ptr != (char *) NULL)
        {
            // field was found the value can be broken up into the label and value for the GIGE register requested
            // the value should be in the form of the gige register label (the define string) and the actual register
            // seperated with a dot '.'
            const char *p_dot = strchr((const char *)field_ptr, (int)'.');
            if (p_dot == (const char *)NULL)
            {
                // if this is a notification request - and the request is ALL -
                // must take it as a string literal otherwise assume it is a 0x register
                // value.
                if (strcmp((const char *) field_ptr, fcp_all_notifications) == 0)
                {
                    strcpy((char *)gige_reg_label, (const char *)field_ptr);
                    // no register value just set it to zero value
                    strcpy((char *)gige_reg_value, (const char *)"0x0000");
                }
                else
                {
                    // just use the register value for this
                    //sprintf(gige_reg_label, "%s%d", "field",field_index);
                    // when send over register values without a label use same register value for label in JSON
                    strcpy((char *)gige_reg_value, (const char *)field_ptr);
                    strncpy((char *)gige_reg_label, (const char *)&gige_reg_value[0], 6);
                    gige_reg_label[6] = 0;
                }
                // unexpected key found for loading request
                //p_request->pkgstatus = FCP_PKGSTATUS_PARSE_BAD_FORMAT;
            }
            else
            {
                // can pull out the label and the register and convert the register to the unsigned 32bit value needed
                strcpy((char *)gige_reg_label, (const char *)field_ptr);
                // null terminate at the dot.
                char *put_null = (char *) strchr((const char *) gige_reg_label, '.');
                if (put_null != (char *) NULL)
                    *put_null = '\0';
                // should add check for the 0x for the register value
                p_dot++;
                if (*p_dot == '0' && (*(p_dot + 1) == 'x'))
                    strcpy((char *) gige_reg_value, (char *) p_dot);
                else
                {
		    lwsl_debug("%s: Bad register specified [%s]!\n", __FUNCTION__, gige_reg_label);
                    p_request->pkgstatus = FCP_PKGSTATUS_PARSE_BAD_FCP_REG;
                }
            }
        }
        else
	{
	    lwsl_debug("%s: field_ptr was NULL!\n", __FUNCTION__);
            p_request->pkgstatus = FCP_PKGSTATUS_PARSE_FIELDS_FAILED;
	}

        // if package status is still good populate the request data with this field
        if (p_request->pkgstatus == FCP_PKGSTATUS_OK)
        {
            uint32_t cs_reg = strtoul((char *)gige_reg_value, NULL, 16);

            // is this field a memory field or a register field
            uint16_t nbytes_in_fcp_mem = 0;
            bool is_fcp_mem_req = false;
            bool response_needs_encoding = false;

            // check for case where this is a request to load a memory request
            is_cs_reg_mem(cs_reg, &is_fcp_mem_req, &nbytes_in_fcp_mem, &response_needs_encoding);

            // add this field to the request record.
            if (add_field_to_request(p_request, gige_reg_label, cs_reg, is_fcp_mem_req, nbytes_in_fcp_mem, response_needs_encoding))
            {
                lwsl_debug("%s: Received Valid FCP Field: %s [%u]\n", __FUNCTION__, p_request->p_fcp_request_fields[field_index]->cs_reg_name, p_request->p_fcp_request_fields[field_index]->cs_reg);
                // if storing install the request store data into the field that was just created
            }
            else
            {
                p_request->pkgstatus = FCP_PKGSTATUS_BUILD_FAILED;
            }


            // if storing take care of the data requested to store
            if ((p_request->action_type == FCP_REQUEST_ACTION_STORE) && (p_request->pkgstatus == FCP_PKGSTATUS_OK))
            {
                // grab a shortcut to the field pointer to make coding easier
                fcp_request_field *p_field = p_request->p_fcp_request_fields[field_index];

                // the data should have been presented in the request if not the store request is invalid
                if (data_ptr == (char *) NULL)
                    p_request->pkgstatus = FCP_PKGSTATUS_PARSE_BAD_FCP_DATA;
                else
                {
                    // the data should can be a memory array or it can be a register value
                    if (p_field->is_register_an_array)
                    {
                        // is the data being stored b64 encoded?
                        if (strncmp(data_ptr, "b64_", 4) == 0)
                        {
                            // need to verify that the data presented does not exceed the size of the field
                            // if so warn the user that the store package exceeded the maximun size allowed
                            char gige_buffer[MAX_FCP_MEMORY_BUFFER];
                            // decode the data into the gige memory array prior to adding it to the request field
                            gige_bufsize = lws_b64_decode_string((const char *) &data_ptr[4], gige_buffer, MAX_FCP_MEMORY_BUFFER);

                            //check for the oversize condition
                            if (gige_bufsize > p_field->cs_reg_array_len)
                            {
                                p_request->pkgstatus = FCP_PKGSTATUS_FIELDSIZE_EXCEEDED;
                            }
                            else
                            {
                                // copy over the data into the request record
                                memcpy(p_field->p_cs_reg_array, gige_buffer, gige_bufsize);
                            }
                        }
                        else
                        {
                            // copy over the information from the request JSON
                            memcpy(p_field->p_cs_reg_array, data_ptr, p_field->cs_reg_array_len - 1);
                            // put a zero at the very end
                            p_field->p_cs_reg_array[p_field->cs_reg_array_len] = '\0';
                            gige_bufsize = strlen((const char *) p_field->p_cs_reg_array);
                        }
                        // TODO any type of preprocessing?
                    }
                    else
                    {
                        // storing a 32-bit value if not a memory array convert the string to
                        // the 32 bit value based on whether or not there is a 0x prepended to data
                        if (strncmp(data_ptr, "0x", 2) == 0)
                            p_field->cs_reg_data = strtoul((char *)data_ptr, NULL, 16);
                        else
                            p_field->cs_reg_data = strtoul((char *)data_ptr, NULL, 10);

                        // TODO any type of preprocessing?
                    }
                }
            }
        }

        field_index++;  // set up for next iteration

    }   // end of the field processing while loop

    // at this point whether the request prepared itself properly or not it needs to be
    // sent back to the session caller so mark the session request record as in use
    p_request->state = FCP_REQUEST_STATE_READY;

    // free the json parsed data structure
    free_fi_json_parse_pkg(&my_json_data);

    // if made it here there is a request record that needs to be written back to caller
    return true;
}


/* --------------------------------------------------------------------------------
* @brief Common logic to safely copy over memory for a JSON response buffer
*
* Called from the fcp_handle_response function to build up the
* callers json buffer
*
* @param p_add:     pointer to the data to add to the destination buffer
* @param addsize:   how many bytes to copy to the destination buffer
* @param p_dest:    pointer to the beginning of the destination buffer ready to receive data
* @param destmax:   how many total bytes can the destination buffer hold
* @param dest_used: how many bytes in the destination buffer are already used
*
* @return number of bytes used in the destination buffer - if same as input the
* copy did not take place either the size would exceed the buffer or the pointers
* were null
* --------------------------------------------------------------------------------*/
static uint16_t copy_json_mem(uint8_t *p_add, uint16_t addsize, uint8_t *p_dest, uint16_t destmax, uint16_t dest_used)
{
    uint16_t result = dest_used;    // assume same in use
    // if pointers are valid add the data if room
    if ((p_add != (uint8_t *)NULL) && (p_dest != (uint8_t *)NULL))
    {
        if ((dest_used + addsize) < destmax)
        {
            // should be able to add the data
            memcpy((void *)&p_dest[dest_used], (void *) p_add, addsize);
            result += addsize;  // increment the count of bytes used in the destination
        }
    }
    return result;  // new destination used
}

/* --------------------------------------------------------------------------------
* @brief Common logic store a register value or buffer from camera for a field request
*
* Called from the fcp_handle_response function to build up the
* callers json buffer as it processes each field in the request list
*
* @param p_buffer:      pointer to a buffer will receive status of store operation
* @param max_bufsize:   how many bytes can be stored in the result buffer
* @param p_field:       pointer to the field requested for storing and it contains all of the request information
*
* @return package status code
* --------------------------------------------------------------------------------*/
static int store_cs_field(uint8_t *p_buffer, uint16_t /*max_bufsize*/, fcp_request_field *p_field)
{
    int store_result =  FCP_PKGSTATUS_OK;   // assume the store will work successfully
    uint32_t cs_reg;

    // get the connection to the camera server for the wss
    fiCameraServerIO *p_fiCS = fiGetCS();
    if (p_fiCS == (fiCameraServerIO *) NULL)
    {
        store_result = FCP_PKGSTATUS_NO_CS_AVAIL;
        return store_result;
    }

    // cs register value in numeric form

    // if pointers are proceed with the load request
    if ((p_buffer != (uint8_t *)NULL) && (p_field != (fcp_request_field *)NULL))
    {
        p_buffer[0] = '\0'; // put a null terminator for the result string

        // pull camera server register information from the request record
        int nbytes_written  = 0;
        cs_reg = p_field->cs_reg;

        // if this is an array request want to read memory from camera server
        if (p_field->is_register_an_array == 1)
        {
            // write array to the Camera Server
            nbytes_written = p_fiCS->writeFCPArray(cs_reg, p_field->p_cs_reg_array, p_field->cs_reg_array_len);
            if (nbytes_written <= 0)
                store_result = FCP_PKGSTATUS_CS_ERROR;   // the Camera Server was unable to write the data
        }
        else
        {
            int istat;  // the set register sets this to a value
            uint32_t cs_result = p_fiCS->writeFCPReg(cs_reg, p_field->cs_reg_data, &istat);
            if (cs_result != 1)
            {
                if (istat == GEV_STATUS_ACCESS_DENIED)
                {
                    // We assume the only reason for this would be loss of
                    // FASTEC_CCP lock. Might not be a good assumption.
                    store_result = FCP_PKGSTATUS_CS_DENIED_ACCESS;
                }
                else
                {
                    // the Camera Server returned no data
                    store_result = FCP_PKGSTATUS_CS_ERROR;
                }
            }
        }
        // if the store was successful mark the output buffer as complete
        if (store_result == FCP_PKGSTATUS_OK)
        {
            // mark the store with a complete string
            strcpy((char *)p_buffer, fcp_pkgstatus_complete);
        }
    }
    else
        store_result = FCP_PKGSTATUS_BUILD_FAILED;   // the pointers are invalid something went wrong with the request package generation

    return store_result;  // return the state of the store
}


/* --------------------------------------------------------------------------------
* @brief Common logic load a register value or buffer from camera for a field request
*
* Called from the fcp_handle_response function to build up the
* callers json buffer as it processes each field in the request list
*
* @param p_buffer:      pointer to the buffer to receive the data for a field array or register
* @param max_bufcnt:    maximun size the buffer can handle
* @param p_field:       pointer to the field requested it contains all of the request information
* @param p_cs_len:      pointer to the result value that came back from the camera server
*
* @return package status code
* --------------------------------------------------------------------------------*/
static int load_cs_field(uint8_t *p_buffer, uint16_t /*max_bufcnt*/, fcp_request_field *p_field, uint16_t *p_cs_len)
{
    int load_result =  FCP_PKGSTATUS_OK;    // assume the load will work successfully
    uint32_t cs_reg;                            // cs register value in numeric form
    int len_of_field_data = 0;

    // get the connection to the camera server for the wss
    fiCameraServerIO *p_fiCS = fiGetCS();
    if (p_fiCS == (fiCameraServerIO *) NULL)
    {
        load_result = FCP_PKGSTATUS_NO_CS_AVAIL;
        return load_result;
    }

    // if pointers are proceed with the load request
    if ((p_buffer != (uint8_t *)NULL) &&
            (p_field != (fcp_request_field *)NULL) && (p_cs_len != (uint16_t *)NULL))
    {
        *p_cs_len = 0;   // set to 0 bytes
        p_buffer[0] = '\0';

        // pull camera server register information from the request record
        cs_reg = p_field->cs_reg;

        // if this is an array request want to read memory from camera server
        if (p_field->is_register_an_array == 1)
        {
            *p_cs_len = p_field->cs_reg_array_len;    // expected length

            // read array from the Camera Server
            *p_cs_len = p_fiCS->readFCPArray(cs_reg, p_buffer, *p_cs_len);
            if (*p_cs_len <= 0)
            {
                load_result = FCP_PKGSTATUS_CS_ERROR;   // the Camera Server returned no data
            }
            else
            {
                //handle encoding the data if necessary
                if (p_field->encode_response)
                {
                    char csEncoded[(MAX_FCP_MEMORY_BUFFER * 2)]; // base 64 encoded
                    char base64_str[] = "b64_";
                    uint16_t start_encoding_index = strlen(base64_str);
                    uint16_t nbytes_left = (MAX_FCP_MEMORY_BUFFER * 2) - start_encoding_index;

                    // prepend the base 64 encoded string
                    strcpy(csEncoded, base64_str);

                    // encode into a temporary buffer
                    int nbytes_encoded = lws_b64_encode_string((const char *)p_buffer, *p_cs_len, (char *) &csEncoded[start_encoding_index], (int)nbytes_left);

                    // if the encoding succeeded add it to the output buffer
                    if (nbytes_encoded > 0)
                    {
                        // MAKE SURE IT CAN TAKE THE TOTAL BUFFER SIZE TODO max_bufcnt
                        nbytes_encoded += start_encoding_index; // add in the base 64 identifying string length
                        memcpy(p_buffer, csEncoded, nbytes_encoded);
                        p_buffer[nbytes_encoded] = '\0';
                        *p_cs_len = nbytes_encoded; // update the return value for the size of the buffer
                    }
                    else
                    {
                        // encoding failed -
                        load_result = FCP_PKGSTATUS_B64_ENCODE_FAILED;
                    }
                }
            }
        }
        else
        {
            uint32_t cs_val = 0;

            // read the FCP register
            int cs_result = p_fiCS->readFCPReg(cs_reg, &cs_val);

            // if all is good continue...
            if (cs_result == 1)
            {
                // good request put info in field data
                sprintf((char *)p_buffer, "%u", cs_val);
                len_of_field_data = strlen((const char *)p_buffer);
                *p_cs_len = (uint16_t) len_of_field_data;
            }
            else
            {
                load_result = FCP_PKGSTATUS_CS_ERROR;   // the Camera Server returned no data
            }
        }
    }
    else
        load_result = FCP_PKGSTATUS_BUILD_FAILED;   // the pointers are invalid something went wrong with the request package generation

    return load_result;  // return the state of the load
}

/*!
 * @brief Utility function to get the Fastec FCP lock.
 */
static bool get_lock(void)
{
    int lock_attempts = 0;
    int status = 0; // hold the result of the CCP on and off requests
    bool locked;
    struct timespec sleep_req;

    locked = false;
    while (! locked)
    {
        if (cs_set_reg_l(GIGE_REG_FASTEC_CCP, GIGE_FASTEC_CCP_CA_BE, &status) != 1)
        {
            lock_attempts++;
            fi_log(LOG_INFO, "%s: cs_set_reg_l(GIGE_REG_FASTEC_CCP, GIGE_FASTEC_CCP_CA_BE) attempt %d fails: %d\n",
                   __FUNCTION__, lock_attempts, status);

            // Willing to wait for 3.2 seconds--enough for FASTEC_CCP timeout.
            if (lock_attempts >= 32)
                break;

            // Sleep for 100 milliseconds.
            sleep_req.tv_sec = 0;
            sleep_req.tv_nsec = 100000000;
            nanosleep(&sleep_req, NULL);
        }
        else
        {
            locked = true;
        }
    }

    // Return our status.
    return locked;
}

/*! --------------------------------------------------------------------------------
* @brief Processes request package and translates into JSON Response Data
*
* Called from the fcp protocol callback handler when time is available to write
* a response to requesting clients -
*
* @param p_request:     The fcp request record to service
* @param p_fcp_buffer:  The fcp protocol has allocated a buffer for the JSON response
* @param fcp_max_bufsize:   The maximun number of bytes that can be in a fcp buffer
*
* @return true if successful parse or false if error occurred
 * --------------------------------------------------------------------------------*/
int fcp_handle_response(fcp_request *p_request, uint8_t *p_fcp_buffer, uint16_t fcp_max_bufsize)
{
    int status = 0; // hold the result of the CCP on and off requests
    bool locked = false;

    int checkpoint_field_index;
    uint16_t checkpoint_json_field_size;

    // how many bytes were populated in the fcp buffer
    int n_fcp_bytes = 0;

    // how many  fields have been added to the json - use this to determine field
    // separator characters that may be needed
    int n_fields_in_json = 0;

    // how many bytes are in use for the data being built here
    uint16_t json_pkg_size = 0;

    // keep track of the package status
    uint16_t pkgstatus;

    // keep track of fields being processed
    uint16_t    field_index = 0;
    char        field_buffer[FCP_MAX_RX_BUFFER];    // make a local buffer to build up field responses
    uint16_t    json_field_size = 0;

    // camera server request variables
    char        tmp_buffer[MAX_FCP_MEMORY_BUFFER * 2];         // make it *2 in case it is encoded in b64 form for bin-transfer
    uint16_t    max_tmpbufsize = MAX_FCP_MEMORY_BUFFER * 2;    // store the maximum size of the memory buffer

    // make sure that the pointers passed are not null
    if ((p_request != (fcp_request *) NULL) && (p_fcp_buffer != (uint8_t *) NULL))
    {
        // set the package status to match the request status
        pkgstatus = p_request->pkgstatus;

        // fill the entire output buffer with zeros
        memset((void *)p_fcp_buffer, 0, fcp_max_bufsize);

        // add the initial JSON data for the response package
        n_fcp_bytes = copy_json_mem((uint8_t *)fcp_response_pkg_start, (uint16_t) strlen(fcp_response_pkg_start),
                                    p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
        // keep building response if was able to write bytes to output buffer
        if (n_fcp_bytes != json_pkg_size)
        {
            json_pkg_size = n_fcp_bytes;   // all ok

            // set up the action information
            if (p_request->action_type == FCP_REQUEST_ACTION_LOAD)
            {
                // add the initial JSON data for the response package
                n_fcp_bytes = copy_json_mem((uint8_t *)fcp_response_action_load, (uint16_t) strlen(fcp_response_action_load),
                                            p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
                snprintf(tmp_buffer, sizeof(tmp_buffer), ",\"sn\":%u", fcp_sn++);

                json_pkg_size = n_fcp_bytes;
                n_fcp_bytes = copy_json_mem((uint8_t *)tmp_buffer, (uint16_t) strlen(tmp_buffer),
                                            p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
            }
            else if (p_request->action_type == FCP_REQUEST_ACTION_STORE)
            {
                // Get the Fastec lock
                locked = get_lock();
                if (! locked)
                {
                    pkgstatus = FCP_PKGSTATUS_CS_LOCK_FAILED;
                }

                // add the initial JSON data for the response package
                n_fcp_bytes = copy_json_mem((uint8_t *)fcp_response_action_store, (uint16_t) strlen(fcp_response_action_store),
                                            p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
            }
            else if (p_request->action_type == FCP_REQUEST_ACTION_NOTIFY)
            {
                // add the initial JSON data for the response package
                n_fcp_bytes = copy_json_mem((uint8_t *)fcp_response_action_notify, (uint16_t) strlen(fcp_response_action_notify),
                                            p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
                snprintf(tmp_buffer, sizeof(tmp_buffer), ",\"sn\":%u", fcp_sn++);

                json_pkg_size = n_fcp_bytes;
                n_fcp_bytes = copy_json_mem((uint8_t *)tmp_buffer, (uint16_t) strlen(tmp_buffer),
                                            p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
            }
            else
            {
                // add the initial JSON data for the response package
                n_fcp_bytes = copy_json_mem((uint8_t *)fcp_response_action_bad, (uint16_t) strlen(fcp_response_action_bad),
                                            p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
            }
        }

        // continue building if writing is still going well
        if (n_fcp_bytes != json_pkg_size)
        {
            json_pkg_size = n_fcp_bytes;

            // add the field separater after the action key value
            n_fcp_bytes = copy_json_mem((uint8_t *)fcp_response_field_sep, (uint16_t) strlen(fcp_response_field_sep),
                                        p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
        }

        // at this point can't install package status because still more processing of request
        // so need to build up a local buffer to hold all of the field data which can be added
        // after we determine field request status
        if (n_fcp_bytes != json_pkg_size)
        {

            // the first part of the fields is the key introducing the array of field records
            // install the fields keystring for the response field array
            sprintf(&field_buffer[json_field_size], "%s%s%s",
                    fcp_response_pkg_field_list,
                    fcp_response_field_delimeter,
                    fcp_response_field_list_start);

            json_field_size = strlen(field_buffer); // catch up size

            // process the request field(s) and build up the response field(s) data
            // all errors should be reported in the output
            for (field_index = 0; (field_index < p_request->nfields); field_index++)
            {
                int field_pkgstatus;  // keep short-term result status for field response value support
                uint16_t cs_len;     // how big is the camera server data

                // grab a shortcut to the field pointer to make coding easier
                fcp_request_field *p_field = p_request->p_fcp_request_fields[field_index];

                // if notification mode - only process if the data is marked to be sent
                if (p_request->action_type == FCP_REQUEST_ACTION_NOTIFY)
                {
                    if (p_field->is_this_notification != 1)
                        continue;   // skip this field!
                    else
                        p_field->is_this_notification = 0;  // reset it's being processed for writing right now.
                }

                // otherwise process the field always for a load/store

                // Record state, in case we need to roll-back and re-try
                checkpoint_field_index = field_index - 1;
                checkpoint_json_field_size = json_field_size;

                // if a field has already been processed put a comma before it in the buffer
                if (n_fields_in_json > 0)
                    sprintf(&field_buffer[json_field_size], "%s", fcp_response_field_sep);

                // update the size of the field buffer
                json_field_size = strlen(field_buffer);

                // set up the field buffer with the first part of the response array item each field request should have response
                sprintf(&field_buffer[json_field_size], "%s%s%s%s%s%s",
                        fcp_response_pkg_start,
                        fcp_response_field_keyquote,
                        p_field->cs_reg_name,
                        fcp_response_field_keyquote,
                        fcp_response_field_delimeter,
                        fcp_response_field_keyquote);

                // update the size of the field buffer
                json_field_size = strlen(field_buffer);

                // only try to load or store the data if the overall package status is ok
                if (pkgstatus == FCP_PKGSTATUS_OK)
                {
                    // if loading call the load logic
                    if (p_request->action_type == FCP_REQUEST_ACTION_LOAD)
                        field_pkgstatus = load_cs_field((uint8_t *)tmp_buffer, max_tmpbufsize, p_field, &cs_len);
                    else if (p_request->action_type == FCP_REQUEST_ACTION_STORE)
                    {
                        field_pkgstatus = store_cs_field((uint8_t *)tmp_buffer, max_tmpbufsize, p_field);
                        cs_len = (uint16_t) strlen((const char *)tmp_buffer);
                    }
                    else if (p_request->action_type == FCP_REQUEST_ACTION_NOTIFY)
                    {
                        // good request put info in field data
                        sprintf((char *)tmp_buffer, "%u", p_field->cs_reg_data);
                        cs_len = (uint16_t) strlen((const char *)tmp_buffer);
                        field_pkgstatus = p_request->pkgstatus;
                    }
                    else
                        // action type is not recognized
                        field_pkgstatus = FCP_PKGSTATUS_PARSE_ACTION_FAILED;
                }
                else
                    field_pkgstatus = FCP_PKGSTATUS_SKIPPED;    // indicate that we did not attempt this load/store

                // if the field package status is not ok - put a value in the output for the field package status error
                if (field_pkgstatus != FCP_PKGSTATUS_OK)
                {
                    if ((p_request->action_type == FCP_REQUEST_ACTION_STORE) && (field_pkgstatus == FCP_PKGSTATUS_CS_DENIED_ACCESS))
                    {
                        // (Apparently) lost our FASTEC_CCP lock. Get it back and soldier on....
                        // Get the Fastec lock
                        locked = get_lock();
                        if (locked)
                        {
                            // Got it back; try to pick up where we left off
                            field_index = checkpoint_field_index;
                            json_field_size = checkpoint_json_field_size;
                            continue;
                        }
                        else
                        {
                            pkgstatus = FCP_PKGSTATUS_CS_LOCK_FAILED;
                        }
                    }

                    // translate package status to the result and will put into the output
                    char *pdata = xlate_pkgstatus_to_str((int)field_pkgstatus);
                    strcpy(tmp_buffer, pdata);
                    cs_len = strlen(tmp_buffer);

                    // if the overall package is not in error mark it with this error
                    if (pkgstatus == FCP_PKGSTATUS_OK)
                        pkgstatus = field_pkgstatus;
                }

                // without regard to any errors populate the field data with state

                // if there is enough space left in field buffer add the data
                if ((json_field_size + cs_len) >= fcp_max_bufsize)
                {
                    pkgstatus = FCP_PKGSTATUS_FIELDSIZE_EXCEEDED; // not enough space for this field request!
                    sprintf(&field_buffer[json_field_size], "%s", fcp_pkgstatus_exceeds_size);

                    // update size of the field buffer
                    json_field_size = strlen(field_buffer);
                }
                else
                {
                    strcat((char *) &field_buffer[json_field_size], (const char *) tmp_buffer);
                    json_field_size = strlen(field_buffer);
                }

                // terminate this field information in the response package

                // add the closing data for the field in the JSON package
                sprintf(&field_buffer[json_field_size], "%s%s",
                        fcp_response_field_keyquote,
                        fcp_response_pkg_end);

                // update size of the field buffer
                json_field_size = strlen(field_buffer);

                // increment the number of fields that have been added to the JSON output
                n_fields_in_json++;

            } // end of the field processing loop

            // install the fields keystring for the response field array
            sprintf(&field_buffer[json_field_size], "%s", fcp_response_field_list_end);

            // update the size of the field buffer
            json_field_size = strlen(field_buffer);
        }

        // continue building if writing is still going well
        if (n_fcp_bytes != json_pkg_size)
        {
            // sync the json_pkg_size to be the start of the pkginfo key in the response package
            uint16_t save_pkgstatus_place = json_pkg_size = n_fcp_bytes;

            // want a way to loop back into this logic in case the fields
            // cannot be added to the json response package
            bool try_again = true;

            // keep state of whether or not the field buffer was written
            bool add_fields_failure = false;

            // want to have ability to loop again in case error happens when trying to add
            // the field buffer to the output - exceeding json package size is the error
            // that can occur, if so want to catch the error and rewrite the pkgstatus
            while (try_again)
            {
                // create the package status information JSON
                char cs_temp[256];
                get_pkg_status_msg(pkgstatus, p_request->json_return_code, cs_temp);

                // the next thing to add to the package status, fields and closing package delimeter
                sprintf(tmp_buffer, "%s%s%s%s",
                        fcp_response_pkgstatus,
                        fcp_response_field_delimeter,
                        cs_temp,
                        fcp_response_field_sep);

                // add the package status data to the output buffer
                n_fcp_bytes = copy_json_mem((uint8_t *)tmp_buffer, (uint16_t) strlen(tmp_buffer),
                                            p_fcp_buffer, fcp_max_bufsize, json_pkg_size);

                // attempt to add the field buffer
                if (!add_fields_failure)
                {
                    json_pkg_size = n_fcp_bytes;
                    // now try to install the field buffer
                    n_fcp_bytes = copy_json_mem((uint8_t *)field_buffer, (uint16_t) json_field_size,
                                                p_fcp_buffer, fcp_max_bufsize, json_pkg_size);

                    // if not enough room in json output buffer too much data was requested
                    if (n_fcp_bytes <= json_pkg_size)
                    {
                        pkgstatus = FCP_PKGSTATUS_FIELDSIZE_EXCEEDED;   // not enough space for this fields request!
                        add_fields_failure = true;                          // mark the add failure condition
                        json_pkg_size = save_pkgstatus_place;               // go back to write field size exceeded error
                    }
                    else
                    {
                        try_again = false;              // all is good no need to try again!
                        json_pkg_size = n_fcp_bytes;   // update the package size to accomodate fields added
                    }
                }
                else
                    try_again = false;  // only loop one more time to report the add fields failure in package status
            }  // end of the while loop

            // close off the data with the end of the package
            n_fcp_bytes = copy_json_mem((uint8_t *)fcp_response_pkg_end, (uint16_t) strlen(fcp_response_pkg_end),
                                        p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
            json_pkg_size = n_fcp_bytes;
        }
    }

    if (locked)
    {
        cs_set_reg_l(GIGE_REG_FASTEC_CCP, 0, &status);
        locked = false;
    }
    return n_fcp_bytes;    // this holds the number of bytes written to the output file and to send to requester
}


/* --------------------------------------------------------------------------------
* @brief Builds Notification Response Buffer for ALL notifications in session data
*
* Called from the fcp-listener-protocol callback handler when time is available to write
* a response to requesting clients - for case of ALL notifications versus specific requested
* registers
*
* @param pss:               The fcp-listener protocol session data pointer
* @param p_fcp_buffer:     The fcp protocol has allocated a buffer for the JSON response
* @param fcp_max_bufsize:  The maximun number of bytes that can be in a fcp buffer
*
* @return the number of bytes that were written to the fcp response buffer
 * --------------------------------------------------------------------------------*/
int
fcp_bld_all_notifications(per_session_data_fcp_listener *pss, uint8_t *p_fcp_buffer, uint16_t fcp_max_bufsize)
{
    // how many bytes were populated in the fcp buffer
    int n_fcp_bytes = 0;

    // how many bytes are in use for the data being built here
    uint16_t json_pkg_size = 0;

    // keep track of the package status
    uint16_t pkgstatus;

    // keep track of register being processed
    char        field_buffer[FCP_MAX_RX_BUFFER * 2];  // make a local buffer to build up field responses
    uint16_t    json_field_size = 0;

    char        tmp_buffer[MAX_FCP_MEMORY_BUFFER]; // Intermediate buffer

    // make sure that the pointers passed are not null
    if ((pss != (per_session_data_fcp_listener *) NULL) && (p_fcp_buffer != (uint8_t *) NULL))
    {
        // set the package status in this case to be ok as this is
        // an ALL notification stream
        pkgstatus = FCP_PKGSTATUS_OK;

        // fill the entire output buffer with zeros
        memset((void *)p_fcp_buffer, 0, fcp_max_bufsize);

        // add the initial JSON data for the response package
        json_pkg_size = n_fcp_bytes = copy_json_mem((uint8_t *)fcp_response_pkg_start, (uint16_t) strlen(fcp_response_pkg_start),
                                      p_fcp_buffer, fcp_max_bufsize, json_pkg_size);

        // action must match notify
        if (pss->request.action_type == FCP_REQUEST_ACTION_NOTIFY)
        {
            // add the initial JSON data for the response package
            n_fcp_bytes = copy_json_mem((uint8_t *)fcp_response_action_notify, (uint16_t) strlen(fcp_response_action_notify),
                                        p_fcp_buffer, fcp_max_bufsize, json_pkg_size);

            snprintf(tmp_buffer, sizeof(tmp_buffer), ",\"sn\":%u", fcp_sn++);

            json_pkg_size = n_fcp_bytes;
            n_fcp_bytes = copy_json_mem((uint8_t *)tmp_buffer, (uint16_t) strlen(tmp_buffer),
                                        p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
        }
        else
        {
            // add the initial JSON data for the response package
            n_fcp_bytes = copy_json_mem((uint8_t *)fcp_response_action_bad, (uint16_t) strlen(fcp_response_action_bad),
                                        p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
        }

        // continue building if writing is still going well
        if (n_fcp_bytes != json_pkg_size)
        {
            json_pkg_size = n_fcp_bytes;

            // add the field separater after the action key value
            n_fcp_bytes = copy_json_mem((uint8_t *)fcp_response_field_sep, (uint16_t) strlen(fcp_response_field_sep),
                                        p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
        }

        // add the package status from the request to the output response package
        if (n_fcp_bytes != json_pkg_size)
        {
            char cs_temp[256];

            json_pkg_size = n_fcp_bytes;

            // create the package status information JSON
            get_pkg_status_msg(pkgstatus, pss->request.json_return_code, cs_temp);


            // the next thing to add to the package status, fields and closing package delimeter
            sprintf(&field_buffer[0], "%s%s%s%s",
                    fcp_response_pkgstatus,
                    fcp_response_field_delimeter,
                    cs_temp,
                    fcp_response_field_sep);

            // add the package status to the fcp buffer
            n_fcp_bytes = copy_json_mem((uint8_t *)field_buffer, (uint16_t) strlen(field_buffer),
                                        p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
        }

        // install the field information
        if (n_fcp_bytes != json_pkg_size)
        {
            int x;
            int n_fields_found = 0;

            // the first part of the fields is the key introducing the array of field records
            // install the fields keystring for the response field array
            sprintf(&field_buffer[json_field_size], "%s%s%s",
                    fcp_response_pkg_field_list,
                    fcp_response_field_delimeter,
                    fcp_response_field_list_start);

            json_field_size = strlen(field_buffer); // catch up size

            // process the registers that are ready to be output to the notification buffer
            for (x = 0; x < MAX_FI_PENDING_NOTIFICATIONS; x++)
            {
                if (pss->all_async_data[x].notify_state == FI_ASYNC_NOTIFY_CS_RECEIVED)
                {
                    // this register is ready to be put into the notification buffer

                    // if a field has already been processed put a comma before it in the buffer
                    if (n_fields_found > 0)
                        sprintf(&field_buffer[json_field_size], "%s", fcp_response_field_sep);

                    // update the size of the field buffer
                    json_field_size = strlen(field_buffer);

                    // set up the field buffer with the first part
                    // of the response array.
                    sprintf(&field_buffer[json_field_size], "%s%s0x%X%s%s%s%u%s%s",
                            fcp_response_pkg_start,
                            fcp_response_field_keyquote,
                            pss->all_async_data[x].cs_reg,
                            fcp_response_field_keyquote,
                            fcp_response_field_delimeter,
                            fcp_response_field_keyquote,
                            pss->all_async_data[x].cs_data,
                            fcp_response_field_keyquote,
                            fcp_response_pkg_end);

                    // update size of the field buffer
                    json_field_size = strlen(field_buffer);

                    // increment the number of fields that have been added to the JSON output
                    n_fields_found++;

                    pss->all_async_data[x].notify_state = FI_ASYNC_NOTIFY_BUILT_FOR_WRITE;

                } // end of the field processing for this notification register
            } // for loop checking through the ALL notification queue for writable notifications

            // install the fields keystring for the response field array
            sprintf(&field_buffer[json_field_size], "%s", fcp_response_field_list_end);

            // update the size of the field buffer
            json_field_size = strlen(field_buffer);
        }

        // continue building if writing is still going well
        if (n_fcp_bytes != json_pkg_size)
        {
            json_pkg_size = n_fcp_bytes;
            // now try to install the field buffer
            n_fcp_bytes = copy_json_mem((uint8_t *)field_buffer, (uint16_t) json_field_size,
                                        p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
        }

        if (n_fcp_bytes != json_pkg_size)
        {
            json_pkg_size = n_fcp_bytes;
            n_fcp_bytes = copy_json_mem((uint8_t *)fcp_response_pkg_end, (uint16_t) strlen(fcp_response_pkg_end),
                                        p_fcp_buffer, fcp_max_bufsize, json_pkg_size);
        }
    }
    return n_fcp_bytes;    // this holds the number of bytes written to the output file and to send to requester
}


/* --------------------------------------------------------------------------------
* @brief CS Notification Callback
*
* Asynchronous callback from Camera Server that is invoked when a register
* has changed -
*
* @param reg    The register which changed.
* @param val    The new value of the register.
*
* @return void
* --------------------------------------------------------------------------------*/
void cs_fcp_reg_notify(void *arg, uint32_t reg, uint32_t val)
{
    (void)arg; // unused

    int x;

    // dont' want to service the GIGE_REG_FASTEC_CCP register return if that is what is being set
    if ((reg == GIGE_REG_FASTEC_CCP)       ||
            (reg == GIGE_REG_JPEG_FRAME_ADDR)  ||
            (reg == GIGE_REG_JPEG_FRAME_LEN))
        return;

    lwsl_debug("%s: Received FCP Notification: 0x%08X -> reg 0x%08X", __FUNCTION__, val, reg);

    // go through the possible entries for fcp listeners and see if any of them
    // are registered to take this notification
    fcp_listeners_lock();
    for (x = 0; x < MAX_FCP_LISTENERS; x++)
    {
        bool is_a_listener = false;

        per_session_data_fcp_listener *pss = all_fcp_listeners[x];

        if (pss != (per_session_data_fcp_listener *) NULL)
            is_a_listener = (pss->listening == 1) ? true : false;

        // if not listening just continue
        if (!is_a_listener)
            continue;

        // if made it here than the session data is listening
        fcp_request *p_request = &pss->request;

        // if this listener is catching ALL notifications put the notification into
        // the notification list
        if (pss->catch_all_notifications == 1)
        {
            // need to find the location to add to the notification list
            int pos = -1;
            if (pss->async_pos < MAX_FI_PENDING_NOTIFICATIONS)
                pos = pss->async_pos;
            else
            {
                pss->async_pos = 0; // reset (should not happen)
                pos = 0;
            }

            pss->all_async_data[pos].cs_reg = reg;
            pss->all_async_data[pos].cs_data = val;
            pss->all_async_data[pos].notify_state = FI_ASYNC_NOTIFY_CS_RECEIVED;

            lwsl_debug("%s: CS ALL Notification SET for reg=%x val=%x async_pos=%d",
                       __FUNCTION__, reg, val, pss->async_pos);

            // increment for the next position
            pss->async_pos++;
            if (pss->async_pos == MAX_FI_PENDING_NOTIFICATIONS)
                pss->async_pos = 0;     // wrap around.

            //TODO - should check position 0 if wrapped around to make sure it is available
        }
        else
        {
            int field_index;    // field index to walk list of fields
            bool set_data = false;

            // if not catching all notification need to see if the listener
            // is waiting for notifications for this register
            for (field_index = 0; !set_data && (field_index < p_request->nfields); field_index++)
            {
                // grab a shortcut to the field pointer to make coding easier
                fcp_request_field *p_field = p_request->p_fcp_request_fields[field_index];

                // if there is a field request record see if there is a register match
                if (p_field != (fcp_request_field *) NULL)
                {
                    // register match? - if so prepare the notification data in the field
                    if (p_field->cs_reg == reg)
                    {
                        set_data = true;

                        lwsl_debug("%s: CS Notification SET for regname=%s reg=0x%x val=%d pkg_status=%d",
                                   __FUNCTION__, p_field->cs_reg_name, reg, val, p_request->pkgstatus);

                        // listener wants to know about this register change so install the new
                        // data into the request field inforamtion and mark it pending for write
                        p_field->cs_reg = reg;  // in case of the all state need to store register
                        p_field->cs_reg_data = val;
                        p_field->is_this_notification = 1;  // mark the field data as a notification
                        p_request->pkgstatus = FCP_PKGSTATUS_OK;
                    }
                }
            }
        }
    }
    fcp_listeners_unlock();
}


/* --------------------------------------------------------------------------------
* @brief Initialize the contents of fi fcp listener session data
*
* Called when the session for a registered FCP listener has been established
*
* @return 1 if successful or 0 if an error occurred
 * --------------------------------------------------------------------------------*/
int fcp_init_listener(per_session_data_fcp_listener *pss)
{
    int status = 0;
    int x;  // iterator variable

    lwsl_notice("%s: Initializing fcp-listener-protocol session data %p",
                __FUNCTION__,
                pss);

    if (pss != (per_session_data_fcp_listener *) NULL)
    {
        pss->listening = 0;
        pss->catch_all_notifications = 0;
        pss->async_pos = 0;

        // set up the all pending notification queue to empty state
        for (x = 0; x < MAX_FI_PENDING_NOTIFICATIONS; x++)
        {
            pss->all_async_data[x].cs_data = 0;
            pss->all_async_data[x].cs_reg = 0;
            pss->all_async_data[x].notify_state = FI_ASYNC_NOTIFY_AVAILABLE;
        }

        // capture the camara server pointer and have it available for
        // the callback routine handlers.
        pss->p_fiCS = fiGetCS();

        // initialize the fcp request record to empty state
        if ((pss->p_fiCS != (fiCameraServerIO *) NULL) && init_request_rec(&pss->request))
            status = 1;
    }

    lwsl_info("(%s) Initialized fcp-listener-protocol session data status=%d", __FUNCTION__, status);
    return status;
}

/* --------------------------------------------------------------------------------
* @brief Reset the contents of an fi fcp listener session data
*
* Called when the session for a registered FCP listener has been established and
* being reset for a different request
*
* @return 1 if successful or 0 if an error occurred
 * --------------------------------------------------------------------------------*/
int fcp_reset_listener(per_session_data_fcp_listener *pss)
{
    int status = 0;

    lwsl_info("(%s) Resetting fcp-listener-protocol for new request", __FUNCTION__);

    // don't want any segment faults make sure the pointer is valid
    if (pss != (per_session_data_fcp_listener *) NULL)
    {
        // free up any allocated resources that may be set in previous requests
        free_field_data_for_request(&pss->request);

        // now simply call the initialization procedure
        status = fcp_init_listener(pss);
    }

    lwsl_info("(%s) Initialized fcp listener pss returning %d", __FUNCTION__, status);

    return status;
}

/* --------------------------------------------------------------------------------
* @brief Determines if session data is already in listening queue
*
* Called when the fcp-listener-protocol receives a notification request - there
* is only one active notification request per session data and that session data is
* only allowed one slot in the listening queue - this will return the status of
* the session data in the listening queue
*
* @return 1 if session data is already in listening queue or 0 if it is not found
 * --------------------------------------------------------------------------------*/
int is_pss_in_fcp_listening_queue(per_session_data_fcp_listener *pss)
{
    int x;              // iterator variable

    lwsl_info("(%s) Looking for session data in listening queue", __FUNCTION__);

    // was the fcp listener collection initialized?
    if (fcp_listeners_initialized != 1)
    {
        lwsl_info("(%s) Error the fcp listener queue is not initialized", __FUNCTION__);
        return 0;   // this should be an error - TODO
    }

    if (pss == (per_session_data_fcp_listener *) NULL)
    {
        lwsl_info("(%s) Error session data pointer is null", __FUNCTION__);
        return 0;
    }

    // go through the possible entries for fcp listeners and locate
    // the next available position for this information
    for (x = 0; x < MAX_FCP_LISTENERS; x++)
    {
        per_session_data_fcp_listener *temp_pss = all_fcp_listeners[x];

        if (temp_pss == pss)
        {

            lwsl_info("(%s) Found fcp-listening protocol session in index=%d", __FUNCTION__, x);
            return 1;
        }
    }
    return 0;   // pss is not in the queue already
}

/* --------------------------------------------------------------------------------
* @brief Install a requested FCP Listener Session
*
* Called when the session for a registered FCP listener has been created this manages
* the list of all session listeners in the WSS
*
* @return 1 if successful or 0 if an error occurred
 * --------------------------------------------------------------------------------*/
int fcp_add_listener(per_session_data_fcp_listener *pss)
{
    int x;

    lwsl_info("(%s) Starting", __FUNCTION__);

    // was the fcp listener collection initialized?
    if (fcp_listeners_initialized != 1)
        return 0;   // this should be an error - TODO

    lwsl_info("(%s) Validating pss", __FUNCTION__);
    if (pss == (per_session_data_fcp_listener *) NULL)
        return 0;

    if (is_pss_in_fcp_listening_queue(pss) == 1)
    {
        lwsl_notice("%s: PSS %p already in all_fcp_listeners[]; not adding again (but not reporting as an error either)",
                    __FUNCTION__,
                    pss);
        return 1;
    }

    // go through the possible entries for fcp listeners and locate
    // the next available position for this information
    for (x = 0; x < MAX_FCP_LISTENERS; x++)
    {
        lwsl_info("(%s) Assigning PSS Pointer for %d", __FUNCTION__, x);
        per_session_data_fcp_listener *temp_pss = all_fcp_listeners[x];

        if (temp_pss == (per_session_data_fcp_listener *)NULL)
        {
            lwsl_debug("%s: Installed pss %p into slot %d", __FUNCTION__, pss, x);

            // this is the available position set it and mark it as a listener
            all_fcp_listeners[x] = pss;
            return 1;   // successfully installed the listener session pointer
        }
    }

    lwsl_err("%s: No space in all_fcp_listeners[] for new pss %p",
             __FUNCTION__,
             pss);

    return 0;   // if make it here there was no available place for the pss
}

/* --------------------------------------------------------------------------------
* @brief Cleanup FCP Listener Session Data
*
* Called when the session for a registered FCP listener is being destroyed and
* this takes care of any cleanup for allocated memory and frees the listening
* position from the list of FCP listeners
*
* @return true if successful or false if an error occurred
 * --------------------------------------------------------------------------------*/
bool fcp_listener_cleanup(per_session_data_fcp_listener *pss)
{
    int x;



    // was the fcp listener collection initialized?
    if (fcp_listeners_initialized != 1)
        return false;   // this should be an error - TODO

    if (pss == (per_session_data_fcp_listener *) NULL)
        return false;

    // go through the possible entries for fcp listeners and initialize
    // the records to the empty state
    for (x = 0; x < MAX_FCP_LISTENERS; x++)
    {
        // is this the session pointer position? it is if the pointers are equal
        if (pss == all_fcp_listeners[x])
        {
            // reset the pointer from list of listeners to null
            all_fcp_listeners[x] = (per_session_data_fcp_listener *) NULL;

            // cleanup and reset the session data even though it's going away
            fcp_reset_listener(pss);

            // all was good and are done

            lwsl_notice("%s: Removed pss %p and cleared slot %d of all_fcp_listeners[]",
                        __FUNCTION__,
                        pss,
                        x);

            return true;
        }
    }

    lwsl_err("%s: Could not find pss %p in all_fcp_listeners[]",
             __FUNCTION__,
             pss);

    return false;   // if got here the session data pointer was not found in the list of listeners
}

/* --------------------------------------------------------------------------------
* @brief Initialize the array of FCP Listeners
*
* Called once from the Web Socket Server to allow the data for the fcp listener
* protocol to initialize itself to an empty state.
*
* @return true if successful parse or false if error occurred
 * --------------------------------------------------------------------------------*/
bool
fi_init_fcp_listeners()
{
    bool result = true;
    int x;

    // if this has already been initialized don't do it again
    if (fcp_listeners_initialized == 1)
        return false;   // this should be an error - TODO

    // Initialize our mutex - used to serialize access to fcp_listeners data
    pthread_mutex_init(&fcp_listeners_mutex, NULL);

    // set all available positions for listeners to null pointers
    for (x = 0; x < MAX_FCP_LISTENERS; x++)
    {
        all_fcp_listeners[x] = (per_session_data_fcp_listener *) NULL;
    }

    // get the connection to the camera server for the wss
    fiCameraServerIO *p_fiCS = fiGetCS();
    if (p_fiCS != (fiCameraServerIO *) NULL)
    {
        // Register for cs change notifications one time during initialization
        p_fiCS->regFCPListener(cs_fcp_reg_notify);
    }

    // if all is ok we are initialized
    if (result)
    {
        fcp_listeners_initialized = 1;
    }
    return result;
}

void fcp_listeners_lock()
{
    pthread_mutex_lock(&fcp_listeners_mutex);
}

void fcp_listeners_unlock()
{
    pthread_mutex_unlock(&fcp_listeners_mutex);
}

// end of fcp_datamgr.cxx
