/*!
 * @file fcp_listen_protocol.cxx
 *
 * @brief all support for managing request/response data for
 * fi-liveview-protocol
 *
 * Copyright 2016 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>

#include "libwebsockets.h"

// this module stores utility functions that were in the
// original open source code from libwebsockets testserver
// useful for debugging etc.
#include "fi_wss_utils.h"


// currently the logic in fcp_datamgr is shared between
// the FCP Command and FCP Listener protocols - this info
// should be incorporated into fcp_listen_protocol.cxx and
// fcp_protocol.cxx but there is a lot of effort written in C
// and out of time to incorporate logic and data
#include "fcp_datamgr.h"

// this includes support for the connected camera server
#include "fi_cs.h"

// the header file for this protocol
#include "fcp_listen_protocol.h"

// what is the protocol id name for this handler
static const char  *FCP_LISTEN_PROTOCOL_ID	= "fi-fcp-listener-protocol";

/*! --------------------------------------------------------------------------------
 * @brief Translate Package Status Code into User Readable String
 *
 * translates the package status integer value into a character readable message
 *
 * @param myProtocolRecPtr  the protocol record to use for the fcp protocol
 *
 */
fiWSSFCPListenProtocol::fiWSSFCPListenProtocol(  struct lws_protocols* myProtocolRecPtr )
{
    // not initialized yet
    IAmInitialized = false;

    if( myProtocolRecPtr != (struct lws_protocols*) NULL )
    {
        // setup the protocol data for this handler
        // these are the important components of the protocol record that
        // are specific to the protocol - it is important to note that the
        // rx_buffer_size is the size that defines the communication package
        // exchanged between the clients and the wss - it's how big the buffer
        // needs to be to support the payloads expected for this protocol
        myProtocolRecPtr->name = FCP_LISTEN_PROTOCOL_ID;
        myProtocolRecPtr->callback = callback_fcp_listener;
        myProtocolRecPtr->per_session_data_size = sizeof(per_session_data_fcp_listener);
        myProtocolRecPtr->rx_buffer_size = FCP_LISTENER_MAX_RX_BUFFER;

        // initialize the listener queues
        if( fi_init_fcp_listeners() )
            IAmInitialized = true;
    }
}

/*! --------------------------------------------------------------------------------
 * @brief Destructs the FI FCP Protocol Handler Object
 */
fiWSSFCPListenProtocol::~fiWSSFCPListenProtocol()
{
    IAmInitialized = false;
}

/*! --------------------------------------------------------------------------------
 * @brief Is the Liveview Command Protocol Handler ready for use?
 *
 * determines if the object is in a state ready for use in web socket server
 *
 * @return true if ready or false if not
 */
bool fiWSSFCPListenProtocol::isReady()
{
    if( IAmInitialized )
        return true;	// right now this is only test but may want to add
    return false;		// not ready
}

/*! --------------------------------------------------------------------------------
 * @brief Callback handler for processing FCP Listener WSS Communications
 *
 * the standard callback handler for the fi-fcp-listener-protocol.
 * the user input parameter is the per session data used to manage the FCP
 * listener for this session - all registered listeners will be stored in a single
 * array managed by the fi_datamgr that gets inserted when session established
 * and removed when the session is destroyed
 *
 * @param wsi                   Libwebsockets session pointer
 * @param reason                Libwebsockets callback reason
 * @param user                  User data pointer for protocol
 * @param p_json_pkg            Input data (where applicable)
 * @param len                   Length of input data
 *
 * @return 0 if successfully handled or -1 if an error occurred
 */
int fiWSSFCPListenProtocol::callback_fcp_listener( struct lws *wsi,
        enum lws_callback_reasons reason,
        void *user, void *p_json_pkg, size_t len )
{
    int status;
    int return_value = 0; // Return -1 to indicate error (asks libwebsockets to close session)
    int n, m;
    unsigned char buf[LWS_SEND_BUFFER_PRE_PADDING + FCP_LISTENER_MAX_RX_BUFFER +
                                                  LWS_SEND_BUFFER_POST_PADDING];
    unsigned char *p = &buf[LWS_SEND_BUFFER_PRE_PADDING];

    // this is the maximun size the Response JSON data can fill - anything above this
    // size will cause the response handler to generate an exceeded size error.
    uint16_t max_json_response_bufsize = FCP_LISTENER_MAX_RX_BUFFER;

    // cast and install the session pointer
    per_session_data_fcp_listener *pss = (per_session_data_fcp_listener *)user;

    fcp_listeners_lock();
    switch (reason)
    {

    case LWS_CALLBACK_ESTABLISHED:
        /* this callback is invoked when the session connection is first established */
        /* this reason provides a way for the protocol to initialize this session */

        /* allocate and initialize any data for this session */
        lwsl_debug("%s: LWS_CALLBACK_ESTABLISHED. pss %p",
                   __FUNCTION__,
                   pss );

        // initialize the content of the session data structure now that a client
        // has established a connection
        fcp_init_listener(pss);
        break;

    case LWS_CALLBACK_CLOSED:
        /* cleanup any data from this session */
        lwsl_debug("%s: LWS_CALLBACK_CLOSED. pss %p",
                   __FUNCTION__,
                   pss );

        fcp_listener_cleanup(pss);
        break;

    case LWS_CALLBACK_SERVER_WRITEABLE:
    {
        /* this callback condition is invoked when the server yields a slice of time */
        /* for the protocol to write any pending messages to the session client */

        // was there any Camera Server notifications processed between time slices
        // for this session?  -- if so want to be able to build the notification
        // response to send to the client listening.

        // set the number to write and number written to zero.
        n = m = 0;

        // If pss is an invalid pointer can't continue.
        if (pss == (per_session_data_fcp_listener *)NULL)
        {
            lwsl_err("%s: NULL session pointer.", __PRETTY_FUNCTION__);
            return_value = -1;
            break;
        }

        fcp_request *p_result = &pss->request;

        //lwsl_info("fi-fcp-listener pss=%u\n", (unsigned int) pss);
        //lwsl_info("fi-fcp-listener-protocol IN WRITE TIME SLICE\n");

        // can only process if there is pending request for this session but need to check
        // two cases - if ALL notifications something will be in the notification list marked
        // as received from the cs - if specific field that field will have an indication in
        // it that it is a notification
        if( ( p_result != (fcp_request*)NULL )  )
        {
            bool notification_waiting = false;
            int x;

            // check to see if the registered client wants ALL notifications
            if( pss->catch_all_notifications )
            {
                // walk the list of notications to see if one is waiting to be sent
                for(x=0; !notification_waiting && (x < MAX_FI_PENDING_NOTIFICATIONS); x++ )
                {
                    if( pss->all_async_data[x].notify_state == FI_ASYNC_NOTIFY_CS_RECEIVED )
                        notification_waiting = true;
                }
            }
            else
            {
                for(x=0; !notification_waiting && (x < pss->request.nfields); x++ )
                {
                    fcp_request_field *p_field = pss->request.p_fcp_request_fields[x];
                    if( p_field != (fcp_request_field*) NULL )
                    {
                        // if a field is marked as a notification pull it out in this write
                        if( p_field->is_this_notification == 1 )
                            notification_waiting = true;
                    }
                }
            }

            // if made it here and need to write notifcation data do so
            if( notification_waiting )
            {
                // is there a notification request pending for output to the caller
                lwsl_debug("fi-fcp-listener-protocol PENDING NOTIFICATION BUILDING RESPONSE\n");

                // if field notification use the handle response function to take care of
                // building up the output buffer for writing
                if( !pss->catch_all_notifications )
                {
                    // call the response handler to process the request and build the
                    // response JSON package to write for the caller
                    n = fcp_handle_response( &pss->request, p, max_json_response_bufsize );
                }
                else
                {
                    // special handler to walk the ALL notification data to build the
                    // response JSON package to write for the caller
                    n = fcp_bld_all_notifications( pss, p, max_json_response_bufsize );
                }

                // mark the result in the log
                lwsl_debug("fi-fcp-listener-protocol REQUEST BUILT RETURNED %d bytes to write\n", n);

                // if there is something to write to the websocket do so
                if( n > 0 )
                {
                    lwsl_info("Writing %d bytes to FCP listener Client buffer=%s\n", n, p);
                    m = lws_write(wsi, (unsigned char *)p, n, LWS_WRITE_TEXT);
                    if (m < n)
                    {
                        lwsl_err("ERROR %d writing to di socket\n", n);
                    }
                }
            }

            // if the socket was unable to take all of the data something is wrong
            if( m < n )
            {
                lwsl_err( "Error writing to FCP listener socket (closing session)\n" );
                return_value = -1;
                break;
            }
        }
    }
    break;

    case LWS_CALLBACK_RECEIVE:
    {
        // this is invoked when data comes to the Web Socket
        // for service - a json request package is expected
        status = 0; // Assuming fault here
        bool in_listener_queue = false;

        lwsl_info("fi-fcp-listener-protocol LWS_CALLBACK_RECEIVE\n");

        // make sure the pss is a valid pointer
        if( pss != (per_session_data_fcp_listener *)NULL )
        {
	    char *json_str = (char *)p_json_pkg;

	    // Make sure we null terminate the JSON string.
	    json_str[len] = 0;

            // since the wss client can send multiple recieve requests, once the connection
            // is established the data that is managed by the session needs to reset the
            // notification data to initialized state before parsing the data requested
            // and it will supercede any existing noification requests - if the client wants
            // to have different levels of notifications receiving data at the same time than
            // that client should set up multiple wss connections to this protocol - otherwise
            // it is possible to change the notification request of an established connection

            // determine if the data is already in the listener queue
            if( is_pss_in_fcp_listening_queue(pss) == 1 )
            {
                in_listener_queue = true;    // note that this resides in the listener queue

                lwsl_debug("fi-fcp-listener-protocol REPLACING AN OLD LISTENER REQUEST\n");

                // already in the queue need to reset it to empty state to prepare for a
                // new parse of the request
                status = fcp_reset_listener(pss);

                // report error in the information stream - but keep going ...
                if( status == 0 )
                    lwsl_err("fi-fcp-listener-protocol error while resetting session data");
            }


            // Process the request - pulls either available or empty pointer to the
            // request record passed as first argument to the fcp parser -
            if (fcp_parse_request(&pss->request, p_json_pkg, false))
            {
                // Mark this as a listening session.
                if (pss->request.nfields > 0)
                {
                    // If this field is an ALL notification field, the state
                    // needs to be marked in the pss session data so that
                    // special handling can take place for the asynchronous
                    // notification messages.
                    fcp_request_field *p_field = pss->request.p_fcp_request_fields[0];

                    // If the field pointer is NULL something is wrong.
                    if (p_field == (fcp_request_field *)NULL)
                    {
                        lwsl_err("fi-fcp-listener-protocol check for ALL fields is null pointer");
                        status = 0;
                    }
                    else
                    {
                        // Is this an ALL notification request?
                        if (strncmp((const char *)p_field->cs_reg_name, "ALL", 3) == 0)
                        {
                            pss->catch_all_notifications = 1;   // user wants ALL notifications
                            pss->async_pos = 0;                 // wrap around position for avaialble notification
                        }
                        else
                        {
                            pss->catch_all_notifications = 0;       // not catching ALL notifications
                        }

                        // Add the session data to the listener queue if not already there.
                        if (!in_listener_queue)
                            status = fcp_add_listener(pss);
                        else
                            status = 1;

                        // ...and if all is ok, mark the pss as a listener.
                        if (status == 1)
                            pss->listening = 1;

                        // Success has status == 1 here
                    }
                }
                else
                {
		    // Make sure we null terminate the 
                    lwsl_err("fi-fcp-listener-protocol LWS_CALLBACK_RECEIVE ERROR nfields=%d pkgstatus=%d\n  %s\n",
                             pss->request.nfields,
                             pss->request.pkgstatus,
                             (char*)p_json_pkg );
                    status = 0;
                }

                // mark the result in the log
                lwsl_info("fi-fcp-listener-protocol LWS_CALLBACK_RECEIVE listening=%d status=%d\n", pss->listening, status);
            }
            else
            {
                lwsl_err( "fi-fcp-listener-protocol Could not parse JSON:\n   %s\n", (char*)p_json_pkg );
                status = 0;
            }
        }
        // return the state of this reason
        if( status == 0 )
        {
            lwsl_err("fi-fcp-listener-protocol LWS_CALLBACK_RECEIVE ERROR\n" );
            return_value = -1;
            break;
        }
    }
    break;

    default:
        break;
    }
    fcp_listeners_unlock();

    // ALL is ok
    return return_value;
}

// end of the fcp_listen_protocol module
