/*!
 * @file fcp_protocol.cxx
 * @brief all support for managing request/response data for the FCP protocol
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>

#include "libwebsockets.h"

// utility functions used for wss
#include "fi_wss_utils.h"

// currently the logic in fcp_datamgr is shared between
// the FCP Command and FCP Listener protocols - this info
// should be incorporated into fcp_listen_protocol.cxx and
// fcp_protocol.cxx but there is a lot of effort written in C
// and out of time to incorporate logic and data
#include "fcp_datamgr.h"

// this includes support for the connected camera server
#include "fi_cs.h"

// the header file for this protocol
#include "fcp_protocol.h"

// what is the protocol id name for this handler
static const char  *FCP_PROTOCOL_ID	= "fi-fcp-protocol";

/*! --------------------------------------------------------------------------------
 * @brief Translate Package Status Code into User Readable String
 *
 * translates the package status integer value into a character readable message
 *
 * @param myProtocolRecPtr  the protocol record to use for the FCP protocol
 *
 * @return none
 */
fiWSSFCPProtocol::fiWSSFCPProtocol(struct lws_protocols* myProtocolRecPtr)
{
    // not initialized yet
    IAmInitialized = false;

    if( myProtocolRecPtr != (struct lws_protocols*) NULL )
    {
        // setup the protocol data for this handler
        // these are the important components of the protocol record that
        // are specific to the protocol - it is important to note that the
        // rx_buffer_size is the size that defines the communication package
        // exchanged between the clients and the wss - it's how big the buffer
        // needs to be to support the payloads expected for this protocol
        myProtocolRecPtr->name = FCP_PROTOCOL_ID;
        myProtocolRecPtr->callback = callback_fcp;
        myProtocolRecPtr->per_session_data_size = sizeof(per_session_data_fcp);
        myProtocolRecPtr->rx_buffer_size = FCP_MAX_RX_BUFFER;

        // initialized
        IAmInitialized = true;
    }
}

/*! --------------------------------------------------------------------------------
 * @brief Destructs the FI FCP Protocol Handler Object
 */
fiWSSFCPProtocol::~fiWSSFCPProtocol()
{
    IAmInitialized = false;
}

/*! --------------------------------------------------------------------------------
 * @brief Is the FCP Command Protocol Handler ready for use?
 *
 * determines if the object is in a state ready for use in web socket server
 *
 * @return true if ready or false if not
 */
bool fiWSSFCPProtocol::isReady()
{
    if( IAmInitialized )
        return true;	// right now this is only test but may want to add
    return false;		// not ready
}

/*! --------------------------------------------------------------------------------
 * @brief Callback handler for processing FCP request/response on the protocol
 *
 * the standard callback handler for the FCP protocol note the *user input
 * parameter is the per session data used to manage the FCP protocol and that each
 * session has it's own allocated user data
 *
 * @param wsi:			the opaque wss interface record
 * @param reason:		the callback reason enumeration (why wss calling back)
 * @param user:			the protocol session data record unique to this protocol
 * @param p_json_pkg:	the JSON input package(the input from wss using this protocol reads JSON)
 * @param len:			used for some callback reasons - not used currently in this protocol
 *
 * @return 0 if successfully handled or 1 if an error occurred
 */
int fiWSSFCPProtocol::callback_fcp( struct lws *wsi,
        enum lws_callback_reasons reason,
        void *user, void *p_json_pkg, size_t len)
{
    int status;
    fcp_request *p_fcp_result = (fcp_request*) NULL;

    int n, m;	// used to keep track of bytes written to buffers
    unsigned char buf[LWS_SEND_BUFFER_PRE_PADDING + FCP_MAX_RX_BUFFER +
                                                  LWS_SEND_BUFFER_POST_PADDING];

    // temporary buffer used to manipulate the communication data for FCP Commands
    unsigned char *pFCPIOBuf = &buf[LWS_SEND_BUFFER_PRE_PADDING];

    // this is the maximun size the Response JSON data can fill - anything above this
    // size will cause the response handler to generate an exceeded size error.
    uint16_t max_json_response_bufsize = FCP_MAX_RX_BUFFER;

    // cast and install the session pointer
    per_session_data_fcp *pss = (per_session_data_fcp *)user;

    switch (reason)
    {

    case LWS_CALLBACK_ESTABLISHED:
        lwsl_debug("%s: LWS_CALLBACK_ESTABLISHED pss %p\n",
                   __FUNCTION__,
                   pss );

        /* this callback is invoked when the session connection is first established */
        /* this reason provides a way for the protocol to initialize this session */

        status = init_fcp_session(pss);

        // check the status of the session initialization
        if( status == 0 )
            return -1;  // initialization failed cannot continue.
        break;

    case LWS_CALLBACK_CLOSED:
        lwsl_debug("%s: LWS_CALLBACK_CLOSED pss %p\n",
                   __FUNCTION__,
                   pss );
        /* cleanup any data from this session */
        reset_all_session_requests(pss);
        break;

    case LWS_CALLBACK_SERVER_WRITEABLE:
        //if you  uncomment this with debugging it generates a lot in log
        //lwsl_info("fi-fcp-protocol LWS_CALLBACK_SERVER_WRITEABLE \n");
        /* this callback condition is invoked when the server yields a slice of time */
        /* for the protocol to write any pending messages to the session client */

        // set the number to write and number written to zero.
        n = m = 0;

        // if pss is an invalid pointer can't continue
        if( pss == (per_session_data_fcp*)NULL )
            return -1;  // major problem

        p_fcp_result = fcp_get_next_request(pss);

        // can only process if there is pending request for this session.
        if (p_fcp_result != (fcp_request*)NULL)
        {
            lwsl_info("fcp_protocol PENDING REQUEST BUILDING RESPONSE\n");

            // call the response handler to process the request and build the
            // response JSON package to write for the caller
            n = fcp_handle_response( p_fcp_result, pFCPIOBuf, max_json_response_bufsize );

            // if there is something to write to the websocket do so
            if( n > 0 )
            {
                lwsl_info("Writing %d to FCP Client\n", n);
                m = lws_write(wsi, (unsigned char *)pFCPIOBuf, n, LWS_WRITE_TEXT);
                if (m < n)
                {
                    lwsl_err("ERROR %d writing to di socket\n", n);
                }
            }

            // if we found a request must tell the fi fcp data manager we are done
            // with this request
            fcp_request_sent(pss,p_fcp_result);

            // if the socket was unable to take all of the data something is wrong
            if( m < n )
                return -1;  // socket write failed TODO better handling here?
        }
        break;

    case LWS_CALLBACK_RECEIVE:
        lwsl_info("fi-fcp-protocol LWS_CALLBACK_RECEIVE \n");

        // This is invoked when data comes to the Web Socket
        // for service - a JSON request package is expected.
        status = 0;

        // Make sure the pss is a valid pointer.
        if (pss != (per_session_data_fcp *)NULL)
        {
	    char *json_str = (char *)p_json_pkg;

	    // Make sure we null terminate the JSON string.
	    json_str[len] = 0;

            // process the request - pulls either available or empty pointer to the
            // request record passed as first argument to the fcp parser -
            if (fcp_parse_request(fcp_get_empty_request(pss), p_json_pkg, true))
                status = 1;
        }

        // return the state of this reason
        if (status == 0)
            return -1;  // this receive request failed.
        break;

    /*
    * this just demonstrates how to use the protocol filter. If you won't
    * study and reject connections based on header content, you don't need
    * to handle this callback
    */
    case LWS_CALLBACK_FILTER_PROTOCOL_CONNECTION:
        lwsl_info("fi-fcp-protocol LWS_CALLBACK_FILTER_PROTOCOL_CONNECTION \n");
        dump_handshake_info(wsi);
        /* you could return non-zero here and kill the connection */
        break;

    default:
        break;
    }

    // All is well.
    return 0;
}

// end of the fcp_protocol module
