/*
 * fi_buffer_q.cxx
 *
 *  Created on: Jan 25, 2016
 *      Author: eric
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libwebsockets.h>

#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

#include "fi_streaming_session.h"
#include "fi_buffer_q.h"

//--------------------------------------------------------------------
// Constant, Type and Macro definitions
//--------------------------------------------------------------------

enum BQ_TRACE
{
    BQ_FUNC    	= 0x00000010,
    BQ_DETAIL  	= 0x00000020
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

static uint32_t traceControl =
//        BQ_FUNC		|
        BQ_DETAIL	|
//        TRACE_0;        // Set our trace level to TRACE_0 to trace.
        TRACE_3;        // Set our trace level to TRACE_3 to not trace.


BufferQueue::BufferQueue() :
    bytes_to_read_from_im(0),
    bytes_to_read_from_header(0),
    frame_num(0),
    xfer_base(0),
    xfer_offset(0),
    buf_id(-1),
    fault(false),
    buffer_queue_trace(false),
    producerOwns(NUM_BUFFS),
    consumerCurrent(0),
    producerCurrent(0),
    size(0),
    get_oldest_flag(false),
    txQueueHead(NULL),
    txQueueTail(NULL),
    producer_signal(0),
    producer_signal_arg(0)
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);

    pthread_mutex_init(&lock, NULL);
    pthread_cond_init (&producer_check_ready, NULL);
}

BufferQueue::~BufferQueue()
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);

    pthread_mutex_destroy(&lock);
    pthread_cond_destroy(&producer_check_ready);
}

// Executes only on Consumer Thread
void BufferQueue::finished()
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);

    if ( producer_signal != NULL )
    {
        // Execute callback to signal producer
        pthread_mutex_lock (&lock);

	TP(BQ_DETAIL, "%s: PS C-finished %d (producer owns %d)", __PRETTY_FUNCTION__, consumerCurrent, producerOwns );

        b[consumerCurrent].setProducer();
        producerOwns++;
        consumerCurrent++;
        if (consumerCurrent >= NUM_BUFFS)
        {
            consumerCurrent = 0;
        }

	TP(BQ_DETAIL, "%s: PS consumerCurrent now %d; (consumer owns: %s, producerOwns: %d)",
	   __PRETTY_FUNCTION__, consumerCurrent, (consumerOwnsCurrent() ? "YES" : "NO"), producerOwns);

        pthread_mutex_unlock (&lock);

	// If the producer got a buffer, let's notify the thread.
        if (producerOwns == 1)
        {
	    TP(BQ_DETAIL, "%s: Signal producer via callback", __PRETTY_FUNCTION__ );
            (*producer_signal)( producer_signal_arg );
        }
    }
    else
    {
        // Use built-in BufferQueue signaling
        pthread_mutex_lock (&lock);

	TP(BQ_DETAIL, "%s: C-finished %d (producer owns %d)", __PRETTY_FUNCTION__, consumerCurrent, producerOwns );
        b[consumerCurrent].setProducer();
        producerOwns++;
        if (producerOwns == 1)
        {
	    TP(BQ_DETAIL, "%s: Signal producer via pthread", __PRETTY_FUNCTION__);
            pthread_cond_signal(&producer_check_ready);
        }
        consumerCurrent++;
        if (consumerCurrent >= NUM_BUFFS)
        {
            consumerCurrent = 0;
        }
	TP(BQ_DETAIL, "%s: consumerCurrent now %d; (consumer owns: %s)", 
	   __PRETTY_FUNCTION__, consumerCurrent, (consumerOwnsCurrent() ? "YES" : "NO"));

        pthread_mutex_unlock (&lock);
    }
}

// Executes only on Consumer Thread
void BufferQueue::flush()
{
    pthread_mutex_lock ( &lock );
    int np = producerOwns;

    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);

    while ( b[consumerCurrent].isConsumer() != NULL )
    {
	TP(BQ_DETAIL, "%s: C-finished %d", __PRETTY_FUNCTION__, consumerCurrent);

        b[consumerCurrent].setProducer();
        producerOwns++;
        consumerCurrent++;
        if (consumerCurrent >= NUM_BUFFS)
        {
            consumerCurrent = 0;
        }
    }

    TP(BQ_DETAIL, "%s: complete, consumerCurrent now %d; (consumer owns: %s)",
       __PRETTY_FUNCTION__, consumerCurrent, (consumerOwnsCurrent() ? "YES" : "NO"));

    if ((np == 0) || fault )
    {
        if ( producer_signal != NULL )
        {
            pthread_mutex_unlock ( &lock );
	    TP(BQ_DETAIL, "%s: Signal producer via callback", __PRETTY_FUNCTION__ );
            (*producer_signal)( producer_signal_arg );
        }
        else
        {
            // Use built-in BufferQueue signaling
	    TP(BQ_DETAIL, "%s: Signal producer via pthread", __PRETTY_FUNCTION__ );
            pthread_cond_signal( &producer_check_ready );
            pthread_mutex_unlock ( &lock );
        }
    }
    else
    {
        pthread_mutex_unlock ( &lock );
    }
}

// Executes only on Consumer (WSS) Thread
void BufferQueue::fill(StreamingSession* s)
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);

    s->setBytesSoFar(0);

    pthread_mutex_lock ( &lock );
    TP(BQ_DETAIL, "%s: ( session %p ), txQueueHead=%p", __PRETTY_FUNCTION__, (void*)s, (void*)txQueueHead );

    if (txQueueHead == NULL)
    {
        // Queue is empty; new element goes at queue head
        s->setTxQueueLinks(NULL, NULL);
        txQueueHead = s;
        txQueueTail = s;
    }
    else
    {
        // Queue is not empty; new element appended to tail
        s->setTxQueueLinks(txQueueTail, NULL);
        txQueueTail->setTxQueueNext(s);
        txQueueTail = s;
    }

    pthread_mutex_unlock ( &lock );

    TP(BQ_DETAIL, "%s: Schedule writing", __PRETTY_FUNCTION__);
    s->scheduleWriteCallback();
}

bool BufferQueue::get_fault()
{
    TP(BQ_FUNC, "%s: fault=0x%x", __PRETTY_FUNCTION__, fault);
    return fault;
}

int BufferQueue::get_producer_buffer_count()
{
    TP(BQ_FUNC, "%s: %d", __PRETTY_FUNCTION__, producerOwns);
    return producerOwns;
}

bool BufferQueue::producer_owns_current()
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);
    return (b[producerCurrent].isProducer() != NULL);
}

// Executes only from producer thread
void BufferQueue::producer_wait()
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);

    pthread_mutex_lock (&lock);
    while ( producerOwns == 0 )
    {
	TP(BQ_DETAIL, "%s: Still waiting...", __PRETTY_FUNCTION__);
        pthread_cond_wait(&producer_check_ready, &lock);
    }
    pthread_mutex_unlock (&lock);
}

void BufferQueue::producer_set_flags(uint32_t f)
{
    TP(BQ_FUNC, "%s: flag=0x%x", __PRETTY_FUNCTION__, f);
    b[producerCurrent].setFlags(f);
}

char* BufferQueue::producer_get_data_ptr(size_t offset)
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);
    return b[producerCurrent].getDataPtr(offset);
}

void BufferQueue::producer_reserve_prefix( size_t size )
{
    TP(BQ_FUNC, "%s: size=%d", __PRETTY_FUNCTION__, size);
    b[producerCurrent].prefix_reserve( size );
}

int BufferQueue::producer_finished()
{
    int n = 0;

    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);

    pthread_mutex_lock(&lock);

    n = NUM_BUFFS - producerOwns;
    b[producerCurrent].setConsumer();
    producerOwns--;
    producerCurrent++;
    if (producerCurrent >= NUM_BUFFS)
    {
        producerCurrent = 0;
    }

    TP(BQ_DETAIL, "%s: producer owns %d buffers; current buffer is %d",
       __PRETTY_FUNCTION__, producerOwns, producerCurrent );

    pthread_mutex_unlock(&lock);
    return n; // Caller gets to see the number of consumer-owned buffers before this one was handed over
}

void BufferQueue::set_fault( bool f)
{
    TP(BQ_FUNC, "%s: fault=0x%x", __PRETTY_FUNCTION__, f);
    fault = f;
}

void BufferQueue::removeFromTxQueue(StreamingSession* s)
{
    TP(BQ_FUNC, "%s: session %p", __PRETTY_FUNCTION__, (void *)s);

    pthread_mutex_lock (&lock);
    StreamingSession* next = s->getTxQueueNext();
    StreamingSession* prev = s->getTxQueuePrev();
    s->setTxQueueLinks(NULL, NULL);

    TP(BQ_DETAIL, "%s: Remove session %p from Tx queue", __PRETTY_FUNCTION__, (void*)s );

    if (txQueueHead == s)
    {
        txQueueHead = next;
    }
    if (txQueueTail == s)
    {
        txQueueTail = prev;
    }
    if (next != NULL)
    {
        next->setTxQueuePrev(prev);
    }
    if (prev != NULL)
    {
        prev->setTxQueueNext(next);
    }

    pthread_mutex_unlock (&lock);
}

bool BufferQueue::isOnTxQueue(StreamingSession* s)
{
    TP(BQ_FUNC, "%s: txQueueHead=%p, s=%p, prev=%p, next=%p",
       __PRETTY_FUNCTION__, (void *)txQueueHead, (void *)s, (void *)s->getTxQueuePrev(), (void *)s->getTxQueueNext());

    return ((txQueueHead == s)           ||
            (s->getTxQueuePrev() != NULL) ||
            (s->getTxQueueNext() != NULL));
}

bool BufferQueue::consumerOwnsCurrent()
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);
    return (b[consumerCurrent].isConsumer() != NULL);
}

bool BufferQueue::txQueueIsEmpty()
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);
    return (txQueueHead == NULL);
}

size_t BufferQueue::getSize()
{
    TP(BQ_FUNC, "%s: size=%d", __PRETTY_FUNCTION__, size);
    return size;
}

bool BufferQueue::get_oldest()
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);
    return get_oldest_flag;
}

void BufferQueue::set_oldest(bool mode)
{
    TP(BQ_FUNC, "%s: mode=%d", __PRETTY_FUNCTION__, mode);
    get_oldest_flag = mode;
}

void BufferQueue::producer_set_size(size_t s)
{
    TP(BQ_FUNC, "%s: size=%d", __PRETTY_FUNCTION__, s);
    b[producerCurrent].setSize(s);
}

void BufferQueue::producer_set_signal( void (*func)( void* ), void* arg )
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);

    producer_signal_arg = arg;
    producer_signal = func;
}

size_t BufferQueue::consumer_get_size(size_t sf)
{
    TP(BQ_FUNC, "%s: sf=%d", __PRETTY_FUNCTION__, sf);

    if ( sf == 0 )
    {
        TP(BQ_DETAIL, "%s ->C[%d:%ld]->", __PRETTY_FUNCTION__, consumerCurrent, b[consumerCurrent].getSize());
    }
    return b[consumerCurrent].getSize();
}

size_t BufferQueue::consumer_get_size()
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);
    return b[consumerCurrent].getSize();
}

char* BufferQueue::consumerGetDataPtr(size_t offset)
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);
    return b[consumerCurrent].getDataPtr(offset);
}

uint32_t BufferQueue::getCurrentConsumerFlags()
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);
    return b[consumerCurrent].getFlags();
}

uint32_t BufferQueue::consumer_get_frame_data_size( void )
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);
    return b[consumerCurrent].get_frame_data_size();
}

void BufferQueue::producer_set_frame_data_size( uint32_t fds )
{
    TP(BQ_FUNC, "%s: fds=%d", __PRETTY_FUNCTION__, fds);
    b[producerCurrent].set_frame_data_size(fds);
}

int32_t BufferQueue::consumer_get_frame_id( void )
{
    TP(BQ_FUNC, "%s", __PRETTY_FUNCTION__);
    return b[consumerCurrent].get_frame_id();
}

void BufferQueue::producer_set_frame_id( int32_t n )
{
    TP(BQ_FUNC, "%s: id=%d", __PRETTY_FUNCTION__, n);
    b[producerCurrent].set_frame_id(n);
}

int BufferQueue::producer_index()
{
    TP(BQ_FUNC, "%s: %d", __PRETTY_FUNCTION__, producerCurrent);
    return producerCurrent;
}

int BufferQueue::consumer_index()
{
    TP(BQ_FUNC, "%s: %d", __PRETTY_FUNCTION__, consumerCurrent);
    return consumerCurrent;
}

