/*! --------------------------------------------------------------------------------
 * @file fi_cs.cxx
 * @brief implementation file for the fi_cs.cpp
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

// NOTES
// the idea here was to wrap the interface to fcp.c - in case a) fcp module
// changed or b) the wss wanted multiple camera server connections to be used.
// it may be overkill in any event and is a singleton object in context right
// now.

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <fastec/gige_msg.h>

#include "fi_cs.h"
#include "fcp.h"

//--------------------------------------------------------------------
// Forward Class Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class: fiWebSocketServer
//--------------------------------------------------------------------

/*! --------------------------------------------------------------------------------
 * @brief Camera Server IO Construction
 */
fiCameraServerIO::fiCameraServerIO() :
    cam_ip(0),
    cam_port(GIGE_SERVER_UDP_PORT),  // default the UDP Port Identifier
    cs_state(FI_CSIO_UNINITIALIZED)  // set to uninitialized state
{
    // Make sure our IP address is empty.
    cam_ip_arg.clear();
}

/*! --------------------------------------------------------------------------------
 * @brief Camera Server IO Destruction
 */
fiCameraServerIO::~fiCameraServerIO()
{
    // make sure to close down the connection with the Camera Server
    if (cs_state == FI_CSIO_CONNECTED)
    {
        cs_destroy();
        cs_state = FI_CSIO_UNINITIALIZED;
    }
}

/*! --------------------------------------------------------------------------------
 * @brief Is the Camera Server Connected?
 *
 * call to determine if the camera server is connected and
 * therefore ready to be used
 *
 * @return The number of bytes read, or <= 0 on error.
 */
bool fiCameraServerIO::isCSConnected()
{
    // make sure to close down the connection with the Camera Server
    if (cs_state == FI_CSIO_CONNECTED)
    {
        return true;
    }
    return false;
}


/*! --------------------------------------------------------------------------------
 * @brief Initializes Connection to the Camera Server
 *
 * @param in_p_cam_ip:          The IP string argument from command line or the
 *                              lowercase letter 'l' if camera server is local
 * @param in_cam_port:          The PORT argument or NULL from the command line
 */
FI_CSIO_STATES fiCameraServerIO::initCSConnection(const char *in_p_cam_ip, int in_cam_port)
{
    // if the port is not zero assign to the one supplied.
    if (in_cam_port != 0)
        cam_port = in_cam_port;

    //TODO check the size of the camera ip string is within MAX_IP_ADDRESS_LEN
    //TODO check the cam port and set it to NO_PORT
    // TODO provide this information if ip or port is invalid
    //FI_CSIO_MISSING_PORT,     // there was no port specified for CS use
    //FI_CSIO_MISSING_IP,           // there was no IP specified for CS use


    // pull the camara IP address or set the Local Server if that
    // is what the user specified in the input arguments
    if (in_p_cam_ip != (char *) '\0')
    {
        // assign the ip command line argument to the ip string object
        cam_ip_arg.assign(in_p_cam_ip);

        // if this is a local camera server set the state
        // local means that the Camera Server is resident on same OS and process
        // space as the fi_wss executable in this case there is no port or ip
        if (cam_ip_arg.compare("l") == 0)
        {
            cam_ip_arg.assign("Local Connection");
            cam_ip = 0;
            cam_port = 0;
        }
        else
        {
            // a remote cs connection is typically true if running the wss on
            // ubuntu system for debugging purposes and need to connect to cs
            // running remotely on connected camera.

            unsigned char ip[4];
            if (sscanf(in_p_cam_ip, "%hhu.%hhu.%hhu.%hhu", &ip[0], &ip[1], &ip[2], &ip[3]) == 4)
            {
                cam_ip = ((uint32_t)ip[0] << 24) | ((uint32_t)ip[1] << 16) |
                         ((uint32_t)ip[2] <<  8) | ((uint32_t)ip[3]);
            }
        }
    }

    //TODO what is a local connection
    // attempt to initialize the camera server
    if (cs_init(cam_ip, cam_port) == 0)
    {
        cs_state = FI_CSIO_INIT_FAILED;
    }
    else
    {
        cs_state = FI_CSIO_CONNECTED;
    }
    return cs_state;
}


/*! --------------------------------------------------------------------------------
 * @brief Closes and Quits Connection to the Camera Server
 */
void fiCameraServerIO::quitCSConnection()
{
    // make sure to close down the connection with the Camera Server
    if (cs_state == FI_CSIO_CONNECTED)
    {
        cs_destroy();
        cs_state = FI_CSIO_UNINITIALIZED;
    }
}

/*! --------------------------------------------------------------------------------
 * @brief Reads a FCP Register from Camera Server IO Object
 *
 * @param reg:          The fcp register to read
 * @param val:          A pointer to value storage for 32 bit register
 *
 * @return 1 on success, 0 on failure.
 */
int fiCameraServerIO::readFCPReg(uint32_t reg, uint32_t *val)
{
    int result = 0;

    // make sure to cs is connected
    if (cs_state == FI_CSIO_CONNECTED)
    {
        result = cs_get_reg_l(reg, val);
    }
    return result;
}

/*! --------------------------------------------------------------------------------
 * @brief Write a 32-bit register to the camera server.
 * @note This will update the cache.
 *
 * @param reg   Register ID we are interested in.
 * @param val   Value we want to write to the register.
 * @param istat Pointer for returning the status information.
 *
 * @return 1 on success, 0 on failure. On failure, refer to the
 *     contents of istat.
 */
int fiCameraServerIO::writeFCPReg(uint32_t reg, uint32_t val, int *istat)
{
    int result = 0;
    // make sure to cs is connected
    if (cs_state == FI_CSIO_CONNECTED)
    {
        result = cs_set_reg_l(reg, val, istat);
    }
    return result;
}

/*! --------------------------------------------------------------------------------
 * @brief This reads camera server array structures, such as strings.
 *
 * @param array_id The ID of the array we want to read.
 * @param data_ptr A pointer to a buffer we are going to fill with data.
 * @param data_len The maximum number of bytes which can go into that buffer.
 *
 * @return The number of bytes read, or <= 0 on error.
 */
int fiCameraServerIO::readFCPArray(int array_id, void *data_ptr, uint16_t data_len)
{
    int result = 0;
    // make sure to cs is connected
    if (cs_state == FI_CSIO_CONNECTED)
    {
        result = cs_read_array(array_id, data_ptr, data_len);
    }
    return result;
}


/*! --------------------------------------------------------------------------------
 * @brief This writes camera server array structures, such as strings.
 *
 * @param array_id The ID of the array we want to write.
 * @param data_ptr A pointer to a buffer with data to write to the server.
 * @param data_len The number of bytes which we are to write to the server.
 *
 * @return The number of bytes written, or <= 0 on error.
 */
int fiCameraServerIO::writeFCPArray(int array_id, const void *data_ptr, uint16_t data_len)
{
    int result = 0;
    // make sure to cs is connected
    if (cs_state == FI_CSIO_CONNECTED)
    {
        result = cs_write_array(array_id, data_ptr, data_len);
    }
    return result;
}

/*! --------------------------------------------------------------------------------
 * @brief Register Callback Handler designed to receive fcp register change notifications
 *
 * the FCP Register change notification handler for wss is registered
 * thru this interface - note still uses CS_REG_CALLBACK function signature
 *
 * @param fcp_listener:  The callback handler for live view images.
 *
 * @return 1 on success, 0 if no CS is connected
 */
int fiCameraServerIO::regFCPListener(CS_REG_CALLBACK fcp_listener)
{
    int result = 0;
    // make sure to cs is connected
    if (cs_state == FI_CSIO_CONNECTED)
    {
        cs_register_callback(fcp_listener, NULL);
    }
    return result;
}



// end of fi_cs.cxx

