/*
 * fi_download_protocol.cxx
 *
 *  Created on: Jan 25, 2016
 *      Author: eric
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <string.h>
#include <fastec/fas_log.h>
#include <fastec/lssm.h>
#include <fastec/fas_trace.h>

#include "fi_download_session.h"
#include "fi_download_protocol.h"
#include "fi_buffer_q.h"

//--------------------------------------------------------------------
// Constant, Type and Macro definitions
//--------------------------------------------------------------------

enum DNLD_TRACE
{
    DNLD_FUNC   = 0x00000010,
    DNLD_STATE  = 0x00000020,
    DNLD_BQ = 0x00000040,
    DNLD_DETAIL = 0x00000080
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

static uint32_t traceControl =
    DNLD_FUNC   |
    DNLD_STATE  |
    DNLD_BQ     |
    DNLD_DETAIL |
    TRACE_0;        // Set our trace level to TRACE_0 to trace.
//        TRACE_3;        // Set our trace level to TRACE_3 to not trace.

/**
 * @brief Constructor for the fiWssDownloadProtocol class.
 */
fiWSSDownloadProtocol::fiWSSDownloadProtocol(struct lws_protocols *p) :
    StreamingProtocol(p)
{
    TP(DNLD_FUNC, "%s", __PRETTY_FUNCTION__);

    p->name = "fi-download-protocol";
    p->rx_buffer_size = 48 * 1024;
}

fiWSSDownloadProtocol::~fiWSSDownloadProtocol()
{
    TP(DNLD_FUNC, "%s", __PRETTY_FUNCTION__);

    for (WssProtocol::iterator i = WssProtocol::begin();
         i != WssProtocol::end();
         i++)
    {
        delete *i;
    }
}

/**
 * @brief Release the current image memory buffer
 */
void fiWSSDownloadProtocol::release_image_memory_buffer(void *shmem)
{
    TP(DNLD_FUNC, "%s: buf_id=%d", __PRETTY_FUNCTION__, bq.buf_id);

    // Release the LSSM buffer we're finished processing, if we're
    // currently holding one.
    if (bq.buf_id >= 0)
    {
        lssm_lock(shmem);
        lssm_release_buffer_l(shmem, bq.buf_id);
        lssm_server_signal_l(shmem);
        bq.buf_id = -1;
        lssm_unlock(shmem);
    }
}

/**
 * @brief Restrict to one connection at a time.
 * @return 0 to allow the connect, 1 for an error.
 */
int fiWSSDownloadProtocol::accept(struct lws * /*wsi*/, WssSession ** /*user*/)
{
    TP(DNLD_FUNC, "%s", __PRETTY_FUNCTION__);

    // Only allow one session at a time.
    return (sessions.size() == 0) ? 0 : 1;
}

int fiWSSDownloadProtocol::established(struct lws *wsi, WssSession **user)
{
    int retval = 0;

    TP(DNLD_FUNC, "%s", __PRETTY_FUNCTION__);

    try
    {
        StreamingSession *s = new DownloadSession(this, wsi);
        s->setState(StreamingSession::ON_REQ);
        *((StreamingSession **)user) = s;
        sessions.push_back(s);

        bq.set_oldest(true);
        TP(DNLD_FUNC, "%s: new download session.", __PRETTY_FUNCTION__);
    }
    catch (std::exception& e)
    {
        retval = 1;
        fi_log(LOG_ERR, "%s: Exception establishing download session: %s", __PRETTY_FUNCTION__, e.what());
    }
    return retval;
}

int fiWSSDownloadProtocol::closed(struct lws *wsi, WssSession **user)
{
    int retval = 0;
    DownloadSession *s = *(DownloadSession **)user;

    TP(DNLD_FUNC, "%s", __PRETTY_FUNCTION__);

    if (s == NULL)
    {
        // Already closed, or never opened, or something
        fi_log(LOG_WARNING, "%s: Closing absent Download session", __PRETTY_FUNCTION__);
        return retval;
    }

    TP(DNLD_FUNC, "%s: close download session", __PRETTY_FUNCTION__);
    try
    {
        // Tell the session object that we are shutting down.
        s->setState(StreamingSession::OFF);
        for (std::vector<WssSession *>::iterator i = sessions.begin(); i != sessions.end(); i++)
        {
            if (*i == s)
            {
                TP(DNLD_DETAIL, "%s: Erase Download session %p/%d from sessions vector",
                   __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));
                sessions.erase(i);

                // Clear stale pointer
                *user = NULL;
                break;
            }
        }

        if (bq.isOnTxQueue(s))
        {
            TP(DNLD_BQ, "%s: Download session %p/%d on Tx queue (faulted, remove it)",
               __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));

            // If we're on the Tx queue, then there must be data in a Buffer
            // we expected to send. Downloads are not expected to end
            // prematurely, so we really want to propagate a fault condition
            // back to the source of the data. So, we set the 'faulted'
            // indication in the BufferQueue, and let it figure out where to
            // send it and how to send it.
            bq.removeFromTxQueue(s);
            bq.set_fault(true);
        }
        else
        {
            TP(DNLD_BQ, "%s: Download session %p/%d closed; not on Tx queue",
               __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));
        }

        if (bq.consumerOwnsCurrent())
            bq.finished();
        bq.flush(); // Ok to flush buffer queue; only ever one Download session, and it's ending

        // ALM: Do we want to do this here?
        // delete s;

    }
    catch (std::exception& e)
    {
        retval = 1;
        fi_log(LOG_ERR, "%s: Exception shutting down Download session: %p: %s", __PRETTY_FUNCTION__, (void *)s, e.what());
    }

    fi_log(LOG_INFO, "%s: Finished close callback", __PRETTY_FUNCTION__);
    return retval;
}

int fiWSSDownloadProtocol::destroy(struct lws * /*wsi*/, WssSession * /*user*/)
{
    TP(DNLD_FUNC, "%s", __PRETTY_FUNCTION__);
    return 0;
}

/**
 * Implementation notes:
 *
 * Ignoring error cases, each pass through this function will result
 * in one or two calls to lws_write(). There are five cases:
 *
 *   1 - Write 4-bye binary message that carries next Image Frame length (bytes) (FIN)
 *     - Write single binary message that carries entire Image Frame data (FIN)
 *
 *   2 - Write 4-bye binary message that carries next Image Frame length (bytes) (FIN)
 *     - Write binary message that carries initial N bytes of Image Frame data ( ! FIN)
 *
 *   3 - Write Continuation message that carries next N bytes of Image Frame data ( ! FIN)
 *
 *   4 - Write Continuation message that carries final N bytes of Image Frame data (FIN)
 *
 *   5 - Write additional bytes of in-progress message (Binary or Continuation, with or without FIN)
 *
 * We distinguish them as follows:
 *   bytes-so-far > 0:
 *       => case 5
 *   Buffer's START_OF_FRAME and END_OF_FRAME flags both set:
 *       => Case 1
 *   Buffer's START_OF_FRAME flag set (END_OF_FRAME clear):
 *       => Case 2
 *   Buffer's END_OF_FRAME flag set (START_OF_FRAME_CLEAR):
 *       => case 4
 *   Buffer's START_OF_FRAME and END_OF_FRAME flags both clear:
 *       => case 3
 *
 *
 */
int fiWSSDownloadProtocol::writable(struct lws *wsi, WssSession *user)
{
    uint32_t data_size = 0;
    StreamingProtocol::FRAME_HEADER_LWS header;
    DownloadSession *s = (DownloadSession *)user;

    TP(DNLD_FUNC, "%s: state=%d (%s)", __PRETTY_FUNCTION__, (int)s->getState(), s->decodeState());

    // lebwebsockets may issue spurious callbacks. Just let it go
    // if we get 'writable' callback and have nothing to write.
    if (!bq.isOnTxQueue(s))
    {
        TP(DNLD_DETAIL, "%s: Spurious write callback for Download protocol", __PRETTY_FUNCTION__);
        return 0;
    }

    if (s->getState() != StreamingSession::ON_FRAME)
    {
        if (s->getState() == StreamingSession::OFF_REQ)
        {
            TP(DNLD_STATE, "%s: Found session %p/%d in state OFF_REQ while on TxQ",
               __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));
        }
        else
        {
            fi_log(LOG_ERR, "%s: Found session %p/%d in unexpected state %d while on TxQ",
                   __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi), (int)s->getState());
        }
    }

    int soFar = s->getBytesSoFar();
    int total = bq.consumer_get_size();
    size_t bytesToWrite = total - soFar;
    uint32_t flags = bq.getCurrentConsumerFlags();
    bool sof = ((flags & StreamingProtocol::START_OF_FRAME) != 0);
    bool eof = ((flags & StreamingProtocol::END_OF_FRAME) != 0);
    int n = 0;

    if ((total == 0) || (bytesToWrite == 0))
    {
        // This is a logic error of some kind; should not be
        // on the Tx queue if Buffer is empty, or if we have
        // nothing to write.
        fi_log(LOG_ERR, "%s: Session %p/%d On Tx queue, but Buffer empty: total == %d, bytesToWrite == %ld (forcing close)",
               __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), total, bytesToWrite);
        return -1;
    }

    // Track information if we're tracing.
    TP(DNLD_BQ, "%s: id=%d, soFar=%d, total=%d, sof=%d, eof=%d, bytesToWrite=%d", __PRETTY_FUNCTION__, bq.consumer_index(), soFar, total, sof, eof, bytesToWrite);

    // Clear our header structure.
    memset((void *)&header, 0, sizeof(header));

    // Is the write buffered?
    if (lws_send_pipe_choked(wsi))
    {
        TP(DNLD_DETAIL, "%s: Partial buffered, schedule writing", __PRETTY_FUNCTION__);
        s->scheduleWriteCallback();
        return 0;
    }

    if (soFar > 0)
    {
        // Case 5
        // Send additional Image data from in-progress Buffer
        // TODO: Do we need to match the Websockets Frame type
        // (LWS_WRITE_BINARY or LWS_WRITE_CONTINUATION) of the
        // initial write for this Buffer?
        TP(DNLD_BQ, "%s: Session %p/%d: write additional data for frame",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));

        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, LWS_WRITE_BINARY);
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }
        soFar += n;
    }
    else if (sof && eof)
    {
        // Case 1
        data_size = bq.consumer_get_frame_data_size();
        header.h.magic = htonl(StreamingProtocol::MAGIC_ID);
        header.h.size = htonl(data_size);
        header.h.id = htonl(bq.consumer_get_frame_id());

        TP(DNLD_BQ, "%s: Session %p/%d: write first (and last) data for frame",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));

        s->set_frame_length(data_size);

        // Send header as separate Websocket segment
        n = lws_write(wsi, (unsigned char *)(&header.h), sizeof(header.h), LWS_WRITE_BINARY);
        if (n != sizeof(header.h))
        {
            fi_log(LOG_ERR, "%s: Frame Header write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }

        // Is the write buffered?
        if (lws_send_pipe_choked(wsi))
        {
            TP(DNLD_DETAIL, "%s: Partial buffered, schedule writing", __PRETTY_FUNCTION__);
            s->scheduleWriteCallback();
            return 0;
        }

        // Now start Image data
        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, LWS_WRITE_BINARY);
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }
        if (n < (int)bytesToWrite)
        {
            fi_log(LOG_ERR, "%s: Partial write: 0x%08x != 0x%08x", __PRETTY_FUNCTION__, n, bytesToWrite);
            return -1;
        }
        soFar += n;
    }
    else if (sof && (!eof))
    {
        uint32_t frame_id;

        // Case 2
        header.h.magic = htonl(StreamingProtocol::MAGIC_ID);
        data_size = bq.consumer_get_frame_data_size();
        header.h.size = htonl(data_size);
        frame_id = bq.consumer_get_frame_id();
        header.h.id = htonl(frame_id);

        TP(DNLD_BQ, "%s: Session %p/%d: write first data for frame %d (size=%d)",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), frame_id, data_size);

        // Tell the session the total length for the frame.
        s->set_frame_length(data_size);

        // Send header as separate Websocket segment
        n = lws_write(wsi, (unsigned char *)(&header.h), sizeof(header.h), LWS_WRITE_BINARY);
        if (n != sizeof(header.h))
        {
            fi_log(LOG_ERR, "%s: Frame Header write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }

        // Is the write buffered?
        if (lws_send_pipe_choked(wsi))
        {
            TP(DNLD_DETAIL, "%s: Partial buffered after header, schedule writing", __PRETTY_FUNCTION__);
            s->scheduleWriteCallback();
            return 0;
        }

        // Now start Image data
        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, (enum lws_write_protocol)(LWS_WRITE_BINARY | LWS_WRITE_NO_FIN));
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }
        if (n < (int)bytesToWrite)
        {
            fi_log(LOG_ERR, "%s: Partial write: 0x%08x != 0x%08x", __PRETTY_FUNCTION__, n, bytesToWrite);
            return -1;
        }
        soFar += n;
    }
    else if ((!sof) && eof)
    {
        // Case 3:
        // Image data
        TP(DNLD_BQ, "%s: Session %p/%d: write last data for frame",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));

        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, LWS_WRITE_CONTINUATION);
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }
        if (n < (int)bytesToWrite)
        {
            fi_log(LOG_ERR, "%s: Partial write: 0x%08x != 0x%08x", __PRETTY_FUNCTION__, n, bytesToWrite);
            return -1;
        }
        soFar += n;
    }
    else if ((!sof) && (!eof))
    {
        // Case 4:
        // Image data
        TP(DNLD_BQ, "%s: Session %p/%d: write continuation for frame",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));

        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, (enum lws_write_protocol)(LWS_WRITE_CONTINUATION | LWS_WRITE_NO_FIN));
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }
        if (n < (int)bytesToWrite)
        {
            fi_log(LOG_ERR, "%s: Partial write: 0x%08x != 0x%08x", __PRETTY_FUNCTION__, n, bytesToWrite);
            return -1;
        }
        soFar += n;
    }
    else
    {
        // Logic problem....
        fi_log(LOG_ERR, "%s: Got unexpected condition in write callback: %d, %s, %s",
               __PRETTY_FUNCTION__, soFar, (sof ? "SOF" : "! SOF"), (eof ? "EOF" : "! EOF"));
        return -1;
    }

    s->setBytesSoFar(soFar);
    if (soFar >= total)
    {
        TP(DNLD_BQ, "%s: Session %p/%d: finished writing data from Buffer",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));

        // Remove this session from Tx queue
        bq.removeFromTxQueue(s);

        if ((bq.getCurrentConsumerFlags() & StreamingProtocol::END_OF_FRAME) != 0)
        {
            switch (s->getState())
            {
            case StreamingSession::ON_FRAME:
                TP(DNLD_STATE, "%s: Session %p/%d: This buffer had END_OF_FRAME (ON_FRAME -> ON_IDLE)",
                   __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));
                s->setState(StreamingSession::ON_IDLE);
                break;

            case StreamingSession::OFF_REQ:
                TP(DNLD_STATE, "%s: OFF_REQ -> OFF", __PRETTY_FUNCTION__);
                break; // process_consumer() will complete transtition to StreamingSession::OFF

            default:
                fi_log(LOG_ERR, "%s: Found session %p/%d in unexpected state %d after completing Buffer",
                       __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi), (int)s->getState());
                break;
            }

        }
        if (bq.txQueueIsEmpty())
        {
            TP(DNLD_BQ, "%s: Session %p/%d:Tx queue is empty", __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));

            // Close out the buffer queue.
            bq.finished();

            // Anymore to add on our queue?
            process_consumer();
        }
    }
    else
    {
        TP(DNLD_DETAIL, "%s: Schedule writing", __PRETTY_FUNCTION__);
        s->scheduleWriteCallback();
    }

    // All done.
    return 0;
}

int fiWSSDownloadProtocol::receive(struct lws * /*wsi*/, WssSession * /*s*/, uint8_t * /*in*/, size_t /*len*/)
{
    TP(DNLD_FUNC, "%s", __PRETTY_FUNCTION__);
    return 0;
}
