/*!
 * @file fi_jsonpkg.cxx
 * @brief source file for the fi_jsonpkg
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 *
 * The core of the internal JSON read function was pulled from the microjson project
 * in github - this is left in the source to recognize the original author
 * otherwise the interfaces and data structures maintained are completely customized
 * for Fastec Imaging Control Protocol usage
 *
 * PERMISSIONS
 *   This file is Copyright (c) 2014 by Eric S. Raymond
 *   BSD terms apply: see the file COPYING in the distribution root for details.
 *
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <math.h>	/* for HUGE_VAL */

#include "fi_jsonpkg.h"

//--------------------------------------------------------------------
// Private Constant, Typedefs and Macros
//--------------------------------------------------------------------

// these values are used privately by the internal json read method to
// create buffers large enough to handle any data coming over the WSS
#define JSON_ATTR_MAX	48      /* max chars in JSON attribute name */
#define JSON_VAL_MAX	1024	/* max chars in JSON value part */

// #define DEBUG_ENABLE 1
#ifdef DEBUG_ENABLE

static int debuglevel = 3;
static FILE *debugfp = stdout;

void json_enable_debug(int level, FILE * fp)
/* control the level and destination of debug trace messages */
{
    debuglevel = level;
    debugfp = fp;
}

static void json_trace(int errlevel, const char *fmt, ...)
/* assemble command in printf(3) style */
{
    if (errlevel <= debuglevel)
    {
        char buf[BUFSIZ];
        va_list ap;

        (void)strncpy(buf, "json: ", BUFSIZ-1);
        buf[BUFSIZ-1] = '\0';
        va_start(ap, fmt);
        (void)vsnprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), fmt,
                        ap);
        va_end(ap);

        (void)fputs(buf, debugfp);
        fflush(debugfp);
    }
}

# define json_debug_trace(args) (void) json_trace args
#else
# define json_debug_trace(args) /*@i1@*///do { } while (0)
#endif /* DEBUG_ENABLE */

//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////

/* --------------------------------------------------------------------------------
 * @brief Initializes a json_buf record
 *
 * This function initialized a json buffer record to an empty state
 *
 * @param json_buf_ptr:                pointer to the json buffer to initialize
 *
 * @return contents of the json buffer are initialized to empty state
 * --------------------------------------------------------------------------------*/
static int init_json_buf( json_buf_ptr output )
{
    int result=0;
    if( output != (json_buf_ptr)NULL )
    {
        output->len = 0;
        output->p_buf = (uint8_t*) NULL;
        result = 1;
    }
    return result;
}

/* --------------------------------------------------------------------------------
 * @brief Initializes a json record
 *
 * This function initialized a json record to an empty state
 *
 * @param json_rec_ptr:     pointer to the json record to initialize
 *
 * @return contents of the json record are initialized to empty state
 * --------------------------------------------------------------------------------*/
static int init_json_rec( json_rec_ptr output )
{
    int result=0;
    if( output != (json_rec_ptr)NULL )
    {
        init_json_buf(&output->key);
        init_json_buf(&output->value);
        result = 1;
    }
    return result;
}

/* --------------------------------------------------------------------------------
 * @brief Initializes a json array
 *
 * This function initialized a json record to an empty state
 *
 * @param json_array_rec_ptr:     pointer to the json array to initialize
 *
 * @return contents of the json array are initialized to empty state
 * --------------------------------------------------------------------------------*/
static uint8_t init_json_array_rec( json_array_rec_ptr output )
{
    uint8_t result=0;
    uint8_t index;
    if( output != (json_array_rec_ptr)NULL )
    {
        // initialize the primary key for the array
        init_json_buf(&output->array_key);
        // set max available array items
        output->array_items_size = MAX_JSON_ARRAY_SIZE;
        output->array_items_inuse = 0;
        // initialize all of the array items empty
        for(index=0; index < output->array_items_size; index++)
        {
            init_json_rec(&output->array_items[index]);
        }
        result = 1;
    }
    return result;
}

#if 0
/* --------------------------------------------------------------------------------
 * @brief frees a json buffer
 *
 * This function frees any memory associated with a json buffer and initializes
 * contents back to empty state
 *
 * @param json_buf_ptr:     pointer to the json buffer to free
 *
 * @return contents of the json array are initialized to empty state
 * --------------------------------------------------------------------------------*/
static void free_json_buf( json_buf_ptr output )
{
    if( output != (json_buf_ptr)NULL )
    {
        // if allocated need to free the internal pointer
        if( output->p_buf != (uint8_t *) NULL )
        {
            free(output->p_buf);
            output->p_buf = (uint8_t *)NULL;
        }
        init_json_buf(output);		// initialize empty
    }
}

/* --------------------------------------------------------------------------------
 * @brief frees a json record
 *
 * This function frees any memory associated with a json record and initializes
 * contents back to empty state
 *
 * @param json_rec_ptr:     pointer to the json record to free
 *
 * @return contents of the json record are initialized to empty state
 * --------------------------------------------------------------------------------*/
static int free_json_rec( json_rec_ptr output )
{
    int result=0;
    if( output != (json_rec_ptr)NULL )
    {
        // free the json buffers for the key value pair
        free_json_buf(&output->key);
        free_json_buf(&output->value);
        result = 1;
    }
    return result;
}

/* --------------------------------------------------------------------------------
 * @brief frees a json array
 *
 * This function frees any memory associated with a json array and initializes
 * contents back to empty state
 *
 * @param json_array_rec_ptr:     pointer to the json array to free
 *
 * @return contents of the json array are initialized to empty state
 * --------------------------------------------------------------------------------*/
static uint8_t free_json_array_rec( json_array_rec_ptr output )
{
    uint8_t result=0;
    uint8_t index;
    if( output != (json_array_rec_ptr)NULL )
    {
        // free the primary key for the array
        free_json_buf(&output->array_key);
        // set max available array items
        output->array_items_size = MAX_JSON_ARRAY_SIZE;
        output->array_items_inuse = 0;
        // initialize all of the array items empty
        for(index=0; index < output->array_items_size; index++)
        {
            free_json_rec(&output->array_items[index]);
        }
        result = 1;
    }
    return result;
}
#endif

/* --------------------------------------------------------------------------------
 * @brief set the data for a json buffer
 *
 * This function frees any memory associated with the json buffer and allocates new
 * memory with the new size + 1 for an ending zero code and copies over the data putting
 * a zero at the very end.  Some data may not be string data in the case of base 64 encoded
 * strings TODO revisit this for the encoded data
 *
 * @param json_buf_ptr:     pointer to the json buffer to free
 *
 * @return contents of the json array are initialized to empty state
 * --------------------------------------------------------------------------------*/
static int set_json_buf( json_buf_ptr output, uint16_t in_len,  uint8_t *in_p_buf )
{
    int result=0;
    int alloc_len = in_len + 1; // want to put a zero at very end

    if( output != (json_buf_ptr)NULL )
    {
        // if allocated need to free the internal pointer
        if( output->p_buf != (uint8_t*) NULL )
        {
            output->len = 0;
            free(output->p_buf);
            output->p_buf = (uint8_t*)NULL;
        }

        // allocate the buffer
        if( in_len > 0 )
        {
            output->p_buf = (uint8_t*)malloc(alloc_len);
            if( output->p_buf != (uint8_t *)NULL )
            {
                memset(output->p_buf, 0, alloc_len );
                output->len = in_len;
                // copy over the source data to store
                if( in_p_buf != (uint8_t *) NULL )
                    memcpy(output->p_buf, in_p_buf, in_len);
                result = 1;	// successful
            }
        }
        else
            output->len=0;  // should already be zero but make sure.
    }
    return result;
}

/* --------------------------------------------------------------------------------
 * @brief set the data for a json record
 *
 * calls the routine to install data into a json buffer one for the key and one
 * for the data - lengths define the size of the data to install assuming that
 * some data may not be string data with a null terminator TODO - revisit this.
 *
 * @param json_rec_ptr:     pointer to the json record to store the key value data
 *
 * @return contents of the json record contain the new key and data values
 * --------------------------------------------------------------------------------*/
static int set_json_rec( json_rec_ptr output, uint16_t in_key_len,  uint8_t *in_key_str, uint16_t in_val_len,  uint8_t *in_val_str  )
{
    int result=0;
    if( output != (json_rec_ptr)NULL )
    {
        // attempt to install the key information
        result = set_json_buf(&output->key, in_key_len, in_key_str);
        // attempt to install the value information
        result = set_json_buf(&output->value, in_val_len, in_val_str);
    }
    return result;
}

// install an key value pair into the json package being built by the parser
static int add_json_pkg_key_value( fi_json_parse_pkg_ptr output, uint16_t in_key_len,  uint8_t *in_key_str, uint16_t in_val_len,  uint8_t *in_val_str )
{
    int result=0;
    if( output != (fi_json_parse_pkg_ptr)NULL )
    {
        // is there room for this data?
        if( output->key_values_inuse < output->key_values_size )
        {
            result = set_json_rec( &output->key_values[output->key_values_inuse], in_key_len, in_key_str, in_val_len,  in_val_str  );
            if( result )
                output->key_values_inuse++;
        }
    }
    return result;
}


// install an json_pkg_array_key_value into the json package being built by the parser
static int add_json_pkg_array_item( fi_json_parse_pkg_ptr output, uint16_t in_key_len,  uint8_t *in_key_str,
                                    uint16_t in_itemkey_len,  uint8_t *in_itemkey_str, uint16_t in_itemval_len,  uint8_t *in_itemval_str )
{
    int result=0;
    if( output != (fi_json_parse_pkg_ptr)NULL )
    {
        // check to see if primary key already added if found use this record if not found install it
        int x;
        int index = -1;
        for(x=0; (x < output->json_arrays_inuse) && (index == -1); x++)
        {
            // if primary keys match the record is found
            if( memcmp(in_key_str, output->json_arrays[x].array_key.p_buf, in_key_len) == 0 )
                index = x;
        }

        // if the primary key is not there can this be added?
        if( index == -1 )
        {
            uint8_t next_free = output->json_arrays_inuse;

            // is there room for this new entry?
            if( next_free < output->json_arrays_size )
            {
                // the primary key is added first
                result = set_json_buf( &output->json_arrays[next_free].array_key, in_key_len, in_key_str  );
                if( result )
                {
                    index = next_free;  // use this record;
                    output->json_arrays_inuse++;    // we are using the JSON array at index
                }
            }
        }

        // if we have a place to put the array item, set it in the array
        if( index != -1 )
        {
            if( index <  output->json_arrays_size )
            {
                // how many array items are in use
                uint8_t nbr_array_items = output->json_arrays[index].array_items_inuse;
                uint8_t max_array_items = output->json_arrays[index].array_items_size;

                // is there room for another array item?
                if( nbr_array_items < max_array_items )
                {
                    // now add the array information
                    json_rec_ptr json_array_list_ptr = &output->json_arrays[index].array_items[nbr_array_items];
                    result = set_json_rec( json_array_list_ptr, in_itemkey_len, in_itemkey_str, in_itemval_len, in_itemval_str );
                    if( result )
                        output->json_arrays[index].array_items_inuse++; // we have added an array element item to the array list
                }
                else
                    result = 0; // out of space!
            }
            else
                result = 0; // out of space!
        }
    }
    return result;
}

/* --------------------------------------------------------------------------------
 * @brief Query a JSON parsed package for the value of a key
 *
 * looks for a key in a JSON parsed package and returns the value associated with the key
 *
 * @param fi_json_parsed_pkg_ptr:     pointer to JSON parsed package containing the key
 *
 * @return const pointer to the value to use if the key was found otherwise a NULL pointer is returned
 * --------------------------------------------------------------------------------*/
const uint8_t* find_value_for_key( fi_json_parse_pkg_ptr json_parsed_pkg, uint8_t * key_ptr)
{
    // if parsed package pointer is valid start searching for the key requested
    if( json_parsed_pkg != (fi_json_parse_pkg_ptr) NULL )
    {
        uint16_t key_size = strlen((const char *)key_ptr);
        uint16_t index;
        for(index=0; index < json_parsed_pkg->key_values_inuse; index++)
        {
            // if the keys match return the value
            if( memcmp((const void*)json_parsed_pkg->key_values[index].key.p_buf, (const void*)key_ptr, key_size) == 0 )
                return ((const uint8_t*) json_parsed_pkg->key_values[index].value.p_buf );  // return the const pointer
        }
    }
    return NULL;
}

/* --------------------------------------------------------------------------------
 * @brief Query a JSON parsed package for the value of a key
 *
 * return the primary key identifying an array collection from a JSON package
 *
 * @param fi_json_parsed_pkg_ptr:   pointer to JSON parsed package containing the key
 * @param array_key_index:          zero-based index into the array for the return key value
 *
 * @return const pointer to the value to use if the key was found otherwise a NULL pointer is returned
 * --------------------------------------------------------------------------------*/
const uint8_t* get_array_key( fi_json_parse_pkg_ptr parsed_pkg_ptr, uint16_t array_key_index)
{
    const uint8_t * p_key_result = (uint8_t *) NULL;
    if( parsed_pkg_ptr != ( fi_json_parse_pkg_ptr) NULL )
    {
        if( array_key_index < parsed_pkg_ptr->json_arrays_inuse )
        {
            // should have a key index or null pointer
            p_key_result = (const uint8_t *)parsed_pkg_ptr->json_arrays[array_key_index].array_key.p_buf;
        }
    }
    return p_key_result;
}


// initialize the json parse package to empty state
uint8_t init_fi_json_parse_pkg( fi_json_parse_pkg_ptr output )
{
    uint8_t result=0;
    uint8_t index;
    if( output != (fi_json_parse_pkg_ptr)NULL )
    {
        // set max available key value pairs
        output->key_values_size = MAX_FI_JSON_KEYS;
        // set the number in use to 0
        output->key_values_inuse = 0;

        // initialize all of the key value pair records
        for(index=0; index < output->key_values_size; index++)
        {
            init_json_rec(&output->key_values[index]);
        }

        // initialize the array data
        output->json_arrays_size = MAX_FI_JSON_ARRAYS;
        output->json_arrays_inuse = 0;

        // initialize all of the array records
        for(index=0; index < output->json_arrays_size; index++)
        {
            init_json_array_rec(&output->json_arrays[index]);
        }

        result = 1;
    }
    return result;
}

void free_fi_json_parse_pkg( fi_json_parse_pkg_ptr pkg )
{
    uint8_t index;
    uint8_t i;

    if( pkg != (fi_json_parse_pkg_ptr)NULL )
    {
        for (index = 0; index < pkg->key_values_inuse; index++ )
        {
            if ( pkg->key_values[index].key.p_buf != (uint8_t*)NULL )
            {
                free( pkg->key_values[index].key.p_buf );
            }
            if ( pkg->key_values[index].value.p_buf != (uint8_t*)NULL )
            {
                free( pkg->key_values[index].value.p_buf );
            }
            pkg->key_values[index].key.p_buf = (uint8_t*)NULL;
            pkg->key_values[index].key.len = 0;
        }
        pkg->key_values_inuse = 0;
        pkg->key_values_size = MAX_FI_JSON_KEYS;

        for ( index = 0; index < pkg->json_arrays_inuse; index++ )
        {
            for ( i = 0; i < pkg->json_arrays[index].array_items_inuse; i++ )
            {
                if ( pkg->json_arrays[index].array_items[i].key.p_buf != (uint8_t*)NULL )
                {
                    free( pkg->json_arrays[index].array_items[i].key.p_buf );
                }
                pkg->json_arrays[index].array_items[i].key.p_buf = (uint8_t*)NULL;
                pkg->json_arrays[index].array_items[i].key.len = 0;

                if ( pkg->json_arrays[index].array_items[i].value.p_buf != (uint8_t*)NULL )
                    free( pkg->json_arrays[index].array_items[i].value.p_buf );
                pkg->json_arrays[index].array_items[i].value.p_buf = (uint8_t*)NULL;
                pkg->json_arrays[index].array_items[i].value.len = 0;
            }

            if ( pkg->json_arrays[index].array_key.p_buf != (uint8_t*)NULL )
            {
                free( pkg->json_arrays[index].array_key.p_buf );
            }
            pkg->json_arrays[index].array_key.p_buf = (uint8_t*)NULL;
            pkg->json_arrays[index].array_key.len = 0;

            pkg->json_arrays_size = MAX_FI_JSON_ARRAYS;
            pkg->json_arrays_inuse = 0;
        }
    }
}




/* this will parse out the contents of a json package into key data pairs */
static int json_internal_read_object(const char *cp,    /* this is the input JSON buffer to parse */
                                     fi_json_parse_pkg_ptr json_data_out)
{
    enum
    {
        init = 0,
        await_attr,
        in_attr,
        await_value,
        in_val_string,
        in_val_token,
        post_val,
    } state;
#ifdef DEBUG_ENABLE
    const char *statenames[] =
    {
        "init", "await_attr", "in_attr", "await_value", "in_val_string",
        "in_val_token", "post_val",
    };
#endif /* DEBUG_ENABLE */

    state = init;

    char attrbuf[JSON_ATTR_MAX + 1], *pattr = NULL;
    char valbuf[JSON_VAL_MAX + 1], *pval = NULL;
    char array_attrbuf[JSON_ATTR_MAX + 1];//, *parray_attr = NULL;

    //bool value_quoted = false;

    //char uescape[5];		/* enough space for 4 hex digits and a NUL */
    //int n;
    int maxlen = 0;
    // unsigned int u;

    // if processing an array this will be set
    bool in_array = false;

    uint16_t attrlen = 0;
    uint16_t vallen = 0;
    uint16_t array_attrlen = 0;

    /* prevents gripes about buffers not being completely defined */
    memset(valbuf, '\0', sizeof(valbuf));
    memset(attrbuf, '\0', sizeof(attrbuf));
    memset(array_attrbuf, '\0', sizeof(array_attrbuf));

    json_debug_trace((1, "JSON parse of '%s' begins.\n", cp));

    /* parse input JSON */
    for (; *cp != '\0'; cp++)
    {
        json_debug_trace((2, "State %-14s, looking at '%c' (%p)\n",
                          statenames[state], *cp, cp));

        switch (state)
        {
        case init:
            if (isspace((unsigned char) *cp))
            {
                continue;
            }
            else if (*cp == '{')
            {
                state = await_attr;
            }
            else
            {
                json_debug_trace((1, "Non-WS when expecting object start.\n"));
                return JSON_ERR_OBSTART;
            }
            break;
        case await_attr:
            // reset the keys
            attrlen = 0;
            attrbuf[0] = '\0';
            vallen = 0;
            valbuf[0] = '\0';

            if (isspace((unsigned char) *cp))
            {
                continue;
            }
            else if (*cp == '"')
            {
                state = in_attr;
                pattr = attrbuf;
            }
            else if (*cp == '}')
            {
                break;
            }
            else if( in_array )
            {
                cp--;   // don't want to lose this char in next pass for state change
                state = await_value;    // no key in the array value
            }
            else
            {
                json_debug_trace((1, "Non-WS when expecting attribute.\n"));
                return JSON_ERR_ATTRSTART;
            }
            break;
        case in_attr:
            if (pattr == NULL)
            {
                /* don't update end here, leave at attribute start */
                return JSON_ERR_NULLPTR;
            }
            if (*cp == '"')
            {
                *pattr++ = '\0';
                json_debug_trace((1, "Collected attribute name %s\n",
                                  attrbuf));
                /* changing state to await the value component of the key */
                state = await_value;
                maxlen = JSON_VAL_MAX;
                pval = valbuf;
            }
            else if (pattr >= attrbuf + JSON_ATTR_MAX - 1)
            {
                json_debug_trace((1, "Attribute name too long.\n"));
                /* don't update end here, leave at attribute start */
                return JSON_ERR_ATTRLEN;
            }
            else
            {
                *pattr++ = *cp;
            }
            break;
        case await_value:
            // if space or key value separater just continue
            if (isspace((unsigned char) *cp) || (*cp == ':') )
            {
                continue;
            }

            // if already in an array check for case where it's only values separated by commas
            if ( in_array )
            {
                // if embedded braces don't want to count the characters just want
                // to skip them --
                if( *cp == '{' || *cp == '}' )
                {
                    // We may be in any object.
                    if ( *cp == '{' )
                    {
                        state = await_attr;
                    }

                    // still looking for array field separater or the closing array character
                    // so just continue on in the process
                    continue;
                }

                // has the current array item come to an end yet?
                if( *cp == ',' || *cp == ']')
                {
                    // this is case where array item is only values separated by commas so attribute
                    // already parsed is the value in this case
                    state = post_val;

                    // copy over the attribute since there is no key
                    memcpy(&valbuf[0], &attrbuf[0], JSON_ATTR_MAX );
                    valbuf[JSON_ATTR_MAX] = '\0';
                    attrbuf[0] = '\0';
                    attrlen = 0;
                    pattr = attrbuf;

                    // if end of the array backup the cp pointer
                    if( *cp == ']' )
                    {
                        cp--;   // catch end of array in post_val processing
                    }
                    continue;
                }
            }
            else if (*cp == '[')
            {
                // want to save off the array primary key attribute opening the array
                memcpy(&array_attrbuf[0], &attrbuf[0], JSON_ATTR_MAX );
                array_attrbuf[JSON_ATTR_MAX] = '\0';

                // also want to clear the old attribute buffer to locate internal
                // array keys if they exist
                attrbuf[0] = '\0';
                attrlen = 0;
                pattr = attrbuf;

                // set the state of processing an internal array
                in_array = true;
                continue;
            }

            // check for next value
            if (*cp == '"')
            {
                state = in_val_string;
                pval = valbuf;
            }
            else
            {
                //value_quoted = false;
                state = in_val_token;
                pval = valbuf;
                *pval++ = *cp;
            }
            break;
        case in_val_string:
            if (pval == NULL)
            {
                /* don't update end here, leave at value start */
                return JSON_ERR_NULLPTR;
            }
            if (*cp == '"')
            {
                *pval++ = '\0';
                json_debug_trace((1, "Collected string value %s\n", valbuf));
                state = post_val;
            }
            else if ((pval > (valbuf + JSON_VAL_MAX - 1)) || (pval > (valbuf + maxlen)))
            {
                json_debug_trace((1, "String value too long.\n"));
                /* don't update end here, leave at value start */
                return JSON_ERR_STRLONG;	/*  */
            }
            else
            {
                *pval++ = *cp;
            }
            break;

        case in_val_token:
            if (pval == NULL)
            {
                /* don't update end here, leave at value start */
                return JSON_ERR_NULLPTR;
            }
            if (isspace((unsigned char) *cp) || *cp == ',' || *cp == '}' || *cp == ']' )
            {
                *pval = '\0';
                json_debug_trace((1, "Collected token value %s.\n", valbuf));
                state = post_val;
                if (*cp == '}' || *cp == ',' || *cp == ']')
                {
                    --cp;
                }
            }
            else if (pval > valbuf + JSON_VAL_MAX - 1)
            {
                json_debug_trace((1, "Token value too long.\n"));
                /* don't update end here, leave at value start */
                return JSON_ERR_TOKLONG;
            }
            else
            {
                *pval++ = *cp;
            }
            break;

        case post_val:
            /*
             * We know that cursor points at the first spec matching
             * the current attribute.  We don't know that it's *the*
             * correct spec; our dialect allows there to be any number
             * of adjacent ones with the same attrname but different
             * types.  Here's where we try to seek forward for a
             * matching type/attr pair if we're not looking at one.
             */
            // determine length of the attribute and value information
            array_attrlen = (uint16_t) strlen(array_attrbuf);
            attrlen = (uint16_t) strlen(attrbuf);
            vallen = (uint16_t) strlen(valbuf);

            // are we in an array?
            if( in_array )
            {
                // if this is a key:data separator : character need to go back and get the value part
                if( *cp == ':' )
                {
                    if( attrlen == 0 )
                    {
                        // flip flop the values because there is a key in this array item
                        attrlen = vallen;
                        memcpy( &attrbuf[0], &valbuf[0], JSON_ATTR_MAX );
                        attrbuf[JSON_ATTR_MAX] = '\0';
                        state = await_value;    // the key was saved now let's get the value
                        continue;
                    }
                }

                // need to add the array value - if the array element is not in a key:value form the
                // value is sitting in the attribute buffer - otherwise there are both values.
                // should have a primary key and either a key:value for array item or only a value
                add_json_pkg_array_item(json_data_out,
                                        array_attrlen, (uint8_t*)&array_attrbuf[0],
                                        attrlen, (uint8_t*)&attrbuf[0],
                                        vallen, (uint8_t*)&valbuf[0] );

            }
            else
            {
                // it's a key:value pair in a non-array format
                add_json_pkg_key_value(json_data_out, attrlen, (uint8_t*)&attrbuf[0], vallen, (uint8_t*)&valbuf[0]);
            }

            /*@fallthrough@*/
            if( in_array )
            {
                // the array value has been installed.  depending on the
                // format of the array the end of an array item can look like
                // '}', or '}]' or ',' or ']' determine what it is and how to setup the
                // next iteration
                if( *cp == '}')
                {
                    cp++;
                }

                // push past the end of the array item which can be different depending on
                // the configuration of the JSON request
                while( (state == post_val) && (*cp != '\0') )
                {
                    json_debug_trace((2, "State %-14s, looking at '%c' (%p)\n",
                                      statenames[state], *cp, cp));

                    if( isspace((unsigned char) *cp) )
                    {
                        cp++;
                    }
                    else if( *cp == ',' )
                    {
                        state = await_value;
                    }
                    else if( *cp == ']' )
                    {
                        in_array = false;   // no more array elements
                        state = await_attr;
                    }
                    else    // this is not valid
                    {
                        in_array = false;   // not in array anymore invalid json
                        return JSON_ERR_BADTRAIL;
                    }
                }
            }
            // keep loop to rid whitespace
            else if (isspace((unsigned char) *cp))
            {
                continue;
            }
            // comma deliminates key value pair in non-array condition
            else if (*cp == ',')
            {
                state = await_attr;
            }
            // end parenthesis defines end of the parse
            else if (*cp == '}')
            {
                ++cp;
                goto good_parse;
            }
            else
            {
                json_debug_trace((1, "Garbage while expecting comma or }\n"));
                return JSON_ERR_BADTRAIL;
            }
            break;
        }   // end of the switch statement on the state
    }   // end of the parsing loop

good_parse:
    /* in case there's another object following, consume trailing WS */
    while (isspace((unsigned char) *cp))
        ++cp;

    json_debug_trace((1, "JSON parse ends.\n"));
    return 0;
    /*@ +nullstate +nullderef +mustfreefresh +nullpass +usedef @*/
}

int fi_json_parse(const char *cp, fi_json_parse_pkg_ptr json_data_out)
{
    int st;

    json_debug_trace((1, "json_parse() sees '%s'\n", cp));
    st = json_internal_read_object(cp, json_data_out);
    return st;
}

const /*@observer@*/ char *json_error_string(int err)
{
    const char *errors[] =
    {
        "unknown error while parsing JSON",
        "non-whitespace when expecting object start",
        "non-whitespace when expecting attribute start",
        "unknown attribute name",
        "attribute name too long",
        "saw [ when not expecting array",
        "array element specified, but no [",
        "string value too long",
        "token value too long",
        "garbage while expecting comma or } or ]",
        "didn't find expected array start",
        "error while parsing object array",
        "too many array elements",
        "garbage while expecting array comma",
        "unsupported array element type",
        "error while string parsing",
        "check attribute not matched",
        "can't support strings in parallel arrays",
        "invalid enumerated value",
        "saw quoted value when expecting nonstring",
        "didn't see quoted value when expecting string",
        "other data conversion error",
        "unexpected null value or attribute pointer",
    };

    if (err <= 0 || err >= (int)(sizeof(errors) / sizeof(errors[0])))
        return errors[0];
    else
        return errors[err];
}

/* end */

