/*
 * fi_liveview_protocol.cxx
 *
 *  Created on: Apr 25, 2016
 *      Author: eric
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <exception>
#include <string.h>

#include <fastec/gige_msg.h>
#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

#include "fcp.h"

#include "fi_jsonpkg.h"
#include "fi_liveview_session.h"
#include "fi_liveview_protocol.h"
#include "fi_buffer_q.h"

//--------------------------------------------------------------------
// Constant, Type and Macro definitions
//--------------------------------------------------------------------

enum LVP_TRACE
{
    LVP_FUNC    = 0x00000010,
    LVP_STATE   = 0x00000020,
    LVP_BQ  = 0x00000040,
    LVP_DETAIL  = 0x00000080
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

static uint32_t traceControl =
    LVP_FUNC    |
    LVP_STATE   |
    LVP_BQ      |
    LVP_DETAIL  |
//        TRACE_0;        // Set our trace level to TRACE_0 to trace.
    TRACE_3;        // Set our trace level to TRACE_3 to not trace.

/** \class fiWSSLiveViewProtocol
 *
 * ### Implementaton notes:
 *
 * The Live View protocol uses session states available from StreamingSession
 * to manage the live view stream to each client, so each client will see
 * the stream start and end between complete image frames. That is, the
 * first data on a live view stream seen by a client joining the stream will
 * be the header of a new image. All of these state transitions take place in
 * one of the fiWSSLiveViewProtocol callback methods. Specifically:
 *     - fiWSSLiveViewProtocol::established
 *     - fiWSSLiveViewProtocol::writable
 *     - fiWSSLiveViewProtocol::receive
 *
 * ![Streaming Session States](liveview_states.svg)
 */

/**
 * @brief Trampoline for custom signaling hook
 *
 * This just provides a static function so we can pass an address
 * to cs_register_callback(). CS thread will use this to tell us
 * about changes in the Camera Server's GIGE_REG_CAMERA_STATE register.
 */
static void liveview_callback(void *arg, uint32_t reg, uint32_t val)
{
    static_cast<fiWSSLiveViewProtocol *>(arg)->notify(reg, val);
}

fiWSSLiveViewProtocol::fiWSSLiveViewProtocol(struct lws_protocols *p) :
    StreamingProtocol(p),
    cs_is_liveview(false),
    cs_is_liveview_update(false),
    frame_id(0)
{
    TP(LVP_FUNC, "%s", __PRETTY_FUNCTION__);

    p->name = "fi-liveview-protocol";
    p->rx_buffer_size = 48 * 1024;      // ALM: Why 48K?
    p->tx_packet_size = Buffer::BUFF_DATA_SIZE + (2 * Buffer::BUFF_PREFIX_SIZE);

    // Register a callback handler with the camera server interface.
    cs_register_callback(liveview_callback, this);
}

fiWSSLiveViewProtocol::~fiWSSLiveViewProtocol()
{
    TP(LVP_FUNC, "%s", __PRETTY_FUNCTION__);

    for (WssProtocol::iterator i = WssProtocol::begin(); i != WssProtocol::end(); i++)
    {
        delete *i;
    }
}

uint32_t fiWSSLiveViewProtocol::get_next_id()
{
    uint32_t next_id = ++frame_id;

    TP(LVP_FUNC, "%s: frame_id=%d (0x%x)", __PRETTY_FUNCTION__, next_id, next_id);
    return next_id;
}

bool fiWSSLiveViewProtocol::has_data()
{
    // We have listeners, and we are in live view, and there is data queued.
    bool rc = ((listeners > 0) && cs_is_liveview && (bq.get_producer_buffer_count() < BufferQueue::NUM_BUFFS));
    TP(LVP_FUNC, "%s: listeners=%d, cs_is_liveview=%c buffers=%d -> %c",
       __PRETTY_FUNCTION__, listeners, (cs_is_liveview ? 'T' : 'F'), bq.get_producer_buffer_count(),
       (rc ? 'T' : 'F'));
    return rc;
}

void fiWSSLiveViewProtocol::process_consumer()
{
    int cs_ret = 0;

    // TOO NOISY! Only enable if absolutely necessary!
    // LOG("%s: listeners=%d", __PRETTY_FUNCTION__, listeners);

    // Do we need to shutdown the stream?
    if ((listeners == 0) && cs_is_liveview)
    {
        TP(LVP_STATE, "%s: No listeners, shutdown live view.", __PRETTY_FUNCTION__);
        cs_is_liveview_update = false;
    }

    // Do we need to enable the live view interface?
    if (!cs_is_liveview && cs_is_liveview_update && (listeners > 0))
    {
        TP(LVP_STATE, "%s: Turn on LIVE view", __PRETTY_FUNCTION__);

        // The camera server entered a live view state.
        if (cs_set_reg_l(GIGE_REG_LOCALSTREAM_CONTROL, GIGE_LOCALSTREAM_LIVE, &cs_ret) == 0)
        {
            fi_log(LOG_ERR, "%s: Could not set GIGE_REG_LOCALSTREAM_CONTROL (%d)", __PRETTY_FUNCTION__, cs_ret);
        }

        // We are in live view.
        cs_is_liveview = true;
    }
    else if (cs_is_liveview && !cs_is_liveview_update)
    {
        uint32_t current_state = (uint32_t) -1;

        // Make sure we're in live view before we turn it off.
        if (cs_get_reg_l(GIGE_REG_LOCALSTREAM_CONTROL, &current_state) == 0)
            fi_log(LOG_ERR, "%s: cs_get_reg_l(GIGE_REG_LOCALSTREAM_CONTROL) failed", __PRETTY_FUNCTION__);
        if (current_state == GIGE_LOCALSTREAM_LIVE)
        {
            TP(LVP_STATE, "%s: Turn off LIVE view", __PRETTY_FUNCTION__);

            // We are no longer in live view.
            if (cs_set_reg_l(GIGE_REG_LOCALSTREAM_CONTROL, GIGE_LOCALSTREAM_STOP, &cs_ret) == 0)
            {
                fi_log(LOG_ERR, "%s: Could not set GIGE_REG_LOCALSTREAM_CONTROL (%d)", __PRETTY_FUNCTION__, cs_ret);
            }
        }
        else
        {
            TP(LVP_STATE, "%s: No longer in live view, state=0x%x", __PRETTY_FUNCTION__, current_state);
        }

        // No longer in liveview.
        cs_is_liveview = false;
    }

    // Call base-class method to do normal processing
    StreamingProtocol::process_consumer();
}

/**
 * @brief Release the current Live View image memory buffer
 */
void fiWSSLiveViewProtocol::release_image_memory_buffer(void *shmem)
{
    TP(LVP_FUNC, "%s: buf_id=%d", __PRETTY_FUNCTION__, bq.buf_id);

    // Release the LSSM buffer we're finished processing, if we're
    // currently holding one.
    if (bq.buf_id >= 0)
    {
        lssm_lock(shmem);
        TP(LVP_BQ, "%s: Release IM buffer %d", __PRETTY_FUNCTION__, bq.buf_id);

        lssm_release_buffer_l(shmem, bq.buf_id);
        lssm_server_signal_l(shmem);
        bq.buf_id = -1;
        lssm_unlock(shmem);
    }
}

void fiWSSLiveViewProtocol::notify(uint32_t reg, uint32_t val)
{
    bool old_update = cs_is_liveview_update;

    // Make sure we have context.
    if (ctx == NULL)
        return;

    // Only interested in the state register.
    if (reg == GIGE_REG_CAMERA_STATE)
    {
        TP(LVP_STATE, "%s: new_state=%u, live_up=%d", __PRETTY_FUNCTION__, val, cs_is_liveview_update);

        // What state are we being told about?
        switch (val)
        {
        case GIGE_CAMERA_STATE_LIVE_VIDEO:
        case GIGE_CAMERA_STATE_ARMED:
        case GIGE_CAMERA_STATE_TRIGGERED:
            cs_is_liveview_update = true;
            break;

        case GIGE_CAMERA_STATE_IDLE:
        case GIGE_CAMERA_STATE_REVIEW:
        case GIGE_CAMERA_STATE_SNAPSHOT:
        case GIGE_CAMERA_STATE_AUTOSAVE:
        case GIGE_CAMERA_STATE_DEARMING:
        default:
            cs_is_liveview_update = false;
            break;
        }

        // If we have a stream going, and there is a change, request an update.
        if (cs_is_liveview_update != old_update)
        {
            TP(LVP_STATE, "%s: wake-up the service to change state", __PRETTY_FUNCTION__);
            lws_cancel_service(ctx);
        }
    }
}

int fiWSSLiveViewProtocol::established(struct lws *wsi, WssSession **user)
{
    int retval = 0;

    TP(LVP_FUNC, "%s", __PRETTY_FUNCTION__);

    try
    {
        // Create a new live session for this listener.
        StreamingSession *s = new LiveViewSession(this, wsi);
        s->setState(StreamingSession::OFF);
        *((StreamingSession **)user) = s;
        sessions.push_back(s);

        // For live view, we do not want the oldest buffer. We
        // always want the latest.
        bq.set_oldest(false);

        if (sessions.size() == 1)
        {
            uint32_t camera_state = GIGE_CAMERA_STATE_IDLE;

            TP(LVP_STATE, "%s: First streaming session; Producer has %d Buffers", __PRETTY_FUNCTION__, bq.get_producer_buffer_count());
            streaming_protocol_trace = true;    // Deprecated in favor of trace.
            bq.buffer_queue_trace = true;   // Deprecated in favor of trace.

            // Read and proceses the camera state. We won't actually turn the
            // camera server on until we get the "ON" command.
            if (cs_get_reg_l(GIGE_REG_CAMERA_STATE, &camera_state) == 0)
                fi_log(LOG_ERR, "%s: cs_get_reg_l(GIGE_REG_CAMERA_STATE) failed during state initialization", __PRETTY_FUNCTION__);
            notify(GIGE_REG_CAMERA_STATE, camera_state);
        }
    }
    catch (std::exception& e)
    {
        retval = 1;
        fi_log(LOG_ERR, "Exception establishing Live View session: %s", e.what());
    }
    return retval;
}

int fiWSSLiveViewProtocol::closed(struct lws *wsi, WssSession **user)
{
    StreamingSession::SessionState session_state = StreamingSession::NONE;
    int retval = 0;

    TP(LVP_FUNC, "%s", __PRETTY_FUNCTION__);

    LiveViewSession *lvs = *(LiveViewSession **)user;
    if (lvs == NULL)
    {
        // Already closed, or never opened, or something
        fi_log(LOG_WARNING, "%s: Closing absent Live View session", __PRETTY_FUNCTION__);
        return retval;
    }

    session_state = lvs->getState();
    TP(LVP_STATE, "%s: session_state=%d", __PRETTY_FUNCTION__, (int)session_state);
    switch (session_state)
    {
    // For any active state, we need to decrement the listeners count:
    case StreamingSession::ON_FRAME:
    case StreamingSession::ON_IDLE:
    case StreamingSession::ON_REQ:
    case StreamingSession::OFF_REQ:
        listeners--;
        break;

    default:
    case StreamingSession::OFF:
    case StreamingSession::NONE:
        break; // Nothing to do here
    }

    if (listeners == 0)
    {
        fi_log(LOG_INFO, "%s: Close last Live View session %p/%d from state %d (now %d active listeners)",
               __PRETTY_FUNCTION__, (void *)lvs, lws_get_socket_fd(wsi), (int)session_state, listeners);
    }
    else
    {
        fi_log(LOG_INFO, "%s: Close Live View session %p/%d from state %d (now %d active listeners)",
               __PRETTY_FUNCTION__, (void *)lvs, lws_get_socket_fd(wsi), (int)session_state, listeners);
    }

    try
    {
        for (std::vector<WssSession *>::iterator i = sessions.begin(); i != sessions.end(); i++)
        {
            if (*i == lvs)
            {
                if (streaming_protocol_trace)
                {
                    TP(LVP_DETAIL, "%s: Erase Live View session %p/%d from sessions vector",
                       __PRETTY_FUNCTION__, (void *)lvs, lws_get_socket_fd(wsi));
                }
                sessions.erase(i);
                break;
            }
        }

        if (bq.isOnTxQueue(lvs))
        {
            // Check state vs. Tx queue status consistency
            switch (session_state)
            {
            case StreamingSession::OFF:
            case StreamingSession::ON_IDLE:
            case StreamingSession::ON_REQ:
            case StreamingSession::OFF_REQ:
            case StreamingSession::NONE:
                fi_log(LOG_ERR, "%s: Session %p/%d: On Tx queue in unexpected state (%d)",
                       __PRETTY_FUNCTION__, (void *)lvs, (int)lws_get_socket_fd(wsi), (int)session_state);
                break;

            case StreamingSession::ON_FRAME:
                break;
            }

            TP(LVP_BQ, "%s: Live View session %p/%d on Tx queue (remove it)",
               __PRETTY_FUNCTION__, (void *)lvs, lws_get_socket_fd(wsi));

            bq.removeFromTxQueue(lvs);
            if (bq.txQueueIsEmpty())
            {
                TP(LVP_BQ, "%s: Tx queue empty; %d listeners", __PRETTY_FUNCTION__, listeners);

                if (listeners > 0)
                {
                    bq.finished();
                    process_consumer();
                }
                else
                {
                    bq.flush();
                }
            }
        }
        else
        {
            // Check state vs. Tx queue status consistency
            switch (session_state)
            {
            default:
            case StreamingSession::OFF:
            case StreamingSession::ON_IDLE:
            case StreamingSession::ON_REQ:
            case StreamingSession::OFF_REQ:
            case StreamingSession::NONE:
                break;

            case StreamingSession::ON_FRAME:
                fi_log(LOG_ERR, "%s: Session %p/%d: Not on Tx queue in ON_FRAME state (%d)",
                       __PRETTY_FUNCTION__, (void *)lvs, (int)lws_get_socket_fd(wsi), (int)session_state);
                break;
            }

            TP(LVP_BQ, "%s: Live View session %p/%d not on Tx queue (in state %d)",
               __PRETTY_FUNCTION__, (void *)lvs, (int)lws_get_socket_fd(wsi), (int)session_state);
        }

        delete lvs;
    }
    catch (std::exception& e)
    {
        retval = 1;
        fi_log(LOG_ERR, "%s: Exception shutting down Live View session %p: %s", __PRETTY_FUNCTION__, (void *)lvs, e.what());
    }

    fi_log(LOG_INFO, "%s: Finished close callback (%d listeners; %ld sessions)", __PRETTY_FUNCTION__, listeners, sessions.size());
    return retval;
}

int fiWSSLiveViewProtocol::destroy(struct lws * /*wsi*/, WssSession * /*user*/)
{
    TP(LVP_FUNC, "%s", __PRETTY_FUNCTION__);
    return 0;
}

/**
 * ### Implementation notes:
 *
 * Ignoring error cases, each pass through this function will result
 * in one or two calls to lws_write(). There are five cases:
 *
 *   1
 *   - Write 8-bye binary message that carries next Image Frame Header (FIN)
 *   - Write single binary message that carries entire Image Frame data (FIN)
 *
 *   2
 *   - Write 8-bye binary message that carries next Image Frame Header (FIN)
 *   - Write binary message that carries initial N bytes of Image Frame data ( ! FIN)
 *
 *   3
 *   - Write Continuation message that carries next N bytes of Image Frame data ( ! FIN)
 *
 *   4
 *   - Write Continuation message that carries final N bytes of Image Frame data (FIN)
 *
 *   5
 *   - Write additional bytes of in-progress message (Binary or Continuation, with or without FIN)
 *
 * We distinguish them as follows:
 *
 * \verbatim
     bytes-so-far > 0:
         => case 5

     Buffer's START_OF_FRAME and END_OF_FRAME flags both set:
         => case 1

     Buffer's START_OF_FRAME flag set (END_OF_FRAME clear):
         => case 2

     Buffer's END_OF_FRAME flag set (START_OF_FRAME_CLEAR):
         => case 4

     Buffer's START_OF_FRAME and END_OF_FRAME flags both clear:
         => case 3
 * \endverbatim
 */
int fiWSSLiveViewProtocol::writable(struct lws *wsi, WssSession *user)
{
    uint32_t data_size = 0;
    StreamingProtocol::FRAME_HEADER_LWS header;
    LiveViewSession *s = (LiveViewSession *)user;

    TP(LVP_FUNC, "%s: state=%d (%s)", __PRETTY_FUNCTION__, (int)s->getState(), s->decodeState());

    // libwebsockets may issue spurious callbacks. Just let it go
    // if we get 'writable' callback and have nothing to write.
    if (!bq.isOnTxQueue(s))
    {
        TP(LVP_DETAIL, "%s: Spurious write callback for Live view protocol", __PRETTY_FUNCTION__);
        return 0;
    }

    if (s->getState() != StreamingSession::ON_FRAME)
    {
        if (s->getState() == StreamingSession::OFF_REQ)
        {
            TP(LVP_STATE, "%s: Found session %p/%d in state OFF_REQ while on TxQ",
               __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi));
        }
        else
        {
            fi_log(LOG_ERR, "%s: Found session %p/%d in unexpected state %d while on TxQ",
                   __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi), (int)s->getState());
        }
    }

    int soFar = s->getBytesSoFar();
    int total = bq.consumer_get_size();
    size_t bytesToWrite = total - soFar;
    uint32_t flags = bq.getCurrentConsumerFlags();
    bool sof = ((flags & StreamingProtocol::START_OF_FRAME) != 0);
    bool eof = ((flags & StreamingProtocol::END_OF_FRAME) != 0);
    int n = 0;

    if ((total == 0) || (bytesToWrite == 0))
    {
        // This is a logic error of some kind; should not be
        // on the Tx queue if Buffer is empty, or if we have
        // nothing to write.
        fi_log(LOG_ERR, "%s: Session %p/%d On Tx queue, but Buffer empty: total == %d, bytesToWrite == %ld",
               __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), total, bytesToWrite);
        TP(LVP_BQ, "%s: More info, soFar=%d, total=%d, flags=0x%x, sof=%d, eof=%d, bytesToWrite=%d", __PRETTY_FUNCTION__, soFar, total, flags, sof, eof, bytesToWrite);

        // Release this buffer and carry on.
        goto release_buf;
    }

    // Track information if we're tracing.
    TP(LVP_BQ, "%s: soFar=%d, total=%d, sof=%d, eof=%d, bytesToWrite=%d", __PRETTY_FUNCTION__, soFar, total, sof, eof, bytesToWrite);

    // Initialize our header structure.
    memset((void *)&header, 0, sizeof(header));

    // Is the write still buffered?
    if (lws_send_pipe_choked(wsi))
    {
        TP(LVP_DETAIL, "%s: Partial buffered, schedule writing", __PRETTY_FUNCTION__);
        s->scheduleWriteCallback();
        return 0;
    }

    if (soFar >  0)
    {
        // Case 5
        // Send additional Image data from in-progress Buffer
        // TODO: Do we need to match the Websockets Frame type
        // (LWS_WRITE_BINARY or LWS_WRITE_CONTINUATION) of the
        // initial write for this Buffer?
        TP(LVP_BQ, "%s: Session %p/%d: write additional data for frame (now %d listeners)",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), listeners);

        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, LWS_WRITE_BINARY);
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08x", __PRETTY_FUNCTION__, n);
            return -1;
        }
        soFar += n;
    }
    else if (sof && eof)
    {
        // Case 1
        data_size = bq.consumer_get_frame_data_size();
        header.h.magic = htonl(StreamingProtocol::MAGIC_ID);
        header.h.size = htonl(data_size);
        header.h.id = htonl(bq.consumer_get_frame_id());

        TP(LVP_BQ, "%s: Session %p/%d: write first (and last) data for frame (now %d listeners)",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), listeners);

        s->set_frame_length(data_size);

        // Send header as separate Websocket segment
        n = lws_write(wsi, (unsigned char *)(&header.h), sizeof(header.h), LWS_WRITE_BINARY);
        if (n != sizeof(header.h))
        {
            fi_log(LOG_ERR, "%s: Frame Header write failed: 0x%08x", __PRETTY_FUNCTION__, n);
            return -1;
        }

        // Is the write buffered?
        if (lws_send_pipe_choked(wsi))
        {
            TP(LVP_DETAIL, "%s: Partial buffered after header, schedule writing", __PRETTY_FUNCTION__);
            s->scheduleWriteCallback();
            return 0;
        }

        // Now start Image data
        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, LWS_WRITE_BINARY);
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08x", __PRETTY_FUNCTION__, n);
            return -1;
        }
        if (n < (int)bytesToWrite)
        {
            fi_log(LOG_ERR, "%s: Partial write: 0x%08x != 0x%08x", __PRETTY_FUNCTION__, n, bytesToWrite);
            return -1;
        }
        soFar += n;
    }
    else if (sof && (! eof))
    {
        // Case 2
        data_size = bq.consumer_get_frame_data_size();
        header.h.magic = htonl(StreamingProtocol::MAGIC_ID);
        header.h.size = htonl(data_size);
        header.h.id = htonl(bq.consumer_get_frame_id());

        TP(LVP_BQ, "%s: Session %p/%d: write first data for frame (now %d listeners)",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), listeners);

        s->set_frame_length(data_size);

        // Send header as separate Websocket segment
        n = lws_write(wsi, (unsigned char *)(&header.h), sizeof(header.h), LWS_WRITE_BINARY);
        if (n != sizeof(header.h))
        {
            fi_log(LOG_ERR, "%s: Frame Header write failed: 0x%08x", __PRETTY_FUNCTION__, n);
            return -1;
        }

        // Is the write buffered?
        if (lws_send_pipe_choked(wsi))
        {
            TP(LVP_DETAIL, "%s: Partial buffered after header, schedule writing", __PRETTY_FUNCTION__);
            s->scheduleWriteCallback();
            return 0;
        }

        // Now start Image data
        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, (enum lws_write_protocol)(LWS_WRITE_BINARY | LWS_WRITE_NO_FIN));
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08x", __PRETTY_FUNCTION__, n);
            return -1;
        }
        if (n < (int)bytesToWrite)
        {
            fi_log(LOG_ERR, "%s: Partial write: 0x%08x != 0x%08x", __PRETTY_FUNCTION__, n, bytesToWrite);
            return -1;
        }
        soFar += n;
    }
    else if ((!sof) && eof)
    {
        // Case 3:
        // Image data
        TP(LVP_BQ, "%s: Session %p/%d: write last data for frame (now %d listeners)",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), listeners);

        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, LWS_WRITE_CONTINUATION);
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08x", __PRETTY_FUNCTION__, n);
            return -1;
        }
        if (n < (int)bytesToWrite)
        {
            fi_log(LOG_ERR, "%s: Partial write: 0x%08x != 0x%08x", __PRETTY_FUNCTION__, n, bytesToWrite);
            return -1;
        }
        soFar += n;
    }
    else if ((!sof) && (!eof))
    {
        // Case 4:
        // Image data
        TP(LVP_BQ, "%s: Session %p/%d: write continuation for frame (now %d listeners)",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), listeners);

        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, (enum lws_write_protocol)(LWS_WRITE_CONTINUATION | LWS_WRITE_NO_FIN));
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08x", __PRETTY_FUNCTION__, n);
            return -1;
        }
        if (n < (int)bytesToWrite)
        {
            fi_log(LOG_ERR, "%s: Partial write: 0x%08x != 0x%08x", __PRETTY_FUNCTION__, n, bytesToWrite);
            return -1;
        }
        soFar += n;
    }
    else
    {
        // Logic problem....
        fi_log(LOG_ERR, "%s: Got unexpected condition in write callback: %d, %s, %s",
               __PRETTY_FUNCTION__, soFar, (sof ? "SOF" : "! SOF"), (eof ? "EOF" : "! EOF"));
        return -1;
    }

release_buf:
    s->setBytesSoFar(soFar);
    if (soFar >= total)
    {
        TP(LVP_BQ, "%s: Session %p/%d: finished writing data from Buffer",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));

        // Remove this session from Tx queue
        bq.removeFromTxQueue(s);

        if ((bq.getCurrentConsumerFlags() & StreamingProtocol::END_OF_FRAME) != 0)
        {
            switch (s->getState())
            {
            case StreamingSession::ON_FRAME:
                TP(LVP_STATE, "%s: Session %p/%d: This buffer had END_OF_FRAME (ON_FRAME -> ON_IDLE)",
                   __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));
                s->setState(StreamingSession::ON_IDLE);
                break;
            case StreamingSession::OFF_REQ:
                TP(LVP_STATE, "%s: OFF_REQ -> OFF", __PRETTY_FUNCTION__);
                break; // process_consumer() will complete transtition to StreamingSession::OFF
            default:
                fi_log(LOG_ERR, "%s: Found session %p/%d in unexpected state %d after completing Buffer",
                       __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi), (int)s->getState());
                break;
            }
        }
        if (bq.txQueueIsEmpty())
        {
            TP(LVP_BQ, "%s: Session %p/%d:Tx queue is empty", __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi));

            // Close out the buffer queue.
            bq.finished();

            // Anymore to add on our queue?
            process_consumer();
        }
    }
    else
    {
        TP(LVP_DETAIL, "%s: Schedule writing", __PRETTY_FUNCTION__);
        s->scheduleWriteCallback();
    }

    // All done.
    return 0;
}

/*!
 * @brief Process commands from listeners to turn the live view stream on or off.
 */
int fiWSSLiveViewProtocol::receive(struct lws *wsi, WssSession *s, uint8_t *in, size_t /*len*/)
{
    LiveViewSession *lvs = (LiveViewSession *)s;
    fi_json_parse_pkg my_json_data;  // This will contain the intrepreted JSON data
    int json_return_code = 0;        // This will contain the result of the JSON intrepretation

    TP(LVP_FUNC, "%s", __PRETTY_FUNCTION__);

    // Interpret the JSON data for client on/off.
    if (init_fi_json_parse_pkg(&my_json_data) == 1)
    {
        // Parse the JSON data.
        json_return_code = fi_json_parse((const char *)in, &my_json_data);
        if (json_return_code != 0)
        {
            fi_log(LOG_INFO, "%s: JSON parse: %s", __PRETTY_FUNCTION__, json_error_string(json_return_code));
            return 0; // Other than logging the anomaly, just ignore this input
        }

        // Find the value for the liveview request key.
        const char *liveview_state_ptr = (const char *)find_value_for_key(&my_json_data, (uint8_t *)"liveview");

        // Determine if action type of the request
        if (liveview_state_ptr != (char *)NULL)
        {
            // Is it on or off?
            if (strncmp((const char *)liveview_state_ptr, "ON", 2) == 0)
            {
                // Process based on the current state.
                switch (lvs->getState())
                {
                case StreamingSession::OFF:
                    listeners++; // Transition from "inactive" to "active"
                    lvs->setState(StreamingSession::ON_REQ);
                    TP(LVP_STATE, "%s: Start Live View on session %p/%d; OFF -> ON_REQ (now %d active listeners)",
                       __PRETTY_FUNCTION__, (void *)lvs, lws_get_socket_fd(wsi), listeners);

                    // If this is the first one, we need to get the camera server streaming.
                    if (listeners == 1)
                    {
                        uint32_t camera_state = GIGE_CAMERA_STATE_IDLE;

                        // Read and proceses the camera state. We won't actually turn the
                        // camera server on until we get the "ON" command.
                        if (cs_get_reg_l(GIGE_REG_CAMERA_STATE, &camera_state) == 0)
                            fi_log(LOG_ERR, "%s: cs_get_reg_l(GIGE_REG_CAMERA_STATE) failed during state initialization", __PRETTY_FUNCTION__);
                        notify(GIGE_REG_CAMERA_STATE, camera_state);
                        LOG("%s: cs_is_liveview=%d, cs_is_liveview_update=%d", __PRETTY_FUNCTION__, cs_is_liveview, cs_is_liveview_update);
                    }
                    break;

                default:
                case StreamingSession::ON_REQ:
                case StreamingSession::ON_FRAME:
                case StreamingSession::ON_IDLE:
                case StreamingSession::NONE:
                    TP(LVP_STATE, "%s: Nothing to do with ON command from %s", __PRETTY_FUNCTION__, lvs->decodeState());
                    break; // Nothing to do here

                case StreamingSession::OFF_REQ:
                    TP(LVP_STATE, "%s: Restart Live View on session %p/%d; OFF_REQ -> ON_FRAME (%d active listeners)",
                       __PRETTY_FUNCTION__, (void *)lvs, lws_get_socket_fd(wsi), listeners);
                    lvs->setState(StreamingSession::ON_FRAME);
                    break;
                }
            }
            else if (strncmp((const char *)liveview_state_ptr, "OFF", 3) == 0)
            {
                // Process based on the current state.
                switch (lvs->getState())
                {
                case StreamingSession::ON_FRAME:
                    TP(LVP_STATE, "%s: Request stop Live View on session %p/%d; ON_FRAME -> OFF_REQ (%d active listeners)",
                       __PRETTY_FUNCTION__, (void *)lvs, lws_get_socket_fd(wsi), listeners);
                    lvs->setState(StreamingSession::OFF_REQ);
                    break;

                case StreamingSession::ON_IDLE:
                case StreamingSession::ON_REQ:
                    if (bq.isOnTxQueue(lvs))
                    {
                        fi_log(LOG_ERR, "%s: Found session %p/%d on TxQ while in state ON_IDLE/ON_REQ (%d active listeners)",
                               __PRETTY_FUNCTION__, (void *)lvs, lws_get_socket_fd(wsi), listeners);
                    }

                    listeners--; // Transition from "Active" to "Inactive" state
                    lvs->setState(StreamingSession::OFF);

                    TP(LVP_STATE, "%s: Stop Live View on session %p/%d; ON_IDLE/ON_REQ -> OFF (%d now active listeners)",
                       __PRETTY_FUNCTION__, (void *)lvs, lws_get_socket_fd(wsi), listeners);
                    break;

                case StreamingSession::OFF:
                case StreamingSession::OFF_REQ:
                case StreamingSession::NONE:
                default:
                    TP(LVP_STATE, "%s: Nothing to do with OFF command from %s", __PRETTY_FUNCTION__, lvs->decodeState());
                    break; // Nothing to do here
                }
            }
            else
            {
                fi_log(LOG_INFO, "%s: JSON parse: %s", __PRETTY_FUNCTION__, json_error_string(json_return_code));
                return 0; // Other than logging the anomaly, just ignore this input
            }
        }

        // Free allocated memory items in my_json_data
        free_fi_json_parse_pkg(&my_json_data);
    }

    // All done.
    return 0;
}
