/*
 * fi_playback_protocol.cxx
 *
 *  Created on: Jan 25, 2016
 *      Author: eric
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <exception>
#include <stdio.h>
#include <string.h>

#include <fastec/gige_msg.h>
#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

#include "fi_playback_session.h"
#include "fi_playback_protocol.h"
#include "fi_buffer_q.h"
#include "fcp.h"

//--------------------------------------------------------------------
// Constant, Type and Macro definitions
//--------------------------------------------------------------------

enum PBP_TRACE
{
    PBP_FUNC    = 0x00000010,
    PBP_STATE   = 0x00000020,
    PBP_BQ  = 0x00000040,
    PBP_DETAIL  = 0x00000080
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

static uint32_t traceControl =
    PBP_FUNC    |
    PBP_STATE   |
    PBP_BQ      |
    PBP_DETAIL  |
//        TRACE_0;        // Set our trace level to TRACE_0 to trace.
    TRACE_3;        // Set our trace level to TRACE_3 to not trace.

/**
 * @brief Trampoline for custom signaling hook
 *
 * This just provides a static function so we can pass an address
 * to cs_register_callback(). CS thread will use this to tell us
 * about changes in the Camera Server's FCP registers.
 */
static void playback_callback(void *arg, uint32_t reg, uint32_t val)
{
    static_cast<fiWSSPlaybackProtocol *>(arg)->notify(reg, val);
}

fiWSSPlaybackProtocol::fiWSSPlaybackProtocol(struct lws_protocols *p) :
    StreamingProtocol(p),
    playback_state_status(NEED_INIT),
    playable(false),
    playing(false),
    gige_reg_camera_state(0),
    gige_reg_pb_control(0),
    gige_reg_part_state(0)
{
    int err = 0;

    TP(PBP_FUNC, "%s", __PRETTY_FUNCTION__);

    p->name = "fi-playback-protocol";
    p->rx_buffer_size = 48 * 1024;

    if ((err = pthread_mutex_init(&playback_state_mutex, NULL)) != 0)
    {
        fi_log(LOG_ERR, "%s: pthread_mutex_init failed. %s", __PRETTY_FUNCTION__, strerror(err));
    }

    cs_register_callback(playback_callback, this);
}

fiWSSPlaybackProtocol::~fiWSSPlaybackProtocol()
{
    TP(PBP_FUNC, "%s", __PRETTY_FUNCTION__);

    for (WssProtocol::iterator i = WssProtocol::begin();
         i != WssProtocol::end();
         i++)
    {
        delete *i;
    }
    pthread_mutex_destroy(&playback_state_mutex);
}

void fiWSSPlaybackProtocol::process_consumer()
{
    bool new_playable = false;
    bool new_playing = false;
    int cs_ret = 0;
    uint32_t new_localstream_control = 0xFFFFFFFF;

    // Handle playback state changes
    pthread_mutex_lock(&playback_state_mutex);

    if (playback_state_status == NEED_INIT)
    {
        // first time through; get current reg values
        pthread_mutex_unlock(&playback_state_mutex);

        if (cs_get_reg_l(GIGE_REG_CAMERA_STATE, &gige_reg_camera_state) == 0)
            fi_log(LOG_ERR, "%s: cs_get_reg_l(GIGE_REG_CAMERA_STATE) failed during state initialization", __PRETTY_FUNCTION__);
        if (cs_get_reg_l(GIGE_REG_PB_CONTROL, &gige_reg_pb_control) == 0)
            fi_log(LOG_ERR, "%s: cs_get_reg_l(GIGE_REG_PB_CONTROL) failed during state initialization", __PRETTY_FUNCTION__);
        if (cs_get_reg_l(GIGE_REG_PART_STATE, &gige_reg_part_state) == 0)
            fi_log(LOG_ERR, "%s: cs_get_reg_l(GIGE_REG_PART_STATE) failed during state initialization", __PRETTY_FUNCTION__);

        pthread_mutex_lock(&playback_state_mutex);

        TP(PBP_STATE, "%s: NEED_INIT, camera_state=0x%x, pb_control=0x%x, part_state=0x%x",
           __PRETTY_FUNCTION__, gige_reg_camera_state, gige_reg_pb_control, gige_reg_part_state);
        playback_state_status = INIT_DONE;
    }

    if ((playback_state_status == CHANGE_PENDING) || (playback_state_status == INIT_DONE))
    {
        // Compute new state based on registers
        new_playable = ((gige_reg_camera_state == GIGE_CAMERA_STATE_REVIEW) &&
                        (gige_reg_part_state == GIGE_PART_STATE_COMPLETE));
        new_playing = ((gige_reg_pb_control & GIGE_PBC_ENABLE) != 0);
        TP(PBP_STATE, "%s: playable=%d, new_playable=%d, new_playing=%d",
           __PRETTY_FUNCTION__, playable, new_playable, new_playing);

        if (playable != new_playable)
        {
            playable = new_playable;
            playing = new_playing;
            if (playable)
            {
                // Transition into Playable
                if (playing)
                {
                    if (listeners > 0)
                    {
                        new_localstream_control = GIGE_LOCALSTREAM_PLAYBACK;
                    }
                }
                else
                {
                    if (listeners > 0)
                    {
                        new_localstream_control = GIGE_LOCALSTREAM_SINGLESTEP;
                    }
                }
            }
            else
            {
                if (listeners > 0)
                {
                    new_localstream_control = GIGE_LOCALSTREAM_STOP;
                }
            }
        }

        if (playing != new_playing)
        {
            playing = new_playing;
            if (playable && (listeners > 0))
            {
                if (playing)
                {
                    new_localstream_control = GIGE_LOCALSTREAM_PLAYBACK;
                }
                else
                {
                    new_localstream_control = GIGE_LOCALSTREAM_SINGLESTEP;
                }
            }
        }

        playback_state_status = OK;
        pthread_mutex_unlock(&playback_state_mutex);

        TP(PBP_STATE, "%s: new_localstream_control=0x%x", __PRETTY_FUNCTION__, new_localstream_control);
        if (new_localstream_control != 0xFFFFFFFF)
        {
            if (cs_set_reg_l(GIGE_REG_LOCALSTREAM_CONTROL, new_localstream_control, &cs_ret) == 0)
            {
                fi_log(LOG_ERR, "%s: Could not set GIGE_REG_LOCALSTREAM_CONTROL (%d)", __PRETTY_FUNCTION__, cs_ret);
            }
        }
    }
    else
    {
        pthread_mutex_unlock(&playback_state_mutex);
    }

    // Call base-class method to do normal processing
    StreamingProtocol::process_consumer();
}

/**
 * @brief Release the current image memory buffer
 */
void fiWSSPlaybackProtocol::release_image_memory_buffer(void *shmem)
{
    TP(PBP_FUNC, "%s: buf_id=%d", __PRETTY_FUNCTION__, bq.buf_id);

    // Release the LSSM buffer we're finished processing, if we're
    // currently holding one.
    if (bq.buf_id >= 0)
    {
        lssm_lock(shmem);
        lssm_release_buffer_l(shmem, bq.buf_id);
        lssm_server_signal_l(shmem);
        bq.buf_id = -1;
        lssm_unlock(shmem);
    }
}

//! ### Implemenation notes
//!
//! This function executes on the Camera Server event thread
//! (fcp.cxx::event_thread()). It picks up FCP register
//! changes that affect playback operation, and signals the
//! playback protocol, which executes on the _libwebsockets_
//! thread, when an interesting state change happens. The
//! fiWSSPlaybackProtocol::process_consumer() method takes
//! action based on the state changes.
//!
void fiWSSPlaybackProtocol::notify(uint32_t reg, uint32_t val)
{
    // Make sure someone cares.
    if (ctx == NULL)
        return;

    if ((reg == GIGE_REG_PB_CONTROL)   ||
        (reg == GIGE_REG_CAMERA_STATE) ||
        (reg == GIGE_REG_PART_STATE))
    {
        TP(PBP_STATE, "%s: reg=0x%x, val=0%x", __PRETTY_FUNCTION__, reg, val);

        pthread_mutex_lock(&playback_state_mutex);

        if ((playback_state_status == OK) || (playback_state_status == CHANGE_PENDING))
        {
            switch (reg)
            {
            case GIGE_REG_PB_CONTROL:
                if ((gige_reg_pb_control & GIGE_PBC_ENABLE) != (val & GIGE_PBC_ENABLE))
                {
                    gige_reg_pb_control = val;
                    playback_state_status = CHANGE_PENDING;
                }
                break;

            case GIGE_REG_CAMERA_STATE:
                if (gige_reg_camera_state != val)
                {
                    gige_reg_camera_state = val;
                    playback_state_status = CHANGE_PENDING;
                }
                break;

            case GIGE_REG_PART_STATE:
                if (gige_reg_part_state != val)
                {
                    gige_reg_part_state = val;
                    playback_state_status = CHANGE_PENDING;
                }
                break;

            default:
                break;
            }
        }

        // If there is a change, request an update.
        if (playback_state_status == CHANGE_PENDING)
        {
            TP(PBP_STATE, "%s: wake-up the service to change state", __PRETTY_FUNCTION__);
            lws_cancel_service(ctx);
        }

        pthread_mutex_unlock(&playback_state_mutex);
    }
}

int fiWSSPlaybackProtocol::established(struct lws *wsi, WssSession **user)
{
    uint32_t new_localstream_control = 0;
    int cs_ret = 0;
    int retval = 0;

    TP(PBP_FUNC, "%s", __PRETTY_FUNCTION__);

    try
    {
        StreamingSession *s = new PlaybackSession(this, wsi);
        listeners++;
        s->setState(StreamingSession::ON_REQ);
        *((StreamingSession **)user) = s;
        sessions.push_back(s);

        bq.set_oldest(false);

        if (sessions.size() == 1)
        {
            TP(PBP_STATE, "%s: First streaming session; Producer has %d Buffers", __PRETTY_FUNCTION__, bq.get_producer_buffer_count());
            streaming_protocol_trace = true;
            bq.buffer_queue_trace = true;

            // first listener: May need to start local stream
            if (playable)
            {
                if (playing)
                    new_localstream_control = GIGE_LOCALSTREAM_PLAYBACK;
                else
                    new_localstream_control = GIGE_LOCALSTREAM_SINGLESTEP;

                TP(PBP_STATE, "%s: Request state change to %d", __PRETTY_FUNCTION__, new_localstream_control);
                if (cs_set_reg_l(GIGE_REG_LOCALSTREAM_CONTROL, new_localstream_control, &cs_ret) == 0)
                {
                    fi_log(LOG_ERR, "%s: Could not set GIGE_REG_LOCALSTREAM_CONTROL (%d)", __PRETTY_FUNCTION__, cs_ret);
                }
            }
        }
    }
    catch (std::exception& e)
    {
        retval = 1;
        fi_log(LOG_ERR, "%s: Exception establishing Playback session: %s", __PRETTY_FUNCTION__, e.what());
    }
    return retval;
}

int fiWSSPlaybackProtocol::closed(struct lws *wsi, WssSession **user)
{
    StreamingSession::SessionState session_state = StreamingSession::NONE;
    int retval = 0;
    int cs_ret = 0;
    PlaybackSession *s = *(PlaybackSession **)user;

    TP(PBP_FUNC, "%s", __PRETTY_FUNCTION__);

    if (s == NULL)
    {
        // Already closed, or never opened, or something
        fi_log(LOG_WARNING, "%s: Closing absent Playback session", __PRETTY_FUNCTION__);
        return retval;
    }

    session_state = s->getState();
    TP(PBP_STATE, "%s: session_state=%d", __PRETTY_FUNCTION__, (int)session_state);
    switch (session_state)
    {
    // For any active state, we need to decrement the listeners count:
    case StreamingSession::ON_FRAME:
    case StreamingSession::ON_IDLE:
    case StreamingSession::ON_REQ:
    case StreamingSession::OFF_REQ:
        listeners--;
        break;

    case StreamingSession::OFF:
    case StreamingSession::NONE:
        break; // Nothing to do here
    }

    if (listeners == 0)
    {
        fi_log(LOG_INFO, "%s: Close last Playback session %p/%d from state %d (now %d active listeners)",
               __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), (int)session_state, listeners);

        // listeners has gone to zero: may need to stop local stream
        if (playable)
        {
            TP(PBP_STATE, "%s: Request state change to GIGE_LOCALSTREAM_STOP", __PRETTY_FUNCTION__);
            if (cs_set_reg_l(GIGE_REG_LOCALSTREAM_CONTROL, GIGE_LOCALSTREAM_STOP, &cs_ret) == 0)
            {
                fi_log(LOG_ERR, "%s: Could not set GIGE_REG_LOCALSTREAM_CONTROL (%d)", __PRETTY_FUNCTION__, cs_ret);
            }
        }
    }
    else
    {
        fi_log(LOG_INFO, "%s: Close Playback session %p/%d from state %d (now %d active listeners)",
               __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), (int)session_state, listeners);
    }

    try
    {
        for (std::vector<WssSession *>::iterator i = sessions.begin(); i != sessions.end(); i++)
        {
            if (*i == s)
            {
                TP(PBP_DETAIL, "%s: Erase Playback session %p/%d from sessions vector",
                   __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi));
                sessions.erase(i);
                break;
            }
        }

        if (bq.isOnTxQueue(s))
        {
            // Check state vs. Tx queue status consistency
            switch (session_state)
            {
            case StreamingSession::OFF:
            case StreamingSession::ON_IDLE:
            case StreamingSession::ON_REQ:
            case StreamingSession::OFF_REQ:
            case StreamingSession::NONE:
                fi_log(LOG_ERR, "%s: Session %p/%d: On Tx queue in unexpected state (%d)",
                       __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi), (int)session_state);
                break;

            case StreamingSession::ON_FRAME:
                break;
            }

            TP(PBP_BQ, "%s: Playback session %p/%d on Tx queue (remove it)",
               __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi));

            bq.removeFromTxQueue(s);
            if (bq.txQueueIsEmpty())
            {
                TP(PBP_BQ, "%s: Tx queue empty; %d listeners", __PRETTY_FUNCTION__, listeners);
                if (listeners > 0)
                {
                    bq.finished();
                    process_consumer();
                }
                else
                {
                    bq.flush();
                }
            }
        }
        else
        {
            // Check state vs. Tx queue status consistency
            switch (session_state)
            {
            case StreamingSession::OFF:
            case StreamingSession::ON_IDLE:
            case StreamingSession::ON_REQ:
            case StreamingSession::OFF_REQ:
            case StreamingSession::NONE:
                break;

            case StreamingSession::ON_FRAME:
                fi_log(LOG_ERR, "%s: Session %p/%d: Not on Tx queue in ON_FRAME state (%d)",
                       __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi), (int)session_state);
                break;
            }

            TP(PBP_BQ, "%s: Playback session %p/%d not on Tx queue (in state %d)",
               __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi), (int)session_state);
        }

        delete s;
    }
    catch (std::exception& e)
    {
        retval = 1;
        fi_log(LOG_ERR, "%s: Exception shutting down Playback session: %p: %s", __PRETTY_FUNCTION__, (void *)s, e.what());
    }

    fi_log(LOG_INFO, "%s: Finished close callback (%d listeners; %ld sessions)", __PRETTY_FUNCTION__, listeners, sessions.size());
    return retval;
}

int fiWSSPlaybackProtocol::destroy(struct lws * /*wsi*/, WssSession * /*user*/)
{
    TP(PBP_FUNC, "%s", __PRETTY_FUNCTION__);
    return 0;
}

/**
 * Implementation notes:
 *
 * Ignoring error cases, each pass through this function will result
 * in one or two calls to lws_write(). There are five cases:
 *
 *   1
 *   - Write 8-bye binary message that carries next Image Frame Header (FIN)
 *   - Write single binary message that carries entire Image Frame data (FIN)
 *
 *   2
 *   - Write 8-bye binary message that carries next Image Frame Header (FIN)
 *   - Write binary message that carries initial N bytes of Image Frame data ( ! FIN)
 *
 *   3
 *   - Write Continuation message that carries next N bytes of Image Frame data ( ! FIN)
 *
 *   4
 *   - Write Continuation message that carries final N bytes of Image Frame data (FIN)
 *
 *   5
 *   - Write additional bytes of in-progress message (Binary or Continuation, with or without FIN)
 *
 * We distinguish them as follows:
 *
 * \verbatim
     bytes-so-far > 0:
         => case 5

     Buffer's START_OF_FRAME and END_OF_FRAME flags both set:
         => case 1

     Buffer's START_OF_FRAME flag set (END_OF_FRAME clear):
         => case 2

     Buffer's END_OF_FRAME flag set (START_OF_FRAME_CLEAR):
         => case 4

     Buffer's START_OF_FRAME and END_OF_FRAME flags both clear:
         => case 3
 * \endverbatim
 */
int fiWSSPlaybackProtocol::writable(struct lws *wsi, WssSession *user)
{
    uint32_t data_size = 0;
    StreamingProtocol::FRAME_HEADER_LWS header;
    PlaybackSession *s = (PlaybackSession *)user;

    TP(PBP_FUNC, "%s: state=%d (%s)", __PRETTY_FUNCTION__, (int)s->getState(), s->decodeState());

    // libwebsockets may issue spurious callbacks. Just let it go
    // if we get 'writable' callback and have nothing to write.
    if (!bq.isOnTxQueue(s))
    {
        TP(PBP_DETAIL, "%s: Spurious write callback for Playback protocol", __PRETTY_FUNCTION__);
        return 0;
    }

    if (s->getState() != StreamingSession::ON_FRAME)
    {
        if (s->getState() == StreamingSession::OFF_REQ)
        {
            TP(PBP_STATE, "%s: Found session %p/%d in state OFF_REQ while on TxQ",
               __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi));
        }
        else
        {
            fi_log(LOG_ERR, "%s: Found session %p/%d in unexpected state %d while on TxQ",
                   __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi), (int)s->getState());
        }
    }

    int soFar = s->getBytesSoFar();
    int total = bq.consumer_get_size();
    size_t bytesToWrite = total - soFar;
    uint32_t flags = bq.getCurrentConsumerFlags();
    bool sof = ((flags & StreamingProtocol::START_OF_FRAME) != 0);
    bool eof = ((flags & StreamingProtocol::END_OF_FRAME) != 0);
    int n = 0;

    if ((total == 0) || (bytesToWrite == 0))
    {
        // This is a logic error of some kind; should not be
        // on the Tx queue if Buffer is empty, or if we have
        // nothing to write.
        fi_log(LOG_ERR, "%s: Session %p/%d On Tx queue, but Buffer empty: total == %d, bytesToWrite == %ld (forcing close)",
               __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), total, bytesToWrite);
        return -1;
    }

    // Track information if we're tracing.
    TP(PBP_BQ, "%s: id=%d, soFar=%d, total=%d, sof=%d, eof=%d, bytesToWrite=%d", __PRETTY_FUNCTION__, bq.consumer_index(), soFar, total, sof, eof, bytesToWrite);

    // Clear our header structure.
    memset((void *)&header, 0, sizeof(header));

    // Is the write still buffered?
    if (lws_send_pipe_choked(wsi))
    {
        TP(PBP_DETAIL, "%s: Partial buffered, schedule writing", __PRETTY_FUNCTION__);
        s->scheduleWriteCallback();
        return 0;
    }

    if (soFar > 0)
    {
        // Case 5
        // Send additional Image data from in-progress Buffer
        // TODO: Do we need to match the Websockets Frame type
        // (LWS_WRITE_BINARY or LWS_WRITE_CONTINUATION) of the
        // initial write for this Buffer?
        TP(PBP_BQ, "%s: Session %p/%d: write additional data for frame (now %d listeners)",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), listeners);

        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, LWS_WRITE_BINARY);
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }
        soFar += n;
    }
    else if (sof && eof)
    {
        // Case 1
        data_size = bq.consumer_get_frame_data_size();
        header.h.magic = htonl(StreamingProtocol::MAGIC_ID);
        header.h.size = htonl(data_size);
        header.h.id = htonl(bq.consumer_get_frame_id());

        TP(PBP_BQ, "%s: Session %p/%d: write first (and last) data for frame (now %d listeners)",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), listeners);

        s->set_frame_length(data_size);

        // Send header as separate Websocket segment
        n = lws_write(wsi, (unsigned char *)(&header.h), sizeof(header.h), LWS_WRITE_BINARY);
        if (n != sizeof(header.h))
        {
            fi_log(LOG_ERR, "%s: Frame Header write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }

        // Is the write buffered?
        if (lws_send_pipe_choked(wsi))
        {
            TP(PBP_DETAIL, "%s: Partial buffered, schedule writing", __PRETTY_FUNCTION__);
            s->scheduleWriteCallback();
            return 0;
        }

        // Now start Image data
        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, LWS_WRITE_BINARY);
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }
        if (n < (int)bytesToWrite)
        {
            fi_log(LOG_ERR, "%s: Partial write: 0x%08x != 0x%08x", __PRETTY_FUNCTION__, n, bytesToWrite);
            return -1;
        }
        soFar += n;
    }
    else if (sof && (!eof))
    {
        uint32_t frame_id;

        // Case 2
        header.h.magic = htonl(StreamingProtocol::MAGIC_ID);
        data_size = bq.consumer_get_frame_data_size();
        header.h.size = htonl(data_size);
        frame_id = bq.consumer_get_frame_id();
        header.h.id = htonl(frame_id);

        TP(PBP_BQ, "%s: Session %p/%d: write first data for frame %d (size=%d) (now %d listeners)",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), frame_id, data_size, listeners);

        // Tell the session the total length for the frame.
        s->set_frame_length(data_size);

        // Send header as separate Websocket segment
        n = lws_write(wsi, (unsigned char *)(&header.h), sizeof(header.h), LWS_WRITE_BINARY);
        if (n != sizeof(header.h))
        {
            fi_log(LOG_ERR, "%s: Frame Header write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }

        // Is the write buffered?
        if (lws_send_pipe_choked(wsi))
        {
            TP(PBP_DETAIL, "%s: Partial buffered after header, schedule writing", __PRETTY_FUNCTION__);
            s->scheduleWriteCallback();
            return 0;
        }

        // Now start Image data
        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, (enum lws_write_protocol)(LWS_WRITE_BINARY | LWS_WRITE_NO_FIN));
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }
        if (n < (int)bytesToWrite)
        {
            fi_log(LOG_ERR, "%s: Partial write: 0x%08x != 0x%08x", __PRETTY_FUNCTION__, n, bytesToWrite);
            return -1;
        }
        soFar += n;
    }
    else if ((!sof) && eof)
    {
        // Case 3:
        // Image data
        TP(PBP_BQ, "%s: Session %p/%d: write last data for frame (now %d listeners)",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), listeners);

        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, LWS_WRITE_CONTINUATION);
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }
        if (n < (int)bytesToWrite)
        {
            fi_log(LOG_ERR, "%s: Partial write: 0x%08x != 0x%08x", __PRETTY_FUNCTION__, n, bytesToWrite);
            return -1;
        }
        soFar += n;
    }
    else if (!sof && !eof)
    {
        // Case 4:
        // Image data
        TP(PBP_BQ, "%s: Session %p/%d: write continuation for frame (now %d listeners)",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi), listeners);

        n = lws_write(wsi, (unsigned char *)bq.consumerGetDataPtr(soFar), bytesToWrite, (enum lws_write_protocol)(LWS_WRITE_CONTINUATION | LWS_WRITE_NO_FIN));
        if (n < 0)
        {
            fi_log(LOG_ERR, "%s: Frame data write failed: 0x%08X", __PRETTY_FUNCTION__, n);
            return -1;
        }
        if (n < (int)bytesToWrite)
        {
            fi_log(LOG_ERR, "%s: Partial write: 0x%08x != 0x%08x", __PRETTY_FUNCTION__, n, bytesToWrite);
            return -1;
        }
        soFar += n;
    }
    else
    {
        // Logic problem....
        fi_log(LOG_ERR, "%s: Got unexpected condition in write callback: %d, %s, %s",
               __PRETTY_FUNCTION__, soFar, (sof ? "SOF" : "! SOF"), (eof ? "EOF" : "! EOF"));
        return -1;
    }

    s->setBytesSoFar(soFar);
    if (soFar >= total)
    {
        TP(PBP_BQ, "%s: Session %p/%d: finished writing data from Buffer",
           __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));

        // Remove this session from Tx queue
        bq.removeFromTxQueue(s);

        if ((bq.getCurrentConsumerFlags() & StreamingProtocol::END_OF_FRAME) != 0)
        {
            switch (s->getState())
            {
            case StreamingSession::ON_FRAME:
                TP(PBP_STATE, "%s: Session %p/%d: This buffer had END_OF_FRAME (ON_FRAME -> ON_IDLE)",
                   __PRETTY_FUNCTION__, (void *)s, lws_get_socket_fd(wsi));
                s->setState(StreamingSession::ON_IDLE);
                break;

            case StreamingSession::OFF_REQ:
                TP(PBP_STATE, "%s: OFF_REQ -> OFF", __PRETTY_FUNCTION__);
                break; // process_consumer() will complete transtition to StreamingSession::OFF

            default:
                fi_log(LOG_ERR, "%s: Found session %p/%d in unexpected state %d after completing Buffer",
                       __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi), (int)s->getState());
                break;
            }

        }
        if (bq.txQueueIsEmpty())
        {
            TP(PBP_BQ, "%s: Session %p/%d:Tx queue is empty", __PRETTY_FUNCTION__, (void *)s, (int)lws_get_socket_fd(wsi));

            // Close out the buffer queue.
            bq.finished();

            // Anymore to add on our queue?
            process_consumer();
        }
    }
    else
    {
        TP(PBP_DETAIL, "%s: Schedule writing", __PRETTY_FUNCTION__);
        s->scheduleWriteCallback();
    }

    // All done.
    return 0;
}

int fiWSSPlaybackProtocol::receive(struct lws * /*wsi*/, WssSession * /*s*/, uint8_t * /*in*/, size_t /*len*/)
{
    TP(PBP_FUNC, "%s", __PRETTY_FUNCTION__);
    return 0;
}
