/*
 * fi_wss_streamer.cpp
 *
 *  Created on: Jan 25, 2016
 *      Author: eric
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

#include "fi_streaming_protocol.h"
#include "fi_streaming_session.h"

//--------------------------------------------------------------------
// Constant, Type and Macro definitions
//--------------------------------------------------------------------

enum SP_TRACE
{
    SP_FUNC     = 0x00000010,
    SP_STATE    = 0x00000020,
    SP_BQ   = 0x00000040
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

static uint32_t traceControl =
//        SP_FUNC       |
    SP_STATE    |
    SP_BQ       |
//        TRACE_0;        // Set our trace level to TRACE_0 to trace.
    TRACE_3;        // Set our trace level to TRACE_3 to not trace.

StreamingProtocol::StreamingProtocol(struct lws_protocols *p) :
    WssProtocol(p),
    streaming_protocol_trace(false),
    listeners(0)
{
    TP(SP_FUNC, "%s", __PRETTY_FUNCTION__);

    p->per_session_data_size = sizeof(StreamingSession *);
    sessions.clear();
}

StreamingProtocol::~StreamingProtocol()
{
    TP(SP_FUNC, "%s", __PRETTY_FUNCTION__);
}

BufferQueue& StreamingProtocol::buffer_q()
{
    TP(SP_FUNC, "%s", __PRETTY_FUNCTION__);
    return bq;
}

bool StreamingProtocol::consumer_queue_has_data()
{
    TP(SP_FUNC, "%s", __PRETTY_FUNCTION__);
    return bq.consumerOwnsCurrent();
}

bool StreamingProtocol::producer_queue_has_buffer()
{
    TP(SP_FUNC, "%s", __PRETTY_FUNCTION__);
    return bq.producer_owns_current();
}

bool StreamingProtocol::tx_queue_is_empty()
{
    TP(SP_FUNC, "%s", __PRETTY_FUNCTION__);
    return bq.txQueueIsEmpty();
}

bool StreamingProtocol::get_bq_fault()
{
    TP(SP_FUNC, "%s", __PRETTY_FUNCTION__);
    return bq.get_fault();
}

void StreamingProtocol::flush_consumer_queue()
{
    TP(SP_FUNC, "%s", __PRETTY_FUNCTION__);
    for (StreamingProtocol::iterator i = begin(); i != end(); i++)
    {
        StreamingSession *s = (StreamingSession *)(*i);
        s->tx_flush();
        bq.removeFromTxQueue(s);
    }
    if (! bq.txQueueIsEmpty())
        fi_log(LOG_ERR, "%s: Tx queue should be empty here; flushing anyway", __PRETTY_FUNCTION__);
    bq.flush();
}

/**
 * @brief Place streaming sessions onto session transmit queues
 *
 * This method examines a protocol's Consumer queue.
 * If the Consumer queue is not empty (has a Consumer-owned Buffer at its
 * head), then the method examines the corresponding protocol's session
 * transmit queue. If the transmit queue is empty, then currently active
 * sessions belonging to the protocol are placed into the session transmit
 * queue. The protocol's write() callback is scheduled so that the
 * protocol can transmit data from the Buffer in the libwebsockets service
 * loop.
 */
void StreamingProtocol::process_consumer()
{
    StreamingSession::SessionState state = StreamingSession::NONE;

    if (bq.get_fault())
    {
        TP(SP_BQ, "%s: Process fault indication", __PRETTY_FUNCTION__);
        bq.set_fault(false);   // Clear fault indication, now that we've seen it
        flush_consumer_queue();

        // Walk the session list and attempt to shut down each session
        for (StreamingProtocol::iterator i = begin(); i != end(); i++)
        {
            StreamingSession *s = (StreamingSession *)(*i);
            TP(SP_BQ, "%s: Shut down session due to fault", __PRETTY_FUNCTION__);
            s->tx_flush();
        }

        // We're done.
        return;
    }

    while (bq.consumerOwnsCurrent() && bq.txQueueIsEmpty())
    {
        if (sessions.size() == 0)
        {
            TP(SP_BQ, "%s: Flush BufferQueue (empty session list)", __PRETTY_FUNCTION__);
            flush_consumer_queue();
        }
        else
        {
            // Walk the session list and place sessions onto the Tx queue
            for (StreamingProtocol::iterator i = begin(); i != end(); i++)
            {
                StreamingSession *s = (StreamingSession *)(*i);
                state = s->getState();
                switch (state)
                {
                case StreamingSession::ON_FRAME:
                case StreamingSession::ON_IDLE:
                    TP(SP_STATE, "%s: Session %p, ON_FRAME/ON_IDLE -> ON_FRAME", __PRETTY_FUNCTION__, (void *)s);
                    bq.fill(s);
                    s->setState(StreamingSession::ON_FRAME);
                    break;
                case StreamingSession::OFF_REQ:
                    if ((bq.getCurrentConsumerFlags() & StreamingProtocol::END_OF_FRAME) != 0)
                    {
                        s->setState(StreamingSession::OFF);

                        // Transition from "Active" to "Inactive" state
                        listeners--;
                        TP(SP_STATE, "%s: Honor OFF_REQ-> OFF for session %p (now %d active listeners)",
                           __PRETTY_FUNCTION__, (void *)s, listeners);
                    }
                    else
                    {
                        TP(SP_STATE, "%s: OFF_REQ still filling buffers", __PRETTY_FUNCTION__);
                        bq.fill(s);
                    }
                    break;
                case StreamingSession::ON_REQ:
                    // Is this the start of the frame?
                    if ((bq.getCurrentConsumerFlags() & StreamingProtocol::START_OF_FRAME) != 0)
                    {
                        TP(SP_STATE, "%s: Honor ON_REQ -> ON_FRAME for session %p", __PRETTY_FUNCTION__, (void *)s);
                        s->setState(StreamingSession::ON_FRAME);
                        bq.fill(s);
                    }
                    break;
                case StreamingSession::OFF:
                    TP(SP_STATE, "%s: OFF state.", __PRETTY_FUNCTION__);
                    break; // Just skip this one; it's off
                default:
                    fi_log(LOG_ERR, "%s: Found session %p in state %d (%s) (ignoring)", __PRETTY_FUNCTION__, (void *)s, (int)state, s->decodeState());
                    break;
                }
            }

            if (bq.txQueueIsEmpty())
            {
                // We've walked through the list of streaming sessions,
                // but found none that could use the current consumer
                // Buffer. (Perhaps all are waiting for start-of-frame,
                // and the current buffer has no start-of-frame flag,
                // for example.
                //
                // So, toss this Buffer, and move on the next. Otherwise
                // we could sit here forever.
                TP(SP_BQ, "%s: Drop buffer at head of consumer queue", __PRETTY_FUNCTION__);
                bq.finished();
            }
        }
    }
}

void StreamingProtocol::producer_set_signal(void (*func)(void *), void *arg)
{
    TP(SP_FUNC, "%s", __PRETTY_FUNCTION__);
    bq.producer_set_signal(func, arg);
}

void StreamingProtocol::producer_wait()
{
    TP(SP_FUNC, "%s", __PRETTY_FUNCTION__);
}

void StreamingProtocol::producer_set_flags(uint32_t f)
{
    TP(SP_FUNC, "%s: flags=0x%x", __PRETTY_FUNCTION__, f);
    bq.producer_set_flags(f);
}

void StreamingProtocol::producer_set_size(size_t s)
{
    TP(SP_FUNC, "%s: size=0x%x", __PRETTY_FUNCTION__, s);
    bq.producer_set_size(s);
}

void StreamingProtocol::producer_set_frame_id(int32_t n)
{
    TP(SP_FUNC, "%s: id=%d", __PRETTY_FUNCTION__, n);
    bq.producer_set_frame_id(n);
}

void StreamingProtocol::producer_set_frame_data_size(uint32_t fds)
{
    TP(SP_FUNC, "%s: fds=%d", __PRETTY_FUNCTION__, fds);
    bq.producer_set_frame_data_size(fds);
}

char *StreamingProtocol::producer_get_data_ptr(size_t offset)
{
    TP(SP_FUNC, "%s", __PRETTY_FUNCTION__);
    return bq.producer_get_data_ptr(offset);
}

void StreamingProtocol::producer_reserve_prefix(size_t size)
{
    TP(SP_FUNC, "%s: size=%d", __PRETTY_FUNCTION__, size);
    bq.producer_reserve_prefix(size);
}

void StreamingProtocol::set_fault(void)
{
    TP(SP_FUNC, "%s", __PRETTY_FUNCTION__);
    bq.set_fault(true);
}
