/*!
 * @file fi_wss_main.cxx
 * @brief implements liveview support for communicating with the camera server.
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <getopt.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>

#include <sys/time.h>
#include <libwebsockets.h>

#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

// our local include files for the project
#include "fi_buffer_q.h"
#include "fi_streaming_session.h"
#include "fi_download_protocol.h"
#include "fi_download_session.h"
#include "fi_playback_protocol.h"
#include "fi_playback_session.h"
#include "fi_cs.h"
#include "fi_wss_protocols.h"
#include "lssm_thread.h"
#include "status_update.h"

using namespace std;

// The debug_level is used as default it will report any
// LWS log entry of an error, warning or notice to get a
// higher level pass on command line and see libwebsockets.h
// for the enum lws_log_levels
static int debug_level = LLL_ERR | LLL_WARN; // | LLL_NOTICE;

// The Camera Server IO object (currently this is a singleton)
// it was created to isolate the details of a fcp.cxx module and
// provide an interface that maybe one day is needed for a case where
// multiple Camera Server objects might exit but that is probably
// unlikely so it is really only serving as an isolation from the
// FCP details and globals
static fiCameraServerIO fiCSInterface;

// Leftover static variables from testserver.c
static volatile int fiWSSForceExit = 0;

// The Lib Websocket Context Structure
// used to manage WSS communications through the library
static struct lws_context *fiWSSContext = nullptr;

// The WSS Protocol object that is maintaining all of the supported
// protocols of the WSS.  Currently there is only one of these and it
// stores the specifics of the common data shared between all of the
// supported protocols that supports the libwebsocket library.  It is
// a singleton object and will always be one.
static fiWSSProtocols wssProtocols;

// ================================================================================
// ALM TEST TEST TEST =============================================================
static const struct lws_http_mount mount =
{
    /* .mount_next */       nullptr,    /* linked-list "next" */
    /* .mountpoint */       "/",        /* mountpoint URL */
    /* .origin */           "/srv/http",    /* serve from dir */
    /* .def */          "index.html",   /* default filename */
    /* .protocol */         nullptr,
    /* .cgienv */           nullptr,
    /* .extra_mimetypes */      nullptr,
    /* .interpret */        nullptr,
    /* .cgi_timeout */      0,
    /* .cache_max_age */        0,
    /* .auth_mask */        0,
    /* .cache_reusable */       0,
    /* .cache_revalidate */     0,
    /* .cache_intermediaries */ 0,
    /* .origin_protocol */      LWSMPRO_FILE,   /* files in a dir */
    /* .mountpoint_len */       1,      /* char count */
    /* .basic_auth_login_file */    nullptr
};

// ALM TEST TEST TEST =============================================================
// ================================================================================

/* --------------------------------------------------------------------------------
* @brief Get Camera Server Object pointer
*
* Right now the Web Socket Server only maintains a single connection
* to a Camera Server.  In the future this may change and multiple
* Camera Server Connections may be needed for a single WSS - probably
* not likely since the WSS runs on the Camera running the CS but who knows!
*
* @return pointer to the CS object for the WSS or NULL pointer if not available
* --------------------------------------------------------------------------------*/
fiCameraServerIO *fiGetCS()
{
    return &fiCSInterface;
}

/* --------------------------------------------------------------------------------
* @brief LibWebSocket Test Server Sighandler Local Function
*
* pulled over from testserver.c
* --------------------------------------------------------------------------------*/
void fiWSSSigHandler(int sig)
{
    LOG("%s: sig=%d", __PRETTY_FUNCTION__, sig);

    fiWSSForceExit = 1;
    lws_cancel_service(fiWSSContext);
}

void fiWSSLogHandler(int level, const char *line)
{
    fi_log(level, line);
}

/* --------------------------------------------------------------------------------
* @brief Main for Fastec Imaging Web Socket Server
*
* Responsible for creating the Camera Server Connection, the Web Socket Protocols
* And for Server Execution
* --------------------------------------------------------------------------------*/
int main(int argc, char **argv)
{
    // int malloc_stats_counter = 0;
    int opts = 0;
    int n = 0;

    int cam_port = 0;           // default in fi_cs.cxx
    int wss_port = 7681;        // the documented Default port for WSS Communications
    std::string cam_ip_arg;

    // the Lib Websocket Creation Structure
    struct lws_context_creation_info fiWSSCreateInfo;

    // what is the state of Camera Server Initialization?
    FI_CSIO_STATES fiCSInitState;

    // what is the state of creation for the WSS Protocols supported
    FI_WSSPROTOCOLS_STATE wssProtocolsCreated;

    static struct option options[] =
    {
        { "help",           no_argument,        NULL, 'h' },
        { "debug",          required_argument,  NULL, 'd' },
        { "wssport",        required_argument,  NULL, 'w' },
        { "ssl",            no_argument,        NULL, 's' },
        { "ip-cs",          required_argument,  NULL, 'i' },
        { "port-cs",        required_argument,  NULL, 'p' },
        { "local-cs",       required_argument,  NULL, 'l' },
        { "allow-non-ssl",  no_argument,        NULL, 'a' },
        { "libev",          no_argument,        NULL, 'e' },
#ifndef LWS_NO_DAEMONIZE
        { "daemonize",      no_argument,        NULL, 'D' },
#endif
        { "resource_path",  required_argument,  NULL, 'r' },
    };

#if 1
    loggerUseStdout(1);
#else
    loggerFile("/tmp/cs.log");
    loggerUseStdout(0);
    loggerUseSyslog(0);
#endif

    // Catch SIGINT to shutdown gracefully.
    signal(SIGINT, fiWSSSigHandler);
    signal(SIGTERM, fiWSSSigHandler);

    // Parse the command line arguments.
    while (n >= 0)
    {
        n = getopt_long(argc, argv, "hd:w:si:p:l:ae", options, NULL);
        if (n < 0)
        {
            continue;
        }
        switch (n)
        {
        case 'a':
            opts |= LWS_SERVER_OPTION_ALLOW_NON_SSL_ON_SSL_PORT;
            break;
        case 'd':
            debug_level = (int)strtol(optarg, NULL, 0);
            break;
        case 'w':
            wss_port = atoi(optarg);    // can specify non-default wss port
            break;
        case 'p':
            //cam_port = atoi(optarg); // defaults to GIGE_SERVER_UDP_PORT in fiCS constructor
            break;
        case 'i':
            // NOTE:  if the user specifies an IP for the camera server
            // then it is assumed the CS is remote (i.e. testing on remote Ubuntu)
            // otherwise if the argument is lowercase ascii 'l' then it is assumed
            // the CS is local and this wss is running on the Camera OS.
            cam_ip_arg.assign(optarg);
            break;
        case 'h':
            fprintf(stderr, "Usage: fi_wss [options]\n");
            fprintf(stderr, "  Options:\n");
            fprintf(stderr, "  --help, -h                  Display this help message\n");
            fprintf(stderr, "  --debug, -d <level>         Set debug output detail level\n");
            fprintf(stderr, "  --wssport, -w <port>        Websockets listen port\n");
            fprintf(stderr, "  --ssl, -s                   Use SSL on Websockets sessions\n");
            fprintf(stderr, "  --ip-cs, -i <address>       IP Address of Camera Server\n");
            fprintf(stderr, "  --port-cs, -p <port>        IP port for Camera Server\n");
            fprintf(stderr, "  --local-cs, -l 0|1          '1' Camera Server on localhost; '0' Camera Server on remote host\n");
            fprintf(stderr, "  --allow-non-ssl, -a         --\n");
            fprintf(stderr, "  --libev, -e                 --\n");
#ifndef LWS_NO_DAEMONIZE
            fprintf(stderr, "  --daemonize, -D             Execute as a daemon process\n");
#endif
            fprintf(stderr, "  --resource_path, -r         --\n");
            exit(1);
        }
    }

    // set up the libwebsockets log debugging level
    lws_set_log_level(debug_level, fiWSSLogHandler);

// built-in daemonize logic copied over from testserver.c can do this by command line
// will need to put in option and make this call to lws interface
//#if !defined(LWS_NO_DAEMONIZE) && !defined(WIN32)
//  /*
//   * normally lock path would be /var/lock/lwsts or similar, to
//   * simplify getting started without having to take care about
//   * permissions or running as root, set to /tmp/.lwsts-lock
//   */
//  if (daemonize && lws_daemonize("/tmp/.lwsts-lock")) {
//      fprintf(stderr, "Failed to daemonize\n");
//      return 1;
//  }
//#endif

    // mark the copyright notice for log output...
    lwsl_notice("Fastec Imaging Web Socket Server - "
                "(C) Copyright 2010-2018 Fastec Imaging - "
                "licensed under LGPL2.1\n");

    // Create the Camera Server IO Object
    fiCSInitState = fiCSInterface.initCSConnection(cam_ip_arg.c_str(), cam_port);

    // ALM-TODO handle failure conditions for fiCSInitState
    if (fiCSInitState != FI_CSIO_CONNECTED)
    {
        lwsl_err("Unable to Connect to FI Camera Server Return Code:  %d\n", fiCSInitState);
        goto main_bail;
    }

    // report the camera server connection information
    lwsl_notice("FI WSS Connected to Camera Server on %s using Port %d\n", fiCSInterface.getCameraIPStr(), fiCSInterface.getCameraPort());

    // Create supported FI Web Socket Server Protocols
    wssProtocolsCreated = wssProtocols.createProtocols(&fiCSInterface);

    // What is the result of the creation logic?
    if (wssProtocolsCreated != FI_WSSPROTOCOLS_CREATED)
    {
        lwsl_err("Unable to Create FI WSS Protocols\n");
        goto main_bail;
    }
    lwsl_notice("FI WSS Protocols Have Been Created\n");

    // If the protocols were created without any problems, set up the
    // context that is neccessary for our Web Socket Server. Currently
    // there is only one Context correlating to the execute loop below.

    // Initialize the structure to all zeros.
    memset(&fiWSSCreateInfo, 0, sizeof(struct lws_context_creation_info));

    // Set the port we plan on using.
    fiWSSCreateInfo.port = wss_port;    // the WSS Port used for communications

    // The creation information defines how the wss context record is created.
    fiWSSCreateInfo.iface = (const char *)NULL;

    // I have not tested SSL usage at all.
    fiWSSCreateInfo.ssl_cert_filepath = NULL;
    fiWSSCreateInfo.ssl_private_key_filepath = NULL;

    // Install the protocols list.
    wssProtocols.setListOfProtocols(&fiWSSCreateInfo);

// ================================================================================
// ALM TEST TEST TEST =============================================================

    fiWSSCreateInfo.mounts = &mount;
    fiWSSCreateInfo.error_document_404 = "/404.html";

// ALM TEST TEST TEST =============================================================
// ================================================================================

    fiWSSCreateInfo.gid = -1;
    fiWSSCreateInfo.uid = -1;
    fiWSSCreateInfo.options = opts; // OPTIONS SET IN ARGUMENT HANDLING

    // Use TCP keepalives on all our connections; we really really want
    // reasonable notifications if/when clients disappear without saying
    // anything---even when we have nothing in particular to say to them.
    fiWSSCreateInfo.ka_time = 1;        // Start keepalives one second after the
    // last data is sent
    fiWSSCreateInfo.ka_probes = 4;      // Not very patient; declare an error
    // after 4 un-answered probes
    fiWSSCreateInfo.ka_interval = 1;    // Once we start sending probes, send
    // one every second.

    // Create our wss context record.
    fiWSSContext = lws_create_context(&fiWSSCreateInfo);
    if (fiWSSContext == nullptr)
    {
        lwsl_err("Unable to Create FIWSS Context\n");
        goto main_bail;
    }

    lwsl_notice("Successfully Created FIWSS Context\n");
    wssProtocols.getLiveViewProto()->set_lws_context(fiWSSContext);
    wssProtocols.getPlaybackProto()->set_lws_context(fiWSSContext);

    // Create a block for our main loop operation. This prevents complaints
    // about the "gotos" jumping over the creation of the LSSM thread.
    {
        // Create our thread.
        LSSM_Thread lssm(wssProtocols.getLiveViewProto(),
                         wssProtocols.getPlaybackProto(),
                         wssProtocols.getDownloadProto(),
                         fiWSSContext);

        // Create our polling thread object.
        StatusUpdate status_update;

        // Keep going until told to exit, or we error.
        n = 0;
        while ((n >= 0) && !fiWSSForceExit)
        {
            // Give protocols a chance to write any pending data out on wss
            wssProtocols.canWriteIfDataReady(fiWSSContext);

            /* Check status of Streamers (download, live view, and playback) */
            wssProtocols.getDownloadProto()->process_consumer();
            wssProtocols.getPlaybackProto()->process_consumer();
            wssProtocols.getLiveViewProto()->process_consumer();

            /*
             * Run libwebsockets main loop.
             */
            n = lws_service(fiWSSContext, 0);
        }
    }

    // done with the libwebsocket context data we are exiting
    lws_context_destroy(fiWSSContext);

main_bail:
    // TODO log status of exit
    lwsl_notice("FI WSS Exiting\n");

    // done logging
    closelog();

    return 0;
}

// end of fi_wss_main.cxx
