/*!
 * @file fi_wss_protocols.cxx
 * @brief implementation file for the fi_wss_protocols.cpp
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

// NOTE:
// could have designed better with class hierarchy and array of
// prototype subclasses managed in collection but this layout is
// obvious what's happening and easier to transition the old code
// to something more readable
//
// also only have 1 callback per prototype
// and it's a static method (but doesn't have to be static method)
// talked with tony and he felt it was more readable that way sort of
// a developer's preference to be more explicit in what's happening
//
// also to get access to the camera server there is a call on the main
// module that provides a pointer to the one attached - but
// could've probably passed in the Camera Server IO Object used for this
// set of protocols if for some reason there were more than one camera
// servers being used.  etc. etc. etc...

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <string.h>
#include <exception>

#include <fastec/fas_log.h>

#include "fi_cs.h"

#include "fcp_protocol.h"
#include "fcp_listen_protocol.h"
#include "fi_http_protocol.h"
#include "fi_liveview_protocol.h"
#include "fi_playback_protocol.h"
#include "fi_download_protocol.h"
#include "fi_wss_protocols.h"

//--------------------------------------------------------------------
// Forward Class Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class: fiWSSProtocol
//--------------------------------------------------------------------

// Definition for the protocols array - this is where all of the protocols
// are defined for the libwebsocket interfaces of the Fastec Imaging WSS.
// It uses the enum PROTOCOL_COUNT from fi_wss_protocols.h in the declaration
// for the supported protocols.
struct lws_protocols fiWSSProtocols::protocols[PROTOCOL_COUNT + 1];

/*! --------------------------------------------------------------------------------
 * @brief Translate Package Status Code into User Readable String
 *
 * translates the package status integer value into a character readable message
 *
 * @return none
 */
fiWSSProtocols::fiWSSProtocols() :
    pHTTPProto(0),
    pFCPProto(0),
    pFCPListenerProto(0),
    pLVListenerProto(0),
    pPlaybackProto(0),
    pDownloadProto(0)
{
    // not initialized yet
    ProtocolsAreReady = FI_WSSPROTOCOLS_UNINITIALIZED;

    // set the protocol records necessary for the libwebsocket interface
    // to empty clean slate
    memset(protocols, 0, sizeof( protocols ));
}

/*! --------------------------------------------------------------------------------
 * @brief Destructs the Fastec Control Protocol Handler Object
 */
fiWSSProtocols::~fiWSSProtocols()
{
    ProtocolsAreReady = FI_WSSPROTOCOLS_UNINITIALIZED;	// leaving
}

/*! --------------------------------------------------------------------------------
 * @brief setProtocolsListInInfoRec
 *
 * installs the pointer to our static list of protocols so wss context
 * can be created
 *
 * @return protocols list pointer or null if error
 */
void fiWSSProtocols::setListOfProtocols(struct lws_context_creation_info *pfiWSSCreateInfo)
{
    // Install a pointer to our protocols structure into the libwebsocket
    // creation information record.
    if (pfiWSSCreateInfo != (struct lws_context_creation_info *) NULL)
        pfiWSSCreateInfo->protocols = &fiWSSProtocols::protocols[0];
}

/*! --------------------------------------------------------------------------------
 * @brief Protocol Creation is handled here
 *
 * constructs and prepares all supported protocol objects for use in the
 * fastec imaging web socket server
 *
 * @param csInterfacePtr 	ptr to the camera server IO object
 *
 * @return 1 if all protocols are ready or 0 if something went wrong
 */
FI_WSSPROTOCOLS_STATE fiWSSProtocols::createProtocols(fiCameraServerIO* /*csInterfacePtr*/)
{
    // create the FCP Command Protocol Handler Object
    if (ProtocolsAreReady == FI_WSSPROTOCOLS_UNINITIALIZED)
    {
	pHTTPProto = new fiHTTPProtocol(&fiWSSProtocols::protocols[PROTOCOL_HTTP]);
        if (pHTTPProto != (fiHTTPProtocol*)NULL)
        {
            if (!pHTTPProto->isReady())
                ProtocolsAreReady = FI_WSSPROTOCOLS_CREATION_FAILED;
        }

        // create the FCP Command Protocol Handler
        pFCPProto = new fiWSSFCPProtocol(&fiWSSProtocols::protocols[PROTOCOL_FCP]);
        if (pFCPProto != (fiWSSFCPProtocol*)NULL)
        {
            if (!pFCPProto->isReady())
                ProtocolsAreReady = FI_WSSPROTOCOLS_CREATION_FAILED;
        }
    }
    lwsl_debug("%s: FCP protocol result: %d\n",  __FUNCTION__,  ProtocolsAreReady);

    // Create the FCP Listener Protocol Handler
    if (ProtocolsAreReady == FI_WSSPROTOCOLS_UNINITIALIZED)
    {
        // Create the FCP Listener Protocol Handler
        pFCPListenerProto = new fiWSSFCPListenProtocol(&fiWSSProtocols::protocols[PROTOCOL_FCP_LISTENER]);
        if (pFCPListenerProto != (fiWSSFCPListenProtocol *)NULL)
        {
            if (!pFCPListenerProto->isReady())
                ProtocolsAreReady = FI_WSSPROTOCOLS_CREATION_FAILED;
        }
    }
    lwsl_debug("%s: FCP Listener (FCP Notify) protocol result: %d\n",  __FUNCTION__,  ProtocolsAreReady);

    try
    {
        pPlaybackProto   = new fiWSSPlaybackProtocol(&fiWSSProtocols::protocols[PROTOCOL_FI_PLAYBACK]);
        pDownloadProto   = new fiWSSDownloadProtocol(&fiWSSProtocols::protocols[PROTOCOL_FI_DOWNLOAD]);
        pLVListenerProto = new fiWSSLiveViewProtocol(&fiWSSProtocols::protocols[PROTOCOL_FI_LIVEVIEWER]);
    }
    catch (exception& e)
    {
        lwsl_debug("%s: Exception during protocol creation\n",  __FUNCTION__);
        ProtocolsAreReady = FI_WSSPROTOCOLS_CREATION_FAILED;
    }

    // If the state is not set to a failure condition the protocols
    // are ready to go.
    if (ProtocolsAreReady != FI_WSSPROTOCOLS_CREATION_FAILED)
        ProtocolsAreReady = FI_WSSPROTOCOLS_CREATED;

    return ProtocolsAreReady;
}


/*! --------------------------------------------------------------------------------
 * @brief Time Slice for Protocols
 *
 * Walks the protocols and gives them a slice of time. The streaming protocols all
 * post for writes when they have data to send, so they aren't included here.
 */
void fiWSSProtocols::canWriteIfDataReady(struct lws_context *ctx)
{
    if (pFCPProto != nullptr)
    {
        if (pFCPProto->isReady() )
            lws_callback_on_writable_all_protocol(ctx, &fiWSSProtocols::protocols[PROTOCOL_FCP]);
    }

    if (pFCPListenerProto != nullptr)
    {
        if (pFCPListenerProto->isReady() )
            lws_callback_on_writable_all_protocol(ctx, &fiWSSProtocols::protocols[PROTOCOL_FCP_LISTENER]);
    }
}

fiWSSLiveViewProtocol* fiWSSProtocols::getLiveViewProto()
{
    return pLVListenerProto;
}

fiWSSPlaybackProtocol* fiWSSProtocols::getPlaybackProto()
{
    return pPlaybackProto;
}

fiWSSDownloadProtocol* fiWSSProtocols::getDownloadProto()
{
    return pDownloadProto;
}

// end of fi_wss_protocols.cxx
