/*
 * lssm_thread.cxx
 *
 *  Created on: Feb 2, 2016
 *      Author: eric
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <string.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <libwebsockets.h>

#include <fastec/gige_msg.h>
#include <fastec/fas_log.h>
#include <fastec/fas_trace.h>

#include "fcp.h"
#include "fi_streaming_protocol.h"

#include "lssm_thread.h"

//--------------------------------------------------------------------
// Constant, Type and Macro definitions
//--------------------------------------------------------------------

enum LSSM_TRACE
{
    LSSM_FUNC    = 0x00000010,
    LSSM_STATE   = 0x00000020,
    LSSM_BQ = 0x00000040,
    LSSM_DETAIL = 0x00000080
};

const uint64_t WAIT_FOR_WORK_TIME_US = 5000000;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

static uint32_t traceControl =
    LSSM_FUNC   |
    LSSM_STATE  |
    LSSM_BQ     |
    LSSM_DETAIL |
//        TRACE_0;        // Set our trace level to TRACE_0 to trace.
        TRACE_3;        // Set our trace level to TRACE_3 to not trace.


/**
 * @brief Trampoline for custom signaling hook
 *
 * This just provides a static function so we can pass an address
 * to BufferQueue::producer_set_signal(). BufferQueue will use this
 * (instead of its native signaling) to tell us when the Consumer
 * has finished with a Buffer.
 */
static void local_signal(void *arg)
{
    static_cast<LSSM_Thread *>(arg)->consumer_signal();
}

LSSM_Thread::LSSM_Thread(fiWSSLiveViewProtocol *lv,
                         fiWSSPlaybackProtocol *pb,
                         fiWSSDownloadProtocol *dl,
                         struct lws_context *ctx) :
    lws_ctx(ctx),
    mmap_fd(-1),
    current_mode(-1),
    liveview_proto(lv),
    playback_proto(pb),
    download_proto(dl)
{
    pthread_attr_t attr;
    int status;

    TP(LSSM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Register with the various protocols we might be serving.
    liveview_proto->producer_set_signal(local_signal, (void *)this);
    playback_proto->producer_set_signal(local_signal, (void *)this);
    download_proto->producer_set_signal(local_signal, (void *)this);

    // Initialize the shared memory interface.
    if (lssm_client_init(&shmem) != 0)
        fi_log(LOG_ERR, "%s: failed", __PRETTY_FUNCTION__);

    // Open the memory device.
    mmap_fd = open("/dev/mem", O_RDWR /* | O_SYNC */);
    if (mmap_fd < 1)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        return;
    }

    // Create our worker thread.
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    status = pthread_create(&thread, &attr, lssm_entry, (void *)this);
    if (status != 0)
        fi_log(LOG_ERR, "%s: pthread_create() failed [%d, %s]", __PRETTY_FUNCTION__, status, strerror(status));
    pthread_attr_destroy(&attr);
}

LSSM_Thread::~LSSM_Thread()
{
    TP(LSSM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Close the memory device.
    if (mmap_fd >= 0)
    {
        close(mmap_fd);
        mmap_fd = -1;
    }
}

void LSSM_Thread::consumer_signal()
{
    TP(LSSM_FUNC, "%s", __PRETTY_FUNCTION__);

    lssm_lock(shmem);
    lssm_client_signal_l(shmem);
    lssm_unlock(shmem);
}

void LSSM_Thread::run(void *t)
{
    bool lssm_is_idle = true;

    (void)t; // unused

    TP(LSSM_FUNC, "%s", __PRETTY_FUNCTION__);

    // Main loop for LSSM thread:
    // This loop executes transfer operations from memory into Buffers, on
    // behalf of the three Fastec WebSocket streaming protocols:
    //    - Live View
    //    - Playback
    //    - Download
    //
    // It blocks until it receives one of the following events:
    //    - A new image frame has become available for one
    //      of the streaming protocols.
    //    - A Buffer consumer has finished processing a Buffer,
    //      and has released that buffer back to the LSSM thread.
    //    - The LSSM Server has declared a fault condition
    //
    // This loop handles those events as follows:
    //
    // New image frame:
    //    - Determine the protocol to which the image frame belongs
    //    - If the BufferQueue has a Producer Buffer available, transfer
    //      data from memory into the buffer, and pass the Buffer to
    //      the BufferQueue's Consumer(s).
    //
    // Released Buffer:
    //    - If transfer bytes remain in the current image, transfer data from
    //      memory into the buffer, and pass the Buffer to the BufferQueue's
    //      Consumer(s).
    // LSSM Fault:
    //
    for (;;)
    {
        lssm_lock(shmem);
        while (1)
        {
            // Check to see if we have work to do.
            lssm_is_idle = lssm_idle();
            if (!lssm_is_idle)
                break;

            // Nothing going on, let's wait. Forever if necessary.
            lssm_client_wait_l(shmem, WAIT_FOR_WORK_TIME_US);
        }
        lssm_unlock(shmem);

        process_xfer(liveview_proto); // Do work for Liveview if possible
        process_xfer(playback_proto); // Do work for Playback if possible
        process_xfer(download_proto); // Do work for Download if possible
    }
}

void LSSM_Thread::process_xfer(StreamingProtocol *prot)
{
    uint32_t flags = 0;
    size_t buf_size = 0;

    BufferQueue& bq = prot->buffer_q();

    TP(LSSM_FUNC, "%s: prot=%p", __PRETTY_FUNCTION__, (void *)prot);

    // Is there anything to do on this protocol?
    if (bq.bytes_to_read_from_im == 0)
    {
        TP(LSSM_DETAIL, "%s: Nothing to do on %p", __PRETTY_FUNCTION__, (void *)prot);
        return;
    }
    if (!bq.producer_owns_current())
    {
        TP(LSSM_DETAIL, "%s: Producer [%p] doesn't own current buffer.", __PRETTY_FUNCTION__, (void *)prot);
        return;
    }

    // Are we part way into the transfer process? If xfer_offset is 0, then
    // we know we are at the start of the frame.
    if (bq.xfer_offset == 0)
    {
        TP(LSSM_BQ, "%s: First buffer of frame %d", __PRETTY_FUNCTION__, bq.frame_num);

        // First Buffer of frame. It needs to have START_OF_FRAME flag set and the
        // overall frame size and ID set. Also, it may also need a prepended image header.
        flags |= StreamingProtocol::START_OF_FRAME;

        // Initialize the length and ID for this frame.
        bq.producer_set_frame_data_size(bq.bytes_to_read_from_im + bq.bytes_to_read_from_header);
        bq.producer_set_frame_id(bq.frame_num);

        // Are there header bytes for this frame?
        if (bq.bytes_to_read_from_header > 0)
        {
            // Copy the header bytes into our Buffer
            bq.producer_reserve_prefix(bq.bytes_to_read_from_header);
            uint8_t *hdr_dst = (uint8_t *)(bq.producer_get_data_ptr(0));
            lssm_copyin_image_header(shmem, (uint8_t *)hdr_dst, bq.bytes_to_read_from_header);
            buf_size = bq.bytes_to_read_from_header;
            bq.bytes_to_read_from_header = 0;
        }
        else
        {
            // No header, so update the information for 0 header bytes.
            bq.producer_reserve_prefix(0);
        }
    }

    // We'll transfer some pixel data into this currently empty (not counting
    // possible image header bytes) Buffer. Figure out how much. It will be
    // the lesser of the space available (BUFF_DATA_SIZE) or remaining pixel
    // data for this frame (bytes_to_read_from_im). We may also have to truncate
    // the size to a complete pixel row, depending on the image data format.
    uint8_t *dst_addr = (uint8_t *)bq.producer_get_data_ptr(buf_size);
    size_t bytes_to_xfer = compute_xfer_size(Buffer::BUFF_DATA_SIZE, bq.bytes_to_read_from_im);
    buf_size += bytes_to_xfer;
    bq.producer_set_size(buf_size);
    bq.bytes_to_read_from_im -= bytes_to_xfer;

    // If there are no leftover bytes, then make sure to set the END_OF_FRAME flag.
    if (bq.bytes_to_read_from_im == 0)
    {
        TP(LSSM_BQ, "%s: Last buffer of frame %d", __PRETTY_FUNCTION__, bq.frame_num);
        flags |= StreamingProtocol::END_OF_FRAME;
    }
    bq.producer_set_flags(flags);

    // Move the data over.
    deal_with_xfer(bq, bytes_to_xfer, dst_addr);

    // Update our offset for the next time.
    bq.xfer_offset += bytes_to_xfer;

    // Is the producer still active?
    if (bq.producer_finished() == 0)
    {
        // Signal consumer (WSS thread) via lws_cancel_service()
        TP(LSSM_BQ, "%s: Signal (lws_cancel_service())", __PRETTY_FUNCTION__);
        lws_cancel_service(lws_ctx);
    }

    // If we exhausted a complete frame from image memory, notify the
    // source of those frames that we're finished with the buffer in
    // image memory.
    if (bq.bytes_to_read_from_im == 0)
    {
        TP(LSSM_DETAIL, "%s: Release IM buffer %d", __PRETTY_FUNCTION__, bq.buf_id);
        prot->release_image_memory_buffer(shmem);
    }
}

/*!
 * @brief Check to see if there's any work to be done.
 * @note lssm_lock() is held prior to calling this function.
 * @return TRUE if there is no work to be done (i.e. IDLE). FALSE otherwise.
 */
bool LSSM_Thread::lssm_idle()
{
    int lssm_mode;
    StreamingProtocol *prot;

    // What mode are we now operating in?
    lssm_mode = lssm_get_mode(shmem);
    TP(LSSM_FUNC, "%s: current_mode=%d, lssm_mode=%d", __PRETTY_FUNCTION__, current_mode, lssm_mode);

    // Have we switch modes yet?
    if (current_mode != lssm_mode)
    {
        // The mode MAY have changed midway through transmitting a frame. We need to finish
        // transmitting that frame before we honor the mode switch. Check for remnants of
        // frames in any of the protocols.
        if ((current_mode == GIGE_LOCALSTREAM_PLAYBACK) || (current_mode == GIGE_LOCALSTREAM_SINGLESTEP))
            prot = playback_proto;
        else if (current_mode == GIGE_LOCALSTREAM_DOWNLOAD)
            prot = download_proto;
        else if (current_mode == GIGE_LOCALSTREAM_LIVE)
            prot = liveview_proto;
        else
            prot = nullptr;

        if (prot != nullptr)
        {
            // Get the buffer queue for the current mode.
            BufferQueue& bq = prot->buffer_q();
            if ((bq.bytes_to_read_from_im > 0) && bq.producer_owns_current())
            {
                TP(LSSM_BQ, "%s: current_mode (%d) bq is not done yet!", __PRETTY_FUNCTION__, current_mode);
                return false;
            }
        }
    }

    // Switch modes.
    current_mode = lssm_mode;

    // There aren't any fragments of frames hanging out, check the current mode.
    if ((lssm_mode == GIGE_LOCALSTREAM_PLAYBACK) || (lssm_mode == GIGE_LOCALSTREAM_SINGLESTEP))
        prot = playback_proto;
    else if (lssm_mode == GIGE_LOCALSTREAM_DOWNLOAD)
        prot = download_proto;
    else if (lssm_mode == GIGE_LOCALSTREAM_LIVE)
        prot = liveview_proto;
    else
        return true; // Idle: LSSM not running, or not running in a mode we can use

    BufferQueue& bq = prot->buffer_q();
    bool server_fault = lssm_get_server_fault_l(shmem);
    bool client_fault = bq.get_fault();

    // If the buffer queue is reporting a fault, let's reset it here.
    // We'll deal with the fault next.
    if (client_fault)
        bq.set_fault(false);

    if (server_fault && client_fault)
    {
        // Handle simultaneous faults
        fi_log(LOG_INFO, "%s: LSSM Client and Server have both declared faults", __PRETTY_FUNCTION__);
        bq.bytes_to_read_from_im = 0;
        lssm_client_clear_fault_l(shmem);
    }
    else if (server_fault)
    {
        // Propagate fault indication down stream.
        fi_log(LOG_INFO, "%s: LSSM Server has declared a fault", __PRETTY_FUNCTION__);
        lssm_client_clear_fault_l(shmem);
        bq.set_fault(true);
        bq.bytes_to_read_from_im = 0;
    }
    else if (client_fault)
    {
        fi_log(LOG_INFO, "%s: LSSM Client has declared a fault", __PRETTY_FUNCTION__);
        bq.bytes_to_read_from_im = 0;
        bq.bytes_to_read_from_header = 0;
        bq.buf_id = -1;
        lssm_client_cleanup_l(shmem);

        // Propagate fault indication up stream (to LSSM server)
        lssm_client_set_fault_l(shmem);
    }
    else
    {
        // No faults; check progress of the transfer.
        if (!bq.producer_owns_current())
        {
            // Producer does not own current Buffer; no faults to report, so idle
            TP(LSSM_BQ, "%s: No faults, and producer does not own current Buffer: idle", __PRETTY_FUNCTION__);
            return true;
        }

        TP(LSSM_BQ, "%s: producer buffer count=%d", __PRETTY_FUNCTION__, bq.get_producer_buffer_count());

        // Check to see if we're idle (no bytes to read from the IM).
        if (bq.bytes_to_read_from_im == 0)
        {
            // No transfer in progress; no fault. Maybe we can kick off a new transfer.
            // Check LSSM interface for activity
            if ((lssm_mode == GIGE_LOCALSTREAM_LIVE) || (lssm_mode == GIGE_LOCALSTREAM_PLAYBACK) || (lssm_mode == GIGE_LOCALSTREAM_SINGLESTEP))
                bq.buf_id = lssm_get_newest_l(shmem); // For Liveview and Playbacks
            else // lssm_mode == GIGE_LOCALSTREAM_DOWNLOAD
                bq.buf_id = lssm_get_oldest_l(shmem); // For Download

            TP(LSSM_BQ, "%s: buf_id=%d", __PRETTY_FUNCTION__, bq.buf_id);
            if (bq.buf_id >= 0)
            {
                // ALM_TODO: We could do the MMAP operations as soon as we have our bufid.
                //       instead of during the transfer operation. Maybe more efficient?

                // Get the size of the image that we're going to transmit over.
                bq.image_size = lssm_get_processed_size(shmem, bq.buf_id);
                bq.bytes_to_read_from_im = bq.image_size;
                bq.bytes_to_read_from_header = lssm_get_image_header_size(shmem);
                if (bq.bytes_to_read_from_header > Buffer::BUFF_DATA_SIZE)
                {
                    // Sanity check; should never happen
                    fi_log(LOG_ERR, "%s: Got header size %ld bytes, which is too big", __PRETTY_FUNCTION__, bq.bytes_to_read_from_header);
                    bq.bytes_to_read_from_header = 0; // Just skip it
                }

                // Initialize our addresses and offsets to begin the transfer of the frame.
                bq.xfer_base = lssm_get_image_address(shmem, bq.buf_id);

                // Live view handles frame numbers a little differently.
                if (lssm_mode == GIGE_LOCALSTREAM_LIVE)
                    bq.frame_num = liveview_proto->get_next_id();
                else
                    bq.frame_num = lssm_get_frame_num(shmem, bq.buf_id);
                bq.xfer_offset = 0;

                TP(LSSM_BQ, "%s: LSSM Buff %d ready (%d producer-owned). Frame #%d: %ld bytes @ 0x%" PRIx64 " + %ld header.",
                   __PRETTY_FUNCTION__, bq.buf_id, bq.get_producer_buffer_count(), bq.frame_num, bq.bytes_to_read_from_im,
                   bq.xfer_base, bq.bytes_to_read_from_header);
            }
        }
    }

    // At this point we consider the LSSM idle if there is no data to transfer from
    // into our producer-owned Buffer (bq.bytes_to_read_from_im == 0),
    // and if there is no fault declared from the server that needs to propagate
    // down to the WebSocket protocol.
    if ((bq.bytes_to_read_from_im == 0) && !bq.get_fault())
    {
        TP(LSSM_DETAIL, "%s: No LSSM data ready for streaming protocol and no faults declared: idle", __PRETTY_FUNCTION__);
        return true; // Idle
    }

    // Otherwise, there is something we can/should do: LSSM interface is not idle
    TP(LSSM_BQ, "%s: LSSM (%d) not idle (buffer %d): %ld bytes to read; %s", __PRETTY_FUNCTION__,
       bq.buf_id, bq.producer_index(), bq.bytes_to_read_from_im, (server_fault ? "Server declares fault" : "No server fault"));

    return false; // Not idle
}

size_t LSSM_Thread::compute_xfer_size(size_t space_available, size_t data_available)
{
    TP(LSSM_FUNC, "%s: space=%d, data=%d", __PRETTY_FUNCTION__, space_available, data_available);

    if (space_available > data_available)
        space_available = data_available;
    return space_available;
}

void LSSM_Thread::deal_with_xfer(BufferQueue& bq, size_t nbytes, uint8_t *dst)
{
    uint8_t *mmap_ptr;
    uint32_t mmap_len;

    TP(LSSM_FUNC, "%s: id=%d, nbytes=%d, dst=%d/%p", __PRETTY_FUNCTION__, bq.buf_id, nbytes, bq.producer_index(), (void *)dst);

    // Get the size of the frame, and round it up to a page boundary.
    mmap_len = ((bq.image_size + 4095) >> 12) << 12;

    // Now map the memory.
    TP(LSSM_DETAIL, "%s: size: 0x%x, address: 0x%" PRIx64, __PRETTY_FUNCTION__, mmap_len, bq.xfer_base);
    mmap_ptr = (uint8_t *)mmap(NULL, mmap_len, PROT_READ | PROT_WRITE, MAP_SHARED, mmap_fd, bq.xfer_base);
    if (mmap_ptr == MAP_FAILED)
    {
        LOG_UNLIKELY_ERRNO(__PRETTY_FUNCTION__, __LINE__, errno);
        return;
    }

    TP(LSSM_BQ, "%s: Read %d (0x%05X) bytes into producer's buffer #%d @ %p, from lssm #%d @ 0x%" PRIx64 " + 0x%x",
       __PRETTY_FUNCTION__, (int)nbytes, (int)nbytes, bq.producer_index(), (void *)dst, bq.buf_id, bq.xfer_base, bq.xfer_offset);

    // Copy of the data.
    memcpy(dst, mmap_ptr + bq.xfer_offset, nbytes);

    // Unmap our memory.
    munmap(mmap_ptr, mmap_len);
}

void LSSM_Thread::fault_recovery(BufferQueue& bq)
{
    TP(LSSM_FUNC, "%s", __PRETTY_FUNCTION__);

    lssm_lock(shmem);
    lssm_client_clear_fault_l(shmem);
    bq.fault = false;
    bq.bytes_to_read_from_im = 0;
    bq.bytes_to_read_from_header = 0;
    bq.buf_id = -1;
    lssm_unlock(shmem);
}
