/*
 * status_update.cxx
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>

#include <fastec/gige_msg.h>
#include <fastec/fas_log.h>

#include "status_update.h"
#include "fcp.h"
#include "fcp_datamgr.h"

//--------------------------------------------------------------------
// Constant Declarations or Enumerations Used in class
//--------------------------------------------------------------------

// The number of nanoseconds to wait when we're polling.
const uint64_t	SU_POLL_TIME_NS	= 250000000L;

// The number of nanoseconds per second.
const uint64_t NSEC_PER_SEC	= 1000000000L;

// The status update states.
enum SU_STATE
{
    SU_STATE_EXIT,		//!< Terminate the thread.
    SU_STATE_IDLE,		//!< Stop polling and wait for the next command.
    SU_STATE_POLL_RECORD	//!< Start polling record information.
};

StatusUpdate::StatusUpdate()
{
    int status;
    pthread_attr_t attr;

    pthread_mutex_init(&lock, NULL);
    pthread_cond_init(&notify_ready, NULL);

    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    if ((status = pthread_create(&thread, &attr, StatusUpdate::entry, (void*)this)))
    {
        int save_errno = errno;
        fi_log(LOG_ERR, "%s: Unable to create thread! [%d: %s]\n", __PRETTY_FUNCTION__, save_errno, strerror(save_errno));
    }
    pthread_attr_destroy(&attr);

    // Register with the camera service to receive notifications.
    cs_register_callback( su_callback, (void*)this );
}

StatusUpdate::~StatusUpdate()
{
    void *rval;

    // Lock our mutex.
    pthread_mutex_lock( &lock );

    // Tell the thread what to do.
    update_state = SU_STATE_EXIT;
    pthread_cond_signal( &notify_ready );

    // Release the lock.
    pthread_mutex_unlock( &lock );

    // Wait for the thread to exit.
    pthread_join( thread, &rval );

    // Clean it all up.
    pthread_mutex_destroy(&lock);
    pthread_cond_destroy(&notify_ready);
}

void StatusUpdate::cs_notify( uint32_t reg_addr, uint32_t reg_value )
{
    uint32_t old_state = update_state;

    // Only interested in state changes.
    if (reg_addr != GIGE_REG_CAMERA_STATE)
        return;

    // Lock our mutex.
    pthread_mutex_lock( &lock );

    // What is the new state?
    switch (reg_value)
    {
    case GIGE_CAMERA_STATE_ARMED:
    case GIGE_CAMERA_STATE_TRIGGERED:
        update_state = SU_STATE_POLL_RECORD;
        break;
    case GIGE_CAMERA_STATE_IDLE:
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
    case GIGE_CAMERA_STATE_REVIEW:
    case GIGE_CAMERA_STATE_SNAPSHOT:
    case GIGE_CAMERA_STATE_AUTOSAVE:
    case GIGE_CAMERA_STATE_DEARMING:
    default:
        update_state = SU_STATE_IDLE;
        break;
    }

    // Any change?
    if (update_state != (int)old_state)
    {
        // Tell the thread what to do.
        pthread_cond_signal( &notify_ready );
    }

    // Release the lock.
    pthread_mutex_unlock( &lock );
}

void StatusUpdate::run(void * /*t*/)
{
    bool exit_loop = false;
    int	new_state;
    int status;
    struct timespec poll_time;
    struct timespec time_wait;

    // Wait until we are signaled.
    poll_time.tv_sec = 0;
    poll_time.tv_nsec = 0;

    // Loop until we're told to exit.
    while ( !exit_loop )
    {
        // Wait for someone to get our attention.
        pthread_mutex_lock(&lock);

        // Figure out which wait to use.
        if ((poll_time.tv_sec == 0) && (poll_time.tv_nsec == 0))
        {
            // Wait indefinitely.
            status = pthread_cond_wait( &notify_ready, &lock );
        }
        else
        {
            // Figure out when we should timeout the wait.
            clock_gettime( CLOCK_REALTIME, &time_wait );
            time_wait.tv_nsec += poll_time.tv_nsec;
            if ( time_wait.tv_nsec > (int64_t)NSEC_PER_SEC )
            {
                time_wait.tv_nsec -= NSEC_PER_SEC;
                time_wait.tv_sec++;
            }
            time_wait.tv_sec += poll_time.tv_sec;

            // Wait around for a bit.
            status = pthread_cond_timedwait( &notify_ready, &lock, &time_wait );
        }

        // Save the state information.
        new_state = update_state;
        if ( status && ( status != ETIMEDOUT ))
        {
            fi_log(LOG_ERR, "%s: Error on cond_wait()! [%d: %s]\n", __PRETTY_FUNCTION__, errno, strerror(errno));
            new_state = SU_STATE_EXIT;
        }

        // Release the lock.
        pthread_mutex_unlock( &lock );

        // Process based on new state information.
        switch ( new_state )
        {
        case SU_STATE_EXIT:
            exit_loop = true;
            break;
        default:
        case SU_STATE_IDLE:
            // Wait until we are signaled.
            poll_time.tv_sec = 0;
            poll_time.tv_nsec = 0;
            break;
        case SU_STATE_POLL_RECORD:
            // Wait until we are signaled or we timeout.
            poll_time.tv_sec = 0;
            poll_time.tv_nsec = SU_POLL_TIME_NS;

            // Now, go poll some registers!
            record_poll();
            break;
        }
    }

    // We're done.
    pthread_exit(NULL);
}

void StatusUpdate::record_poll( void )
{
    const uint32_t rec_regs[] = { GIGE_REG_ACTIVE_PART, GIGE_REG_USER_SAVE_START_NUM, GIGE_REG_USER_SAVE_END_NUM,
                                  GIGE_REG_AUTOSAVE, GIGE_REG_TRIGGER_REQ_POS, GIGE_REG_CUR_CAPTURE_COUNT,
                                  GIGE_REG_CUR_FRAME_POS
                                };

    int	i;
    uint32_t reg_value;

    // Loop through all the registers we're interested in.
    for (i = 0; i < (int)(sizeof(rec_regs) / sizeof(uint32_t)); ++i )
    {
        // Read the register from the camera server.
        cs_get_reg_l(rec_regs[i], &reg_value);

        // Pass it on to the listener object.
        cs_fcp_reg_notify( NULL, rec_regs[i], reg_value );
    }
}
